package be.iminds.ilabt.jfed.lowlevel.api.test;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.AnyCredential;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.ServerType;
import be.iminds.ilabt.jfed.lowlevel.api.AbstractGeniAggregateManager;
import be.iminds.ilabt.jfed.lowlevel.api.StitchingComputationService;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.*;
import be.iminds.ilabt.jfed.lowlevel.stitching.StitchingDirector;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.rspec.request.geni_rspec_3.*;
import be.iminds.ilabt.jfed.testing.base.ApiTest;
import be.iminds.ilabt.jfed.util.IOUtils;
import be.iminds.ilabt.jfed.util.TargetAuthority;
import be.iminds.ilabt.jfed.util.TextUtil;
import ch.ethz.ssh2.Connection;
import ch.ethz.ssh2.Session;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.Unmarshaller;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StringReader;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.security.NoSuchAlgorithmException;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * StitchingPingTest:
 * - create a stitched RSpec, contact SCS and do all needed CreateSlivers.
 * - Then test login on node and connectivity using ping.
 * - finally, Delete everything
 */
public class StitchingPingTest extends ApiTest {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(StitchingPingTest.class);

    String scsUrl;
    String scsUrn;
    List<String> stitchedAuthorityUrns;
    List<SfaAuthority> stitchedAuthorities;
    SfaAuthority scsAuth;
    private Boolean createSliceIfNotExists = true;
    private Boolean reuseSliceIfExists = false;
    private Boolean disableComplianceTests = false;
    private Boolean disableCleanup = false;
    private CommonAMTest commonAMTest;
    private Map<SfaAuthority, NodeLoginTester> nodeLoginTesters = new HashMap<SfaAuthority, NodeLoginTester>();
    private List<AnyCredential> userCredentials;
    private CommonAMTest.SliceInfo slice;
    private String rspecXmlStr;
    private List<Node> nodes = new ArrayList<Node>();
    private List<Link> links = new ArrayList<Link>();
    private StitchingDirector director;
    private HandleUntrustedCallback handleUntrustedCallback = null;

    private Date expirationDate;

    @Inject
    public StitchingPingTest(Logger logger, TargetAuthority testedAuthority, GeniUser user, JFedConnectionProvider connectionProvider, AuthorityListModel authorityListModel, JFedPreferences jFedPreferences) {
        super(logger, testedAuthority, user, connectionProvider, authorityListModel, jFedPreferences);
    }

    @Override
    public List<String> getReqConfKeys() {
        List<String> res = new ArrayList<String>();
        res.add("stitchedAuthorityUrns"); //space seperated list
        return res;
    }

    @Override
    public List<String> getOptConfKeys() {
        List<String> res = new ArrayList<String>();
        res.add("topology");
        res.add("rspecfile");

        /* fixed_ssh_private_key_file and fixed_ssh_public_key_file option:
              There are 3 cases:
                  option not specified: use randomly generated key that is different each time
                  option specified, with private key file specified: use public and private key from files (public key can be read from PEM certificate)
                  option specified, with private key file specified, AND this file is the same file for both and is a user login file (key+cert file): use public and private key from file.
         */
        res.add("fixed_ssh_public_key_file"); //needed if fixed_ssh_private_key_file is specified (file containing PEM x509certificate is also allowed)
        res.add("fixed_ssh_private_key_file"); //needed if fixed_ssh_public_key_file is specified
        res.add("fixed_ssh_private_key_password"); //may be null,even if fixed_ssh_private_key_file is not (in case no password is needed)

        res.add("disable_slice_cleanup"); //no deleteSliver call

        //there are not much compliance tests here, so no need for this  res.add("disable_compliance_tests"); //test only for bare functionality, do not do extra checks

        res.add("create_slice_if_not_exists");
        res.add("reuse_slice_if_exists");
        res.add("fixed_slice_name"); //uses this slice name. If not specified, will generate a slice name automatically
        res.add("fixed_project_name"); //uses this as the project/subauthority. If not specified, will choose one if needed, or not use one otherwise.
        res.add("try_to_use_project_for_slices"); //if set and no project name set, this will try to find any project for the user, and use it to create slices.

        //fallback to default is not specified
        res.add("scsUrn");
        res.add("scsUrl");

        //future
//        res.add("minimum_slice_lifetime"); //minimum lifetime for slice
//        res.add("minimum_sliver_lifetime"); //minimum lifetime for slice

        return res;
    }

    @Override
    public String getTestDescription() {
        return "Test Stitching between multiple authorities, including node login and ping test between nodes. This version uses the AggregateManagerWrapper, and is thus not limited to AMv2. (and will actually prefer AMv3 if available.)";
    }

    private AggregateManagerWrapper getAMWrapper(SfaAuthority auth) {
        return getAutomaticAggregateManagerWrapperFactory().create(auth);
    }

    @Override
    public void setUp() {
        stitchedAuthorityUrns = getStitchedAuthorityUrns();
        stitchedAuthorities = new ArrayList<SfaAuthority>();
        for (String urn : stitchedAuthorityUrns) {
            SfaAuthority auth = authorityListModel.getByUrnExact(urn);
            if (auth == null)
                errorFatal("Authority \"" + urn +"\" is not known.");
            stitchedAuthorities.add(auth);
        }

        commonAMTest = new CommonAMTest(this, user.getUserAuthority(), stitchedAuthorities);

        // this is production, but iMinds was removed begin March ??
        scsUrl = "http://oingo.dragon.maxgigapop.net:8081/geni/xmlrpc";
        scsUrn = "urn:publicid:IDN+oingo.dragon.maxgigapop.net+auth+scs";
        // this is development
//        scsUrl = "http://geni.maxgigapop.net:8081/geni/xmlrpc";
//        scsUrn = "urn:publicid:IDN+geni.maxgigapop.net+auth+scs";
        if (getTestConfig().getProperty("scsUrl") != null) scsUrl = getTestConfig().getProperty("scsUrl");
        if (getTestConfig().getProperty("scsUrn") != null) scsUrn = getTestConfig().getProperty("scsUrn");
        note("Using SCS " + scsUrn + " @ " + scsUrl);

        assert authorityListModel != null;

        String topologyType = getTestConfig().getProperty("topology");
        String rspecFile = getTestConfig().getProperty("rspecfile");
        if (topologyType == null || topologyType.isEmpty() || topologyType.equals("ring") || topologyType.equals("line") || topologyType.equals("snake")) {
            //supported topology type
        } else
            errorFatal("Unknown topology of type \"" + topologyType + "\"");

        if (rspecFile != null && !rspecFile.isEmpty())
            note("Will use user specified Rspec file (\"" + rspecFile + "\") instead of automatically generating rspec.");
//            errorFatal("support for rspecfile parameter not yet implemented. \""+rspecFile+"\"");

        createSliceIfNotExists = TextUtil.objectToBoolean(getTestConfig().getProperty("create_slice_if_not_exists"));
        reuseSliceIfExists = TextUtil.objectToBoolean(getTestConfig().getProperty("reuse_slice_if_exists"));
        disableComplianceTests = TextUtil.objectToBoolean(getTestConfig().getProperty("disable_compliance_tests"));
        disableCleanup = TextUtil.objectToBoolean(getTestConfig().getProperty("disable_slice_cleanup"));
        if (createSliceIfNotExists == null) createSliceIfNotExists = true;
        if (reuseSliceIfExists == null) reuseSliceIfExists = false;
        if (disableComplianceTests == null) disableComplianceTests = false;
        if (disableCleanup == null) disableCleanup = false;

        expirationDate = new Date(System.currentTimeMillis() + (2*60*60*1000L)); //2 hours
    }

    public List<String> getStitchedAuthorityUrns() {
        List<String> res = new ArrayList<String>();
        String urnList = getTestConfig().getProperty("stitchedAuthorityUrns");
        if (urnList == null)
            throw new RuntimeException("StitchingPingTest required config key stitchedAuthorityUrns is not present.");
        String[] urnsArray = urnList.split(" ");
        for (String urn : urnsArray)
            res.add(urn);
        if (res.isEmpty())
            throw new RuntimeException("StitchingPingTest required config key stitchedAuthorityUrns (space seperated list) does not contain urns: " + urnList);
        if (res.size() == 1)
            throw new RuntimeException("StitchingPingTest required config key stitchedAuthorityUrns (space seperated list) contains only 1 urn: " + urnList);
        return res;
    }

    private List<AnyCredential> getUserCredentialList() {
        assert userCredentials != null;
        assert !userCredentials.isEmpty();
        return userCredentials;
    }

    public SfaConnection getSAConnection() throws JFedException {
        return (SfaConnection) connectionProvider.getConnectionByAuthority(user, user.getUserAuthority(), new ServerType(ServerType.GeniServerRole.PROTOGENI_SA, 1));
    }

    @Test()
    public void getUserCredentials() throws JFedException {
        userCredentials = commonAMTest.getUserCredentials();
    }

    @Test(hardDepends = {"getUserCredentials"}, groups = {"createslice"})
    public void createSlice() throws JFedException {
        //create a slice to use in AM tests
        String fixedSliceName = getTestConfig().getProperty("fixed_slice_name");
        String fixedProjectName = getTestConfig().getProperty("fixed_project_name"); //may be null, in that case, no sub authority is used, unless mandatory
        boolean tryToUseProject = TextUtil.objectToBoolean(getTestConfig().getProperty("try_to_use_project_for_slices"))  != Boolean.FALSE; //so default is tru

        if (reuseSliceIfExists && fixedSliceName != null) {
            //check if exists: reuseSlice returns info if exists, null otherwise
            slice = commonAMTest.reuseSlice(fixedSliceName);

            if (slice != null)
                return;

            //else fall back to creating slice
            if (!createSliceIfNotExists)
                throw new RuntimeException("create_slice_if_not_exists is false, and slice does not exist.");
        }

        if (fixedSliceName == null)
            slice = commonAMTest.createSlice("t", fixedProjectName, tryToUseProject);
        else
            slice = commonAMTest.createSlice(fixedSliceName, false, fixedProjectName, tryToUseProject);

        assertNotNull(slice);
        assertNotNull(slice.urn);
        assertNotNull(slice.name);
        assertNotNull(slice.urnString);
        assertNotNull(slice.credentials);
        for (AnyCredential c : slice.credentials) {
            assertNotNull(c.getCredentialXml());
            assertTrue(c.getCredentialXml().length() > 10, "credential too short " + c.getCredentialXml());
        }
    }

    private List<String> allIps() {
        List<String> res = new ArrayList<String>();
        for (Node node : nodes)
            for (Iface iface : node.ifaces)
                res.add(iface.ip);
        return res;
    }

    private List<String> localIps(SfaAuthority auth) {
        List<String> res = new ArrayList<String>();
        for (Node node : nodes)
            if (node.authUrn.equals(auth.getUrnString()))
                for (Iface iface : node.ifaces)
                    res.add(iface.ip);
        return res;
    }

    private List<String> pingTargetsForAuth(SfaAuthority auth) {
        List<String> res = new ArrayList<String>();
        for (Node node : nodes)
            if (node.authUrn.equals(auth.getUrnString()))
                for (Iface iface : node.ifaces)
                    for (Iface otherSideIface : iface.link.otherIfaces(iface))
                        res.add(otherSideIface.ip);
        return res;
    }

    /**
     * returns true if the authority is one of the authorities with nodes, and not just one of the stitching link auths
     */
    private boolean isNodeAuth(SfaAuthority auth) {
        return stitchedAuthorityUrns.contains(auth.getUrnString());
    }

    private boolean isSiteNodeExclusive(String authUrn) {
//        SfaAuthority auth = authorityListModel.getByUrn(authUrn);
//        assert auth != null;
//        boolean isExclusive = CommonAMTest.getExclusiveForSite(auth);

        if (authUrn.startsWith("urn:publicid:IDN+wall2.ilabt.iminds.be+authority+"))
            return true;
        return false;
    }

    private String getSiteComponentManagerId(String authUrn) {
        SfaAuthority auth =
                authorityListModel.getByUrnExact(authUrn);
        assert auth != null;
        return CommonAMTest.getComponentManagerIdForSite(auth);
    }

    private String getSiteNodeSliverType(String authUrn) {
        SfaAuthority auth =
                authorityListModel.getByUrnExact(authUrn);
        assert auth != null;
        String sliverType = CommonAMTest.getSliverTypeForSite(auth);

        if (authUrn.startsWith("urn:publicid:IDN+wall2.ilabt.iminds.be+authority+") ||
                authUrn.startsWith("urn:publicid:IDN+wall1.ilabt.iminds.be+authority+"))
            return "raw-pc";
        if (sliverType != null && sliverType.equals("raw-pc"))
            return "emulab-openvz";

        if (sliverType == null) //fallback. will probably fail
            return "emulab-openvz";

        return sliverType;
    }

    private void readRspec(String inputRspec) {
        rspecXmlStr = inputRspec;
        try {
            JAXBContext jc =
                    JAXBContext.newInstance(be.iminds.ilabt.jfed.rspec.request.geni_rspec_3.RSpecContents.class);
            Unmarshaller u = jc.createUnmarshaller();
            //            JAXBElement<RSpecContents> doc = (JAXBElement<RSpecContents>) u.unmarshal( new FileInputStream(rspecFile));
            JAXBElement<be.iminds.ilabt.jfed.rspec.request.geni_rspec_3.RSpecContents> doc =
                    (JAXBElement<be.iminds.ilabt.jfed.rspec.request.geni_rspec_3.RSpecContents>) u.unmarshal(new StringReader(inputRspec));
            be.iminds.ilabt.jfed.rspec.request.geni_rspec_3.RSpecContents c = doc.getValue();

            assert c != null;

            //Note: getType() will return null if the type is specified but not "request".
            //      This is because the xsd specifies type as a choice from a list containing only "request".
            if (c != null &&
                    c.getType() != null &&
                    c.getType().value() != null) {
                if (!c.getType().value().equals("request"))
                    LOG.warn("Parsing an RSpec " + c.getType().value() + " as an RSpec request!");
            } else {
                LOG.warn("Parsed RSpec has no type (note: this also occurs if the type is not \"request\"): " + inputRspec);
            }

            Map<String, Iface> nameToIface = new HashMap<String, Iface>();
            Map<String, Node> ifaceNameToRspecNode = new HashMap<String, Node>();

            //first nodes
            for (Object o : c.getAnyOrNodeOrLink()) {
                if (o instanceof JAXBElement) {
                    JAXBElement el = (JAXBElement) o;
                    if (el.getValue() instanceof NodeContents) {
                        NodeContents node = (NodeContents) el.getValue();

                        String compManagerId = node.getComponentManagerId();
                        String id = node.getClientId();
                        //in an advertisement, client ID is typically empty, but component name and ID are set
                        if (id == null) id = node.getComponentName();
                        if (id == null) id = node.getComponentId();

                        assert id != null : "Rspec has a node without any ID: " + el;
                        Node resNode = new Node(id, compManagerId);
                        nodes.add(resNode);

                        for (Object nodeElO : node.getAnyOrRelationOrLocation()) {
                            if (nodeElO instanceof JAXBElement) {
                                JAXBElement nodeEl = (JAXBElement) nodeElO;
                                if (nodeEl.getValue() instanceof InterfaceContents) {
                                    InterfaceContents ic = (InterfaceContents) nodeEl.getValue();
                                    assert ic != null;

                                    //for a request, this is mandatory, but we use the same code to parse the advertisement
                                    //assert ic.getClientId() != null : "InterfaceContents has no clientId: "+ic;
                                    if (ic.getClientId() != null) {
                                        //check if iface with same name is not already added
                                        for (String existingIfaceName : ifaceNameToRspecNode.keySet())
                                            if (ic.getClientId().equals(existingIfaceName))
                                                throw new RuntimeException("Duplicate interfaces in XML Rspec: id=\"" + existingIfaceName + "\"");

                                        ifaceNameToRspecNode.put(ic.getClientId(), resNode);
                                    } else {
                                        LOG.warn("ignoring that InterfaceContents has no clientId");
                                    }
                                }
                            }
                        }
                    }
                }
            }

            //then the rest
            for (Object o : c.getAnyOrNodeOrLink()) {
                if (o instanceof JAXBElement) {
                    JAXBElement el = (JAXBElement) o;
                    boolean handled = false;
                    if (el.getValue() instanceof NodeContents)
                        handled = true;
                    if (el.getValue() instanceof LinkContents) {
                        handled = true;
                        LinkContents link = (LinkContents) el.getValue();

                        String id = link.getClientId();
                        //in an advertisement, client ID is typically empty, but component name and ID are set
                        if (id == null) id = link.getOtherAttributes().get("component_name");
                        if (id == null) id = link.getOtherAttributes().get("component_id");

                        assert id != null : "Rspec has a link without any ID: " + el;

                        List<String> linkComponentManagerUrns = new ArrayList<String>();
                        for (Object oo : link.getAnyOrPropertyOrLinkType()) {
                            //sometimes, it's wrapped in a JAXBElement
                            Object value = oo;
                            if (o instanceof JAXBElement)
                                value = ((JAXBElement) oo).getValue();

                            if (value instanceof ComponentManager) {
                                ComponentManager cm = (ComponentManager) value;
                                linkComponentManagerUrns.add(cm.getName());
                            }
                        }

                        if (linkComponentManagerUrns.size() < 2) {
                            //link without stitching
                            continue;
                        }

                        Link resLink = new Link(); //iface creation will add links
                        links.add(resLink);

                        //check if all interfaces of link are known. We will ignore the link if they are not.
                        for (Object linkElO : link.getAnyOrPropertyOrLinkType()) {
                            if (linkElO instanceof JAXBElement) {
                                JAXBElement linkEl = (JAXBElement) linkElO;
                                if (linkEl.getValue() instanceof InterfaceRefContents) {
                                    InterfaceRefContents ic = (InterfaceRefContents) linkEl.getValue();

                                    Iface iface = nameToIface.get(ic.getClientId());
                                    if (iface == null) {
                                        Node rspecNode = ifaceNameToRspecNode.get(ic.getClientId());

                                        if (rspecNode == null) {
                                            LOG.warn("Interface has not been defined in a <node>, but is referred to in a <link>: " + ic.getClientId());
                                        } else
                                            iface =
                                                    new Iface(rspecNode, resLink); //constructor links Node and Link automatically
                                    }
                                }
                            }
                        }
                    }
                }
            }
        } catch (Throwable e) {
            LOG.warn("Error reading Rspec XML (ping test will not be possible due to this): " + e.getMessage(), e);
            nodes.clear();
            links.clear();
        }
    }

    @Test(groups = {"init"})
    public void generateRspec() throws JFedException, IOException {
        String topologyType = getTestConfig().getProperty("topology");
        String rspecFile = getTestConfig().getProperty("rspecfile");

        if (rspecFile != null && !rspecFile.isEmpty()) {
//            errorFatal("support for rspecfile parameter not yet implemented. \""+rspecFile+"\"");
//            return;

            readRspec(IOUtils.fileToString(rspecFile));

            Set<String> authUrns = new HashSet<String>();
            for (Node node : nodes)
                authUrns.add(node.authUrn);

            note("Parsed Rspec file. node count=" + nodes.size() + "   stitching link count=" + links.size());

            note("  Authority URNs=" + authUrns);
            stitchedAuthorityUrns = new ArrayList<String>(authUrns);

            note("  Stitching IPs=" + allIps());

            return;
        }


        assert authorityListModel != null;
        assert stitchedAuthorityUrns != null;
        assert stitchedAuthorityUrns.size() > 1;

        note("creating topology between " + stitchedAuthorityUrns.size() + " component managers: " + stitchedAuthorityUrns);

        //create nodes
        for (String authUrn : stitchedAuthorityUrns) {
            Node n = new Node(authUrn);
        }

        //create link(s)
        assert nodes.size() > 1;
        if (topologyType == null || topologyType.isEmpty() || topologyType.equals("ring")) {
            if (nodes.size() == 2) {
                note("making topology: simple 2 node connection");
                //create link if there are only 2 nodes
                new Link(nodes.get(0), nodes.get(1));
            } else {
                note("making topology: ring with " + nodes.size() + " nodes");
                //create links (ring topo)
                Node firstNode = nodes.get(0);
                Node lastNode = nodes.get(nodes.size() - 1);
                Node prevNode = lastNode;
                for (Node n : nodes) {
                    new Link(n, prevNode);
                    prevNode = n;
                }
            }
        } else {
            if (topologyType.equals("line") || topologyType.equals("snake")) {
                if (nodes.size() == 2) {
                    note("making topology: simple 2 node connection");
                    //create link if there are only 2 nodes
                    new Link(nodes.get(0), nodes.get(1));
                } else {
                    note("making topology: line with " + nodes.size() + " nodes");
                    //create links
                    Node prevNode = null;
                    for (Node n : nodes) {
                        if (prevNode != null)
                            new Link(n, prevNode);
                        prevNode = n;
                    }
                }
            } else {
                errorFatal("Unknown topology of type \"" + topologyType + "\"");
            }
        }

        //create Rspec
        String nodesXml = "";
        String linksXml = "";
        for (Node node : nodes) {
            String nodeXml =
                    "  <node client_id=\"" + node.name + "\" component_manager_id=\"" + getSiteComponentManagerId(node.authUrn) + "\" exclusive=\"" + (
                            isSiteNodeExclusive(node.authUrn) ? "true" : "false") + "\">\n" +
                            "    <sliver_type name=\"" + getSiteNodeSliverType((node.authUrn)) + "\"/>\n";
//                    "    <sliver_type name=\"emulab-openvz\"/>\n";

            for (Iface iface : node.ifaces)
                nodeXml += "    <interface client_id=\"" + iface.getName() + "\">\n" +
                        "      <ip address=\"" + iface.ip + "\" netmask=\"255.255.255.0\" type=\"ipv4\"/>\n" +
                        "    </interface>\n";

            nodeXml += "  </node>\n";
            nodesXml += nodeXml;
        }

        for (Link link : links) {
            String linkXml = "  <link client_id=\"" + link.name + "\">\n";
            for (String urn : link.componentManagerUrns())
                linkXml += "    <component_manager name=\"" + getSiteComponentManagerId(urn) + "\"/>\n";
            for (Iface iface : link.ifaces)
                linkXml += "    <interface_ref client_id=\"" + iface.getName() + "\"/>\n";


            for (int ifaceIndex1 = 0; ifaceIndex1 < link.ifaces.size(); ifaceIndex1++)
                for (int ifaceIndex2 = ifaceIndex1 + 1; ifaceIndex2 < link.ifaces.size(); ifaceIndex2++) {
                    Iface iface1 = link.ifaces.get(ifaceIndex1);
                    Iface iface2 = link.ifaces.get(ifaceIndex2);
                    //capacity is in demo rspec, and is probably needed:
                    //    I got an error with text "*** ERROR: vtopgen: Cannot mix trivial_ok|emulated with * bw" on emulab.net when not using it
                    //       (but that could also have other cause)
                    linkXml +=
                            "    <property source_id=\"" + iface1.getName() + "\" dest_id=\"" + iface2.getName() + "\" capacity=\"100000\"/>\n" +
                                    "    <property source_id=\"" + iface2.getName() + "\" dest_id=\"" + iface1.getName() + "\" capacity=\"100000\"/>\n";
//                    linkXml += "    <property source_id=\""+iface1.getName()+"\" dest_id=\""+iface2.getName()+"\"/>\n" +
//                               "    <property source_id=\""+iface2.getName()+"\" dest_id=\""+iface1.getName()+"\"/>\n";
                }
            linkXml += "  </link>\n";

            linksXml += linkXml;
        }

        rspecXmlStr = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" +
                "<rspec type=\"request\" \n" +
                "    xmlns=\"http://www.geni.net/resources/rspec/3\" \n" +
                "    xmlns:planetlab=\"http://www.planet-lab.org/resources/sfa/ext/planetlab/1\" \n" +
                "    xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" \n" +
                "       xmlns:stitch=\"http://hpn.east.isi.edu/rspec/ext/stitch/0.1/\" \n" +
                "    xsi:schemaLocation=\"http://www.geni.net/resources/rspec/3 \n" +
                "    http://www.geni.net/resources/rspec/3/request.xsd\">  \n" +
                nodesXml +
                linksXml +
                "</rspec>";

        LOG.info("Generated Rspec: " + rspecXmlStr);
        note("Generated Rspec: " + rspecXmlStr);
    }

    @Test(hardDepends = {"createSlice", "generateRspec"}, groups = {"init"})
    public void initStitching() throws JFedException, NoSuchAlgorithmException, MalformedURLException {
        director = new StitchingDirector(authorityListModel, authorityFinder, expirationDate);


        //create stitching auth if doesn't exist
        scsAuth = null;
        if (authorityListModel.getByUrnExact(scsUrn) == null) {
            Map<ServerType, URL> urlMap = new HashMap<ServerType, URL>();
            urlMap.put(new ServerType(ServerType.GeniServerRole.SCS, 1), new URL(scsUrl));
            scsAuth = new SfaAuthority(scsUrn, "Stitching Test SCS", null, urlMap, null/*proxies*/, null, "scs");
            authorityListModel.addAuthority(scsAuth);
            authorityListModel.fireChange();

            assert authorityListModel.getByUrnExact(scsUrn) != null;
        } else
            scsAuth = authorityListModel.getByUrnExact(scsUrn);
    }

    @Test(hardDepends = {"initStitching"}, groups = {"scs"})
    public void callSCS() throws JFedException {
        assertNotNull(slice);
        assert rspecXmlStr != null;

        //contact stitcher
        StitchingComputationService scs = new StitchingComputationService(logger, jFedPreferences);
        SfaConnection con =
                new SfaPlainHttpConnection(scsAuth.getUrl(ServerType.GeniServerRole.SCS, 1).toExternalForm(), scsAuth,  null/*proxyInfo*/);

        StitchingComputationService.SCSReply<StitchingComputationService.ComputePathResult> scsReply =
                scs.computePath(con, slice.urnString, rspecXmlStr, null);

        assertNotNull(scsReply);
        assertEquals(scsReply.getCode(), 0, "SCS call Result code is not 0 but " + scsReply.getCode());
        assertNotNull(scsReply.getValue());

        StitchingComputationService.ComputePathResult computePathResult = scsReply.getValue();

        director.setComputePathResult(computePathResult);

        assertTrue(director.getAllAuthorities().size() > 0);
        assertTrue(director.getHopsLeft().size() > 0);
        assertTrue(director.areAnyActionsLeft());
        assertEquals(director.getHopsLeft().size(), director.getAllAuthorities().size(),
                "director getHopsLeft().size() != getAllAuthorities().size()  -> " +
                        director.getHopsLeft().size() + " != " + director.getAllAuthorities().size()
        );
    }

    public void setInsecure() {
        handleUntrustedCallback =
                new HttpsClientWithUserAuthenticationFactory.INSECURE_TRUSTALL_HandleUntrustedCallback();
    }

    List<SfaAuthority> toProvision = new ArrayList<SfaAuthority>();
    List<SfaAuthority> toWait = new ArrayList<SfaAuthority>();
    List<SfaAuthority> toDescribe = new ArrayList<SfaAuthority>();

    @Test(hardDepends = {"callSCS", "createSlice"}, groups = {"createslivernowait"})
    public void doStitchActions() throws JFedException, URISyntaxException, NoSuchAlgorithmException {
        assertNotNull(slice.credentials);

        String fixedSshPublicKeyFile = getTestConfig().getProperty("fixed_ssh_public_key_file");
        String fixedSshPrivateKeyFile = getTestConfig().getProperty("fixed_ssh_private_key_file");
        String fixedSshPrivateKeyPassword = getTestConfig().getProperty("fixed_ssh_private_key_password");

        setInsecure();

        while (director.areAnyActionsLeft()) {
            List<StitchingDirector.StitchAction> stitchActions = director.getStitchActions();

            String stitchActionAuthString = "";
            for (StitchingDirector.StitchAction stitchAction : stitchActions)
                stitchActionAuthString += " " + stitchAction.getAuthority().getUrnString();
            note("Stitch Actions needed " + stitchActions.size() + " on: " + stitchActionAuthString);
            LOG.info("Stitch Actions needed " + stitchActions.size() + " on: " + stitchActionAuthString);

            if (stitchActions.isEmpty())
                errorFatal("No stitch actions left: implementation bug.");

            for (StitchingDirector.StitchAction stitchAction : stitchActions) {
                SfaAuthority auth = stitchAction.getAuthority();

                note("Handling StitchAction "+stitchAction.getType()+" on "+auth.getUrnString());

                switch (stitchAction.getType()) {
                    case DELETE: {
                        AggregateManagerWrapper amWrapper = getAMWrapper(auth);

                        try {
                            note("Calling Delete on " + auth.getUrnString() + ", as requested by stitching director.");
                            amWrapper.deleteSliver(slice.urn, slice.credentials);
                        } catch (JFedException e) {
                            //delete call can fail but still work, so we will ignore error
                            errorNonFatal("amWrapper.deleteSliver threw JFedException: "+e.getMessage()+" for "+e.getXmlRpcResult());
                        } finally {
                            toWait.remove(auth);
                            toProvision.remove(auth);
                            director.processDeleteResult(auth);
                        }
                      break;
                    }
                    case LISTRESOURCES:{
                        AggregateManagerWrapper amWrapper = getAMWrapper(auth);

                        String advertisementRspec = null;
                        try {
                            advertisementRspec = amWrapper.listResources(getUserCredentialList(), true/*available*/);
                            note("Called ListResources on " + auth.getUrnString()+". " +
                                    "Got advertisement RSpec of "+(advertisementRspec == null ? "null" : advertisementRspec.length())+" chars");

                            director.processListResourcesResult(auth, advertisementRspec);
                        } catch (JFedException e) {
                            errorFatal("amWrapper.listResources threw JFedException: "+e.getMessage()+" for "+e.getXmlRpcResult());
                        }
                      break;
                    }
                    case ALLOCATE: {
                        String requestRspec = stitchAction.getRequestRspec();

                        note("   CreateSliver call for " + auth.getUrnString());
                        LOG.info("   CreateSliver call for hop=" + stitchAction);
                        LOG.info("   CreateSliver call for auth=" + auth.getUrnString());
                        NodeLoginTester nodeLoginTester =
                                new NodeLoginTester(this, fixedSshPublicKeyFile, fixedSshPrivateKeyFile, fixedSshPrivateKeyPassword);
                        nodeLoginTesters.put(auth, nodeLoginTester);

                        List users = new ArrayList<UserSpec>();
                        UserSpec userSpec =
                                new UserSpec(user.getUserUrnString(), nodeLoginTester.getSshKeyHelper().getSshPublicKeyString());
                        users.add(userSpec);

                        AggregateManagerWrapper amWrapper = getAMWrapper(auth);

                        String manifestRspec = null;
                        AbstractGeniAggregateManager.AggregateManagerReply reply;
                        try {
                            if (amWrapper.hasSeperateAllocateAndProvision()) {
                                String allocateManifestRspec =
                                        amWrapper.allocate(slice.urn, slice.credentials, requestRspec, expirationDate);
                                reply = amWrapper.getLastReply();
                                manifestRspec = null;
                            } else {
                                manifestRspec = amWrapper.createSliver(slice.urn, slice.credentials, requestRspec, users, expirationDate);
                                reply = amWrapper.getLastReply();
                            }
                        } catch (AggregateManagerWrapper.VlanUnavailableException e) {
                            //ignore
                            reply = (AbstractGeniAggregateManager.AggregateManagerReply) e.getSfaReply();
                        } catch (AggregateManagerWrapper.AnyVlanUnsupportedException e) {
                            //ignore
                            reply = (AbstractGeniAggregateManager.AggregateManagerReply) e.getSfaReply();
                        } catch (JFedException e) {
                            //ignore
                            reply = (AbstractGeniAggregateManager.AggregateManagerReply) e.getSfaReply();
                        }

                        LOG.trace("   Processing result of CreateSliver call for " + auth.getUrnString());
                        StitchingDirector.AllocateVerdict verdict = director.processAllocateResult(stitchAction, reply);
                        boolean failure = verdict == StitchingDirector.AllocateVerdict.GIVEUP || verdict == StitchingDirector.AllocateVerdict.CALLER_DELAYED_RETRY;
                        assertFalse(failure, "Stitching director reported failure for call with reply code=" + reply.getGeniResponseCode() + " output=" + reply.getOutput());

                        if (!failure)
                            note("Overview of hop data after CreateSliver call: \n" + director.getSuggestedAvailableOverviewsString());

                        if (verdict == StitchingDirector.AllocateVerdict.OK) {
                            if (amWrapper.hasSeperateAllocateAndProvision()) {
                                toProvision.add(auth);
                            } else {
                                toWait.add(auth);

                                assert manifestRspec != null;
                                assertFalse(nodeLoginTester.hasParsed());
                                nodeLoginTester.parseSshInfoFromGeni3ManifestRspec(manifestRspec, auth, null);
                            }
                        }
                        break;
                    } //case ALLOCATE
                    default: {
                        errorFatal("Stitch Action "+stitchAction+" is not supported by this test.");
                    }
                } //switch
            } //for stitchActions
        } //while actions left
    }

    @Test(hardDepends = {"doStitchActions"}, groups = {"createsliver"})
    public void provision() throws Throwable {
        for (SfaAuthority auth : toProvision) {
            try {
                AggregateManagerWrapper amWrapper = getAMWrapper(auth);

                NodeLoginTester nodeLoginTester = nodeLoginTesters.get(auth);
                List userSpecs = new ArrayList<UserSpec>();
                UserSpec userSpec =
                        new UserSpec(user.getUserUrnString(), nodeLoginTester.getSshKeyHelper().getSshPublicKeyString());
                userSpecs.add(userSpec);

                String manifestRspec = null;
                AbstractGeniAggregateManager.AggregateManagerReply reply;
                try {
                    manifestRspec = amWrapper.provisionAndStart(slice.urn, slice.credentials, userSpecs, expirationDate);
                    note("Called Provision on " + auth.getUrnString());
                    reply = amWrapper.getLastReply();
                } catch (JFedException e) {
                    //ignore
                    note("amWrapper.provisionAndStart threw JFedException: "+e.getMessage()+" for "+e.getXmlRpcResult());
                    reply = (AbstractGeniAggregateManager.AggregateManagerReply) e.getSfaReply();
                }

                if (reply.getGeniResponseCode().isSuccess()) {
                    assert manifestRspec != null;

                    assert nodeLoginTester != null;
                    if (nodeLoginTester != null) {
                        boolean parseFoundLoginInfo = nodeLoginTester.parseSshInfoFromGeni3ManifestRspec(manifestRspec, auth, null);
                        if (!parseFoundLoginInfo)
                            note("Found no SSH login info in manifest received at " + auth);
                    }
                }
            } catch (Throwable e) {
                errorNonFatal("Error while calling Provision and PerformOperationalAction on " + auth.getUrnString() + ".");
                throw e;
            }

            toWait.add(auth);
        }

        toProvision.clear();
    }

    @Test(hardDepends = {"provision", "doStitchActions"}, groups = {"createsliver"})
    public void waitForAllReady() throws JFedException, URISyntaxException {
        long now = System.currentTimeMillis();
        long deadline = now + (20 * 60 * 1000);

        assertTrue(toWait.size() > 0, "There are no authorities to wait for. (Actively involved count = "+director.getActivelyInvolvedAuthorities()+")");

        //Test if the slivers ever becomes ready. We wait for maximum 20 minutes.
        while (now < deadline) {
            List<SfaAuthority> sliverAuthsToCheck = new ArrayList<SfaAuthority>(toWait);
            assertTrue(sliverAuthsToCheck.size() > 0);

            for (SfaAuthority auth : sliverAuthsToCheck) {
                now = System.currentTimeMillis();
                LOG.info("   Calling SliverStatus on " + auth.getUrnString() + ". Deadline in: " + (deadline - now) + " ms");

                AggregateManagerWrapper amWrapper = getAMWrapper(auth);

                StatusDetails statusDetails = null;
                AbstractGeniAggregateManager.AggregateManagerReply reply;
                try {
                    statusDetails = amWrapper.status(slice.urn, slice.credentials);
                    reply = amWrapper.getLastReply();
                } catch (JFedException e) {
                    //ignore
                    note("amWrapper.status threw JFedException: "+e.getMessage()+" for "+e.getXmlRpcResult());
                    reply = (AbstractGeniAggregateManager.AggregateManagerReply) e.getSfaReply();
                }

                if (reply == null)
                    errorFatal("sliver status call failed (reply == null) " + " @ " + auth.getUrn());
                if (!reply.getGeniResponseCode().isSuccess())
                    errorFatal("sliver status call failed (code="+reply.getGeniResponseCode()+")" + " @ " + auth.getUrn());
                if (statusDetails == null)
                    errorFatal("sliver status call failed (returned StatusDetails == null) " + " @ " + auth.getUrn());

                if (statusDetails.getGlobalStatus().equals(StatusDetails.SliverStatus.FAIL)) {
                    errorNonFatal("testCreatedSliverBecomesReady -> sliver failed instead of becoming ready: " + statusDetails.getGlobalStatus() + "" + " @ " + auth.getUrn());
                    errorFatal("sliver failed " + " @ " + auth.getUrn());
                }


                if (statusDetails.getGlobalStatus().equals(StatusDetails.SliverStatus.READY)) {
                    note("testCreatedSliverBecomesReady -> sliver ready: " + statusDetails.getGlobalStatus() + " @ " + auth.getUrn());
                    toWait.remove(auth);
                    toDescribe.add(auth);
                    if (toWait.isEmpty()) {
                        note("testCreatedSliverBecomesReady -> all slivers ready");
                        return;
                    }
                    continue;
                }

                note("testCreatedSliverBecomesReady -> sliver not ready: " + statusDetails.getGlobalStatus() + "" + " @ " + auth.getUrn());
                LOG.info("      Sliver not ready: \"" + statusDetails.getGlobalStatus() + "\". Deadline in: " + (deadline - now) + " ms");

            }

            now = System.currentTimeMillis();
            note("testCreatedSliverBecomesReady -> At least one sliver still not ready.  Trying again in 30 seconds...");
            LOG.info("   At least one sliver still not ready. Sleeping for 30 seconds. Deadline in: " + (deadline - now) + " ms");

            try {
                Thread.sleep(30000 /*ms*/);
            } catch (InterruptedException e) { /* Ignore*/ }

            now = System.currentTimeMillis();
        }
        errorFatal("Some slivers did not become ready within 20 minutes: " + toWait);
    }

    @Test(hardDepends = {"waitForAllReady"}, groups = {"createsliver", "nodelogin"})
    public void describeAllResources() throws JFedException {
        for (SfaAuthority auth : toDescribe) {
            try {
                AggregateManagerWrapper amWrapper = getAMWrapper(auth);

                String manifestRspec = null;
                AbstractGeniAggregateManager.AggregateManagerReply reply;
                try {
                    manifestRspec = amWrapper.describe(slice.urn, slice.credentials);
                    note("Called describe on " + auth.getUrnString());
                    reply = amWrapper.getLastReply();
                } catch (JFedException e) {
                    //ignore
                    note("amWrapper.describe threw JFedException: "+e.getMessage()+" for "+e.getXmlRpcResult());
                    reply = (AbstractGeniAggregateManager.AggregateManagerReply) e.getSfaReply();
                }

                if (reply.getGeniResponseCode().isSuccess()) {
                    assert manifestRspec != null;

                    NodeLoginTester nodeLoginTester = nodeLoginTesters.get(auth);
                    assert nodeLoginTester != null;
                    if (nodeLoginTester != null) {
                        boolean parseFoundLoginInfo = nodeLoginTester.parseSshInfoFromGeni3ManifestRspec(manifestRspec, auth, null);
                        if (!parseFoundLoginInfo)
                            warn("Found no SSH login info in manifest received at "+auth);
                    }
                }
            } catch (Throwable e) {
                errorNonFatal("Error while calling describe on "+ auth.getUrnString()+". Will continue with others.");
            }
        }
    }

    @Test(hardDepends = {"waitForAllReady"}, softDepends = {"describeAllResources"}, groups = {"nodelogin"})
    public void loginAndPing() throws JFedException, IOException {
        List<SfaAuthority> sliverAuths = director.getActivelyInvolvedAuthorities();

        boolean allLoginOk = true;

        //only try login on authorities with nodes.
        List<SfaAuthority> authsWithNode = new ArrayList<SfaAuthority>();
        for (SfaAuthority auth : sliverAuths)
            if (isNodeAuth(auth))
                authsWithNode.add(auth);

        assertTrue(authsWithNode.size() > 1, "Too few authorities with nodes (" + authsWithNode.size() + "): authsWithNode=" + authsWithNode + "   stitchedAuthorityUrns=" + stitchedAuthorityUrns);

        for (SfaAuthority auth : authsWithNode) {
            NodeLoginTester nodeLoginTester = nodeLoginTesters.get(auth);
            if (nodeLoginTester == null) {
                errorNonFatal("No SSH login info was extracted for " + auth.getUrnString());
                allLoginOk = false;
                continue;
            }
            LOG.info("Login to " + auth.getUrnString() + "  " + nodeLoginTester);
            note("Login to " + auth.getUrnString() + "  " + nodeLoginTester);

            if (nodeLoginTester.getSshHostname() == null) {
                errorNonFatal("No node / service / login in manifest RSpec, so SSH login cannot be tested for " + auth.getUrnString());
                allLoginOk = false;
                continue;
            }

            //Test node login using SSH private key
            Connection conn = new Connection(null, nodeLoginTester.getSshHostname(), nodeLoginTester.getSshPort());
            try {
                conn.connect();
            } catch (IOException e) {
                errorNonFatal("Could not connect to node \"" + nodeLoginTester.getSshHostname() + "\" at port " + nodeLoginTester.getSshPort());
                allLoginOk = false;
                continue;
            }
            String privateKeyToPrint = new String(nodeLoginTester.getSshKeyHelper().getPEMPrivateKey());
            if (privateKeyToPrint.length() > 50) privateKeyToPrint = privateKeyToPrint.substring(0, 50);
            note("Trying to log in with PEM private key:\n" + privateKeyToPrint);
            boolean isAuthenticated =
                    conn.authenticateWithPublicKey(nodeLoginTester.getSshUsername(), nodeLoginTester.getSshKeyHelper().getPEMRsaPrivateKey(), "nopass");
            setErrorsNotFatal();
            assertTrue(isAuthenticated, "Could not login to host. " + nodeLoginTester.getSshUsername() + "@" + nodeLoginTester.getSshHostname() + ":" + nodeLoginTester.getSshPort());
            setErrorsFatal();

            if (!isAuthenticated)
                continue;

            Session session = conn.openSession();
            BufferedReader sout = new BufferedReader(new InputStreamReader(session.getStdout()));
            BufferedReader serr = new BufferedReader(new InputStreamReader(session.getStderr()));

            List<String> pingTargets = pingTargetsForAuth(auth);
            for (String targetIP : pingTargets) {
                note("Trying to ping " + targetIP);
                /*
                Ping options on linux:
                       -c count
                              Stop after sending count ECHO_REQUEST packets. With deadline option, ping waits for count ECHO_REPLY packets, until the timeout expires.
                       -n     Numeric output only.  No attempt will be made to lookup symbolic names for host addresses.
                       -w deadline
                              Specify a timeout, in seconds, before ping exits regardless of how many packets have been sent or received. In this case ping does not stop after count packet are sent, it
                              waits either for deadline expire or until count probes are answered or for some error notification from network.
                       -W timeout
                              Time to wait for a response, in seconds. The option affects only timeout in absense of any responses, otherwise ping waits for two RTTs.
                * */

                String command = "ping -c 5 -n -w 30 " + targetIP;
                session.execCommand(command);
                String result = "";
                String err = "";
                String line = sout.readLine();
                while (line != null) {
                    result += line + "\n";
                    line = sout.readLine();
                }
                line = serr.readLine();
                while (line != null) {
                    err += line + "\n";
                    line = serr.readLine();
                }
                note("\"" + "ping -c 5 -n -w 30 " + targetIP + "\" command on " + nodeLoginTester.getSshUsername() + "@" + nodeLoginTester.getSshHostname() + ":" + nodeLoginTester.getSshPort() + " result: \"" + result.trim() + "\". (stderr is: \"" + err + "\")");
                setErrorsNotFatal();
                assertTrue(result.length() > 5, "I executed \"ping\" on the remote host, and expected some reply. Instead I got: \"" + result.trim() + "\". (stderr is: \"" + err + "\")");

                if (result.length() > 5) {
                    //example reply:
                    //   5 packets transmitted, 3 received
                    Pattern patternTrans = Pattern.compile("([0-9]*) packets transmitted");
                    Matcher matcherTrans = patternTrans.matcher(result);
                    assertTrue(matcherTrans.find(), "Did not find \"[0-9]* packets transmitted\" in result");
                    String sent = matcherTrans.group(1);

                    Pattern patternRecv = Pattern.compile("([0-9]*) received");
                    Matcher matcherRecv = patternRecv.matcher(result);
                    assertTrue(matcherRecv.find(), "Did not find \"[0-9]* received\" in result");
                    String recv = matcherRecv.group(1);

                    assertEquals("5", sent, "Packets sent count is not 5: sent=\"" + sent + "\"  (recv=\"" + recv + "\")");
                    assertEquals("5", recv, "Packets recveived count is not 5: \"" + recv + "\"");
                }
                setErrorsFatal();
            }

            session.close();

            conn.close();
        }

        assertTrue(allLoginOk, "one or more login tests failed");
    }

    //DEBUGREUSEPREVTEST comment  callDelete() to reuse previous test. uncomment for normal test
    @Test(softDepends = {"loginAndPing", "doStitchActions", "waitForAllReady", "describeAllResources"}, hardDepends = {
            "createSlice"}, groups = {"createsliver", "nodelogin"})
    public void callDeletes() throws JFedException {
        if (disableCleanup) skip("No cleanup needed.");

        List<SfaAuthority> allAuths = director.getActivelyInvolvedAuthorities();

        if (allAuths.size() == 0) {
            warn("Something went wrong: got empty list of involved authorities! This might be correct, but it could be a bug. Will instead Delete on all authorities to be sure everything is cleaned up.");
            allAuths = director.getAllAuthorities();
        }

        for (SfaAuthority authorityInfo : allAuths) {
            //do delete call
            try {
                AggregateManagerWrapper amWrapper = getAMWrapper(authorityInfo);
                amWrapper.deleteSliver(slice.urn, slice.credentials);
                note("Called delete on " + authorityInfo.getUrnString());
            } catch (JFedException e) {
                warn("Error while calling Delete on "+ authorityInfo.getUrnString()+". Will continue deleting.");
            }
        }
    }

    private class Node {
        private final String name;
        private final String authUrn;
        private List<Iface> ifaces = new ArrayList<Iface>();

        private Node(String authUrn) {
            this.name = "PC" + nodes.size();
            this.authUrn = authUrn;
            nodes.add(this);
        }

        private Node(String name, String authUrn) {
            this.name = name;
            this.authUrn = authUrn;
            nodes.add(this);
        }
    }

    private class Iface {
        private final Node node;
        private final Link link;
        private String ip;

        private Iface(Node node, Link link) {
            this.node = node;
            this.link = link;
            this.node.ifaces.add(this);
            this.link.ifaces.add(this);
            this.ip = link.subnet + "." + (link.ifaces.indexOf(this) + 1);
        }

        private String getName() {
            return node.name + ":if" + node.ifaces.indexOf(this);
        }
    }

    private class Link {
        private final String name;
        private final String subnet;
        private List<Iface> ifaces = new ArrayList<Iface>();

        private Link() {
            this.name = "link" + links.size();
            this.subnet = "192.168." + (4 + links.size());
            links.add(this);
        }

        private Link(Node a, Node b) {
            this();
            new Iface(a, this);
            new Iface(b, this);
        }

        private List<String> componentManagerUrns() {
            List<String> res = new ArrayList<String>();
            for (Iface i : ifaces) {
                Node n = i.node;
                res.add(n.authUrn);
            }
            return res;
        }

        private List<Iface> otherIfaces(Iface iface) {
            List<Iface> res = new ArrayList<Iface>();
            for (Iface i : ifaces)
                if (i != iface)
                    res.add(i);
            return res;
        }
    }
}
