package be.iminds.ilabt.jfed.lowlevel.api.test;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.AnyCredential;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.ServerType;
import be.iminds.ilabt.jfed.lowlevel.api.AbstractFederationApi;
import be.iminds.ilabt.jfed.lowlevel.api.AbstractFederationApi2;
import be.iminds.ilabt.jfed.lowlevel.api.FederationMemberAuthorityApi2;
import be.iminds.ilabt.jfed.lowlevel.api.ProtogeniSliceAuthority;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.SfaConnection;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.SSHKeyHelper;
import be.iminds.ilabt.jfed.util.TargetAuthority;

import javax.inject.Inject;
import java.security.NoSuchAlgorithmException;
import java.util.*;

/**
 * TestFederationMemberAuthority2
 */
public class TestFederationMemberAuthority2 extends AbstractFederationApi2Test {
    @Inject
    public TestFederationMemberAuthority2(Logger logger, TargetAuthority testedAuthority, GeniUser user, JFedConnectionProvider connectionProvider, AuthorityListModel authorityListModel, JFedPreferences jFedPreferences) {
        super(logger, testedAuthority, user, connectionProvider, authorityListModel, jFedPreferences);
    }

    public String getTestDescription() {
        return "Test Federation API Member Authority version 2";
    }

    private FederationMemberAuthorityApi2 ma;


    private GeniUrn testUserUrn;

    public SfaConnection getConnection() throws JFedException {
        return (SfaConnection) connectionProvider.getConnectionByAuthority(user, testedAuthority.getSfaAuthorityToConnect(),
                new ServerType(ServerType.GeniServerRole.GENI_CH_MA, 2));
    }

    List<String> memberDefaultFieldNames = new ArrayList<String>();
    List<AbstractFederationApi2.GetVersionResult.FieldInfo> memberDefaultFields =
            new ArrayList<AbstractFederationApi2.GetVersionResult.FieldInfo>();
    List<String> keyDefaultFieldNames = new ArrayList<String>();
    List<AbstractFederationApi2.GetVersionResult.FieldInfo> keyDefaultFields =
            new ArrayList<AbstractFederationApi2.GetVersionResult.FieldInfo>();

    public void setUp() {

        ma = new FederationMemberAuthorityApi2(logger, jFedPreferences);
        ma.setHandleMalformedReplies(false); //throw exceptions instead of trying to recover from malformed replies

        testUserUrn = GeniUrn.parse(user.getUserUrnString());
        assertNotNull(testUserUrn, "Error in test user urn: " + user.getUserUrnString());

        memberDefaultFields = ma.getMinimumFields("MEMBER");
        keyDefaultFields = ma.getMinimumFields("KEY");
        for (AbstractFederationApi2.GetVersionResult.FieldInfo fi : memberDefaultFields)
            memberDefaultFieldNames.add(fi.getName());
        for (AbstractFederationApi2.GetVersionResult.FieldInfo fi : keyDefaultFields)
            keyDefaultFieldNames.add(fi.getName());
    }

    private boolean hasKeyService = false;

    @Test(groups = {"getversion"})
    public void getVersion() throws JFedException {
        AbstractFederationApi2.FederationApiReply<AbstractFederationApi2.GetVersionResult> reply =
                ma.getVersion(getConnection());
        checkGetVersion(reply, ma, true);

        Hashtable<String, Object> htRes = (Hashtable<String, Object>) reply.getRawValue();
        Vector services = assertHashTableContainsVector(htRes, "SERVICES");

        if (services != null) {
            hasKeyService = services.contains("KEY");
        }
    }

    AnyCredential maTestUserCredential;
    List<AnyCredential> maTestUserCredentialList;

    @Test(softDepends = {"getVersion"}, groups = {"get_credential"})
    public void getTestUserCredential() throws JFedException {
        AbstractFederationApi2.FederationApiReply<List<AnyCredential>> reply =
                ma.getCredentials(getConnection(), new ArrayList<AnyCredential>(), testUserUrn, null);
        checkCorrectnessXmlRpcResult(reply);
        assertTrue(reply.getGeniResponseCode().isSuccess(),
                "The tested call did not return code 0 (success), instead it returned " + reply.getGeniResponseCode());

        assertInstanceOf(reply.getRawValue(), Vector.class,
                "The get_credential call should return an array (Vector) of dictionaries, instead it returned a " +
                        reply.getRawValue().getClass().getName()
        );
        Vector rawVal = (Vector) reply.getRawValue();
        for (Object credO : rawVal)
            assertInstanceOf(credO, Hashtable.class,
                    "The get_credential call should return an array containing dictionaries, " +
                            "instead one of the values in the array is a " + credO.getClass().getName()
            );

        List<AnyCredential> testUserCreds = reply.getValue();
        assertNotNull(testUserCreds, "no credential returned");

        assertEquals(testUserCreds.size(), rawVal.size(), "Error in jFed: raw credential list length (" +
                testUserCreds.size() + ") differs from converted credential list length (" + rawVal.size() + ")");

        assertNotEmpty(testUserCreds,
                "empty list of credentials returned. Expecting a list with credentials, but got  " +
                        (reply.getRawValue() == null ?
                                "null" :
                                (reply.getRawValue() instanceof Vector ?
                                        ("a list: " + reply.getRawValue()) :
                                        ("something of class:" + reply.getRawValue().getClass().getName())
                                )
                        )
        );
        maTestUserCredentialList = new ArrayList<AnyCredential>(testUserCreds);
        maTestUserCredential = testUserCreds.get(0);
    }

    AnyCredential testUserCredential;
    List<AnyCredential> testUserCredentialList;

    @Test(softDepends = {"getTestUserCredential"}, groups = {
            "get_credential"}, description = "Get A Credential for the test user. If the MA get_credential call did not work, this will try to fall back to the old PROTOGENI_SA API.")
    public void retrieveCredentialSomehow() throws JFedException {
        if (maTestUserCredentialList != null && maTestUserCredential != null) {
            testUserCredential = maTestUserCredential;
            testUserCredentialList = maTestUserCredentialList;
        } else {
            if (user.getUserAuthority().getUrl(new ServerType(ServerType.GeniServerRole.PROTOGENI_SA, 1)) == null) {
                errorFatal("The Uniform Federation API Member Authority get_credential call failed to provide a credential. additionally, the test user authority does not have a PROTOGENI_SA available as fallback. => Cannot retrieve a credential for the test user.");
            } else {
                note("The Uniform Federation API Member Authority get_credential call failed to provide a credential: Falling back to the PROTOGENI_SA API to retrieve a credential for the remaining tests.");
                SfaConnection protogeniSaCon =
                        (SfaConnection) connectionProvider.getConnectionByAuthority(user,
                                user.getUserAuthority(), new ServerType(ServerType.GeniServerRole.PROTOGENI_SA, 1));
                ProtogeniSliceAuthority protogeniSa = new ProtogeniSliceAuthority(logger, jFedPreferences);

                ProtogeniSliceAuthority.SliceAuthorityReply<AnyCredential> pgSaReply =
                        protogeniSa.getCredential(protogeniSaCon);
                assertTrue(pgSaReply.getGeniResponseCode().isSuccess());
                AnyCredential pgCred = pgSaReply.getValue();

                assertNotNull(pgCred, "no credential returned by protogeni SA");
                testUserCredentialList = new ArrayList<AnyCredential>();
                testUserCredentialList.add(pgCred);
                testUserCredential = pgCred;
            }
        }
    }

    private void checkMemberLookupResult(AbstractFederationApi2.FederationApiReply<AbstractFederationApi2.LookupResult> reply, List<String> expectedFieldNames, List<String> excludedFieldNames) {
        setErrorsFatal();

        checkCorrectnessXmlRpcResult(reply);
        checkLookupCorrectness(reply);
        Hashtable<String, Hashtable<String, Object>> raw =
                (Hashtable<String, Hashtable<String, Object>>) reply.getRawValue();

        AbstractFederationApi2.LookupResult lookupRes = reply.getValue();
        if (lookupRes == null)
            warn("lookup result is null. This is either due to receiving an incorrect result (will fail later in this test in that case), or due a a bug in the API client implementation");
//        assertNotNull(lookupRes, "lookup result is null");
//        assertNotNull(testUserUrn, "bug in test (testUserUrn is null)");
//
//        assertTrue(lookupRes.size() > 0, "Lookup result is empty for own urn: "+testUserUrn.getValue());

//        Hashtable<String, Hashtable<String, String>> raw = lookupRes.getRaw();
        errorNonFatalIfNot(raw.size() == 1, "lookup must return one user, but it returned " + raw.size());

        Hashtable<String, Object> rawUser = raw.get(testUserUrn.getValue());
        errorNonFatalIfNot(rawUser != null, "lookup must return requested test user (" + testUserUrn.getValue() + ") user.");

        String rawUrn = (String) rawUser.get("MEMBER_URN"); //check that if this users has a member urn, it matches.
        if (rawUrn != null && !testUserUrn.getValue().equals(rawUrn))
            errorNonFatal("lookup must return requested test user (" + testUserUrn.getValue() + ") user, and it did, however its returned MEMBER_URN entry is not matching: " + rawUrn);

//        for (int i = 0; i < raw.size(); i++) {
//            Hashtable<String, String> memberInfo = raw.get(i);
        for (Map.Entry<String, Hashtable<String, Object>> e : raw.entrySet()) {
            Hashtable<String, Object> memberInfo = e.getValue();
            for (String fieldName : expectedFieldNames) {
                if (!memberInfo.containsKey(fieldName))
                    errorNonFatal("returned result does not contain field: \"" + fieldName + "\"");
            }
            for (String fieldName : excludedFieldNames) {
                if (memberInfo.containsKey(fieldName))
                    errorNonFatal("returned result contains unexpected field: \"" + fieldName + "\"");
            }
        }
    }


    @Test(hardDepends = {"retrieveCredentialSomehow"}, groups = {"lookup_member"})
    public void lookupMemberInfoNoFilter() throws JFedException {
        Vector<String> filter = null;
        lookupMemberInfoXFilterHelper(filter, false);
    }

    @Test(hardDepends = {"retrieveCredentialSomehow"}, groups = {"lookup_member"})
    public void lookupMemberInfoEmptyFilter() throws JFedException {
        Vector<String> filter = new Vector<String>();
        note("this test calls a lookup with an EMPTY filter specified. So there is a filter, is just is empty. " +
                "This test is expecting NO fields to be returned.");
        lookupMemberInfoXFilterHelper(filter, false);
    }

    @Test(hardDepends = {"retrieveCredentialSomehow"}, groups = {"lookup_member"})
    public void lookupMemberInfoNoFilterWithMatchList() throws JFedException {
        Vector<String> filter = null;
        lookupMemberInfoXFilterHelper(filter, true);
    }

    @Test(hardDepends = {"retrieveCredentialSomehow"}, groups = {"lookup_member"})
    public void lookupMemberInfoEmptyFilterWithMatchList() throws JFedException {
        Vector<String> filter = new Vector<String>();
        note("this test calls a lookup with an EMPTY filter specified. So there is a filter, is just is empty. " +
                "This test is expecting NO fields to be returned.");
        lookupMemberInfoXFilterHelper(filter, true);
    }

    public void lookupMemberInfoXFilterHelper(Vector<String> filter, boolean makeMatchList /*if true, make match a list instead of a single item*/) throws JFedException {
        Hashtable<String, Object> match = new Hashtable<String, Object>();

        if (makeMatchList) {
            Vector v = new Vector();
            v.add(testUserUrn.getValue());
            match.put("MEMBER_URN", v);
        } else
            match.put("MEMBER_URN", testUserUrn.getValue());

        note("looking up member by MEMBER_URN: " + testUserUrn.getValue() + " with filter: " + filter);
        AbstractFederationApi2.FederationApiReply<AbstractFederationApi2.LookupResult> reply =
                ma.lookupMember(getConnection(), testUserCredentialList, match, filter, null);
        checkCorrectnessXmlRpcResult(reply);

        assertTrue(reply.getGeniResponseCode().isSuccess(), "The tested call did not return code 0 (success), instead it returned " + reply.getGeniResponseCode());

        List<String> expectedFieldNames = new ArrayList<String>();
        List<String> excludedFieldNames = new ArrayList<String>();
        for (AbstractFederationApi2.GetVersionResult.FieldInfo fi : memberDefaultFields)
            if (/*fi.getProtect().equals(AbstractFederationApi2.GetVersionResult.FieldInfo.Protect.PUBLIC) &&*/ (filter == null || filter.contains(fi.getName())))
                expectedFieldNames.add(fi.getName());
            else
                excludedFieldNames.add(fi.getName());
        if (filter != null && filter.isEmpty())
            note("Empty filter provided, so expecting no fields to be returned: excludedFieldNames=" + excludedFieldNames);
        if (filter == null)
            note("No filter provided, so expecting all fields to be returned: " + expectedFieldNames);
        if (filter != null && !filter.isEmpty())
            note("filter provided, so expecting only these fields to be returned: " + filter + " while none of these fields should be returned: " + excludedFieldNames);
        checkMemberLookupResult(reply, expectedFieldNames, excludedFieldNames);
    }


    @Test(hardDepends = {"retrieveCredentialSomehow"}, softDepends = {
            "lookupMemberInfoEmptyFilter"}, description = "", groups = {"lookup_member"})
    public void lookupPublicMemberInfoFiltered() throws JFedException {
        Hashtable<String, String> match = new Hashtable<String, String>();
        Vector<String> filter = new Vector<String>();
        filter.add("MEMBER_UID");

        lookupMemberInfoXFilterHelper(filter, false);
    }

    @Test(hardDepends = {"retrieveCredentialSomehow"}, softDepends = {"lookupMemberInfoEmptyFilter"}, groups = {
            "lookup_member"})
    public void lookupIdentifyingMemberInfoFiltered() throws JFedException {
        Hashtable<String, String> match = new Hashtable<String, String>();
        Vector<String> filter = new Vector<String>();
        filter.add("MEMBER_EMAIL");
        filter.add("MEMBER_FIRSTNAME");

        lookupMemberInfoXFilterHelper(filter, false);
    }

    @Test(hardDepends = {"retrieveCredentialSomehow"}, softDepends = {
            "lookupMemberInfoEmptyFilter"}, description = "", groups = {"lookup_member"})
    public void lookupPublicMemberInfoFilteredWithMatchList() throws JFedException {
        Hashtable<String, String> match = new Hashtable<String, String>();
        Vector<String> filter = new Vector<String>();
        filter.add("MEMBER_UID");

        lookupMemberInfoXFilterHelper(filter, true);
    }

    @Test(hardDepends = {"retrieveCredentialSomehow"}, softDepends = {"lookupMemberInfoEmptyFilter"}, groups = {
            "lookup_member"})
    public void lookupIdentifyingMemberInfoFilteredWithMatchList() throws JFedException {
        Hashtable<String, String> match = new Hashtable<String, String>();
        Vector<String> filter = new Vector<String>();
        filter.add("MEMBER_EMAIL");
        filter.add("MEMBER_FIRSTNAME");

        lookupMemberInfoXFilterHelper(filter, true);
    }


    /////////////////////////////////////////////////////// KEYS ///////////////////////////////////////////////////////

    private SSHKeyHelper sshKeyHelper;
    private Hashtable<String, Object> keyCreationDetails;
    private static String initialKeyDescription = "Randomly generated public key used for automated testing";
    private String keyId;

    @Test(hardDepends = {"retrieveCredentialSomehow"}, softDepends = {
            //softDeps just for order (KEY methods after MEMBER methods):
            "lookupMemberInfoNoFilter",
            "lookupMemberInfoEmptyFilter",
            "lookupPublicMemberInfoFiltered",
            "lookupIdentifyingMemberInfoFiltered"}, groups = {"key"})
    public void createKey() throws JFedException, NoSuchAlgorithmException {
        if (!hasKeyService) skip("No KEY service advertised in get_version");

        sshKeyHelper = new SSHKeyHelper();

        Hashtable<String, String> fields = new Hashtable<String, String>();
        fields.put("KEY_MEMBER", user.getUserUrnString());
        fields.put("KEY_PUBLIC", sshKeyHelper.getSshPublicKeyString());
        fields.put("KEY_DESCRIPTION", initialKeyDescription);

        AbstractFederationApi2.FederationApiReply<Hashtable<String, Object>> reply =
                ma.createKey(getConnection(), testUserCredentialList, testUserUrn, fields, null);
        assertTrue(reply.getGeniResponseCode().isSuccess(), "The tested call did not return code 0 (success), instead it returned " + reply.getGeniResponseCode());

        Object rawValue = reply.getXmlRpcCallDetailsWithCodeValueError().getResultValueObject();
        assertNotNull(rawValue, "The call did not return a value object");
        assertInstanceOf(rawValue, Hashtable.class, "The call did not return a dictionary");
        Hashtable rawValueHt = (Hashtable) rawValue;
        for (Object key : rawValueHt.keySet()) {
            assertInstanceOf(key, String.class, "One of the dictionary keys in the result value is not a String: class=" + key.getClass().getName() + " value=" + key);
            Object v = rawValueHt.get(key);
            if (key.equals("KEY_ID") && !(v instanceof String))
                errorNonFatal("The returned KEY_ID should be a String, but it is a " + v.getClass().getName());
            if (key.equals("KEY_DESCRIPTION") && !(v instanceof String))
                errorNonFatal("The returned KEY_DESCRIPTION should be a String, but it is a " + v.getClass().getName());
            if (!(v instanceof String))
                note("One of the dictionary values in the result value is not a String: key=" + key + " class=" + v.getClass().getName() + " value=" + v);
        }

        keyCreationDetails = reply.getValue();
        assertNotNull(keyCreationDetails, "The jFed code did not return details about the created key. (reply.getValue() == null). internal jFed error?");
        note("key details contains the following fields: " + keyCreationDetails.keySet());

        keyId = assertHashTableContainsNonemptyString(keyCreationDetails, "KEY_ID");
    }


    private void checkKeyLookupResult(AbstractFederationApi2.FederationApiReply<AbstractFederationApi.LookupResult> reply, List<String> expectedFieldNames, List<String> excludedFieldNames) {
        AbstractFederationApi.LookupResult lookupRes = reply.getValue();
        checkCorrectnessXmlRpcResult(reply);
//        checkLookupCorrectness(reply);
        assertNotNull(lookupRes, "lookup result is null");

        assertTrue(lookupRes.size() > 0, "Lookup result is empty for keyId: " + keyId);

        errorNonFatalIfNot(lookupRes.size() == 1, "lookup must return one member with one key, but it returned " + lookupRes.size() + " members");

        Map<String, Object> key = lookupRes.get(0);
        errorNonFatalIfNot(key != null, "lookup must return requested key (" + keyId + ") but 1 member with 1 key that was null.");

        String rawId = (String) key.get("KEY_ID"); //check that if this key has a keyId, it matches.
        if (rawId != null && !keyId.equals(rawId))
            errorNonFatal("lookup must return requested key (" + keyId + "), and it did, however its returned KEY_ID entry is not matching: " + rawId);

        for (String fieldName : expectedFieldNames) {
            if (!key.containsKey(fieldName))
                errorNonFatal("returned result does not contain field: \"" + fieldName + "\"");
        }
        for (String fieldName : excludedFieldNames) {
            if (key.containsKey(fieldName))
                errorNonFatal("returned result contains unexpected field: \"" + fieldName + "\"");
        }
    }

    @Test(hardDepends = {"createKey"}, groups = {"key"})
    public void lookupKeysNoFilter() throws JFedException {
        if (!hasKeyService) skip("No KEY service advertised in get_version");

        Hashtable<String, String> match = new Hashtable<String, String>();
        Vector<String> filter = null;//new Vector<String>();

        match.put("KEY_ID", keyId);

        note("looking up key by KEY_ID: " + keyId);
        AbstractFederationApi2.FederationApiReply<AbstractFederationApi.LookupResult> reply =
                ma.lookupKey(getConnection(), testUserCredentialList, null/*keyId*/, match, filter, null);
        checkCorrectnessXmlRpcResult(reply);

        assertTrue(reply.getGeniResponseCode().isSuccess(), "The tested call did not return code 0 (success), instead it returned " + reply.getGeniResponseCode());

        List<String> expectedFieldNames = new ArrayList<String>();
        List<String> excludedFieldNames = new ArrayList<String>();
        for (AbstractFederationApi2.GetVersionResult.FieldInfo fi : keyDefaultFields)
            expectedFieldNames.add(fi.getName());
        note("No filter provided, so expecting all fields to be returned: " + expectedFieldNames);
        checkKeyLookupResult(reply, expectedFieldNames, excludedFieldNames);

        String desc = (String) reply.getValue().get(0).get("KEY_DESCRIPTION");
        assertNotNull(desc);
        assertEquals(desc, initialKeyDescription);
    }

    private static String updatedKeyDescription =
            "Randomly generated public key used for automated testing with updated description";

    @Test(hardDepends = {"createKey"}, softDepends = {"lookupKeysNoFilter"}, groups = {"key"})
    public void updateKey() throws JFedException {
        if (!hasKeyService) skip("No KEY service advertised in get_version");

        Hashtable<String, String> fields = new Hashtable<String, String>();
        fields.put("KEY_DESCRIPTION", updatedKeyDescription);

        AbstractFederationApi2.FederationApiReply<String> reply =
                ma.updateKey(getConnection(), testUserCredentialList, keyId, fields, null);
        checkCorrectnessXmlRpcResult(reply);

        assertTrue(reply.getGeniResponseCode().isSuccess(), "The tested call did not return code 0 (success), instead it returned " + reply.getGeniResponseCode());
    }

    @Test(hardDepends = {"updateKey"}, groups = {"key"})
    public void lookupKeysNoFilterAfterUpdate() throws JFedException {
        if (!hasKeyService) skip("No KEY service advertised in get_version");

        Hashtable<String, String> match = new Hashtable<String, String>();
        Vector<String> filter = null;// new Vector<String>();

        match.put("KEY_ID", keyId);

        note("looking up key by KEY_ID: " + keyId);
        AbstractFederationApi2.FederationApiReply<AbstractFederationApi.LookupResult> reply =
                ma.lookupKey(getConnection(), testUserCredentialList, null/*keyId*/, match, filter, null);
        checkCorrectnessXmlRpcResult(reply);

        assertTrue(reply.getGeniResponseCode().isSuccess(), "The tested call did not return code 0 (success), instead it returned " + reply.getGeniResponseCode());

        List<String> expectedFieldNames = new ArrayList<String>();
        List<String> excludedFieldNames = new ArrayList<String>();
        for (AbstractFederationApi2.GetVersionResult.FieldInfo fi : keyDefaultFields)
            expectedFieldNames.add(fi.getName());
        note("No filter provided, so expecting all fields to be returned: " + expectedFieldNames);
        checkKeyLookupResult(reply, expectedFieldNames, excludedFieldNames);

        String desc = (String) reply.getValue().get(0).get("KEY_DESCRIPTION");
        assertNotNull(desc);
        assertEquals(desc, updatedKeyDescription);
    }

    @Test(hardDepends = {"retrieveCredentialSomehow"}, groups = {"key"}, softDepends = {"createKey",
                                                                                        "lookupKeysNoFilterAfterUpdate"}, description = "Try to delete key. Always called to ensure an attempt to delete is always made.")
    public void deleteKey() throws JFedException {
        if (!hasKeyService) skip("No KEY service advertised in get_version");
        if (keyId == null)
            skip("No key ID to delete known");

        AbstractFederationApi2.FederationApiReply<Boolean> reply =
                ma.deleteKey(getConnection(), testUserCredentialList, keyId, null);
        checkCorrectnessXmlRpcResult(reply);

        assertTrue(reply.getGeniResponseCode().isSuccess(), "delete_key unsuccessful");

        assertTrue(reply.getValue(), "delete_key returned false");
    }

    @Test(hardDepends = {"deleteKey"}, groups = {"key"})
    public void lookupKeysNoFilterAfterDelete() throws JFedException {
        if (!hasKeyService) skip("No KEY service advertised in get_version");

        Hashtable<String, String> match = new Hashtable<String, String>();
        Vector<String> filter = null;// new Vector<String>();

        match.put("KEY_ID", keyId);

        note("looking up key by KEY_ID: " + keyId);
        note("Expecting no key to be found, since we just deleted the key.");
        AbstractFederationApi2.FederationApiReply<AbstractFederationApi.LookupResult> reply =
                ma.lookupKey(getConnection(), testUserCredentialList, null/*keyId*/, match, filter, null);
        checkCorrectnessXmlRpcResult(reply);

        assertTrue(reply.getGeniResponseCode().isSuccess(), "The tested call did not return code 0 (success), instead it returned " + reply.getGeniResponseCode());

//        checkLookupCorrectness(reply);
        AbstractFederationApi.LookupResult lookupRes = reply.getValue();
        assertNotNull(lookupRes);
        assertTrue(lookupRes.size() == 0, "After delete, a key was still found: " + lookupRes);
    }
}
