package be.iminds.ilabt.jfed.ui.javafx.scanner;

import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.ServerType;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.scanner.AuthorityScanner;
import be.iminds.ilabt.jfed.scanner.AuthorityScannerInput;
import be.iminds.ilabt.jfed.scanner.AuthorityType;
import be.iminds.ilabt.jfed.scanner.ScanFeedback;
import be.iminds.ilabt.jfed.util.IOUtils;
import be.iminds.ilabt.jfed.util.JavaFXLogger;
import be.iminds.ilabt.jfed.util.KeyUtil;
import javafx.application.Platform;
import javafx.beans.property.SimpleStringProperty;
import javafx.beans.property.StringProperty;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.scene.control.*;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.HBox;
import javafx.scene.layout.Pane;
import javafx.stage.Stage;
import javafx.util.Pair;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.PrivateKey;
import java.security.cert.X509Certificate;
import java.util.*;

/**
 * FXML Controller class
 *
 * @author wim
 */
public class ScanAuthorityDialogController {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(ScanAuthorityDialogController.class);

    private final JavaFXLogger logger;
    private final GeniUserProvider geniUserProvider;
    private final JFedPreferences jFedPreferences;


    @FXML
    private BorderPane pane;

    @FXML
    private Pane inputPane;
    @FXML
    private TabPane tabs;

    @FXML
    private Tab securityTab;
    @FXML
    private Tab outputTab;
    @FXML
    private Tab debugTab;

    @FXML
    private TextField amUrlField;
    @FXML
    private TextField urnField;
    @FXML
    private ComboBox<AuthorityType> typeComboBox;

    @FXML
    private HBox postScanButtons;
    @FXML
    private HBox scanLogBox;
    @FXML
    private TextArea certTextArea;
    @FXML
    private TextArea scanLogTextArea;
    @FXML
    private ProgressIndicator progress;
    @FXML
    private Button scanButton;
    @FXML
    private Button cancelButton;

    @FXML
    private CheckBox trustCertsCheckbox;

    @FXML
    private Pane scannedFacts;

    private StringProperty logtext = new SimpleStringProperty();

    private Stage stage;

    @Inject
    public ScanAuthorityDialogController(JavaFXLogger logger, GeniUserProvider geniUserProvider, JFedPreferences jFedPreferences) {
        this.logger = logger;
        this.geniUserProvider = geniUserProvider;
        this.jFedPreferences = jFedPreferences;
    }

    /**
     * Initializes the controller class.
     */
    @FXML
    private void initialize() {
        inputPane.managedProperty().bind(inputPane.visibleProperty());
        tabs.managedProperty().bind(tabs.visibleProperty());
        tabs.visibleProperty().setValue(false);

        String validatorCss = getClass().getResource("validation.css").toExternalForm();
        pane.getStylesheets().add(validatorCss);
        trustCertsCheckbox.getStylesheets().add(validatorCss);

        amUrlField.textProperty().addListener((observableValue, s, s1) -> validate());

        ObservableList<AuthorityType> typeItems = FXCollections.observableArrayList();
        typeItems.addAll(AuthorityType.values());
        typeComboBox.setItems(typeItems);
        trustCertsCheckbox.setSelected(false);

        scanLogTextArea.textProperty().bind(logtext);
    }

    public void setStage(Stage stage) {
        this.stage = stage;
    }

    public void cancel() {
        typeComboBox.getSelectionModel().clearSelection();
        amUrlField.setText("");
        urnField.setText("");

        if (stage != null)
            stage.close();
    }

    Thread scanThread = null;

    public void scan() {
        logtext.setValue("");
        trustCertsCheckbox.setSelected(false);
        trustCertsCheckbox.getStyleClass().removeAll("validation-error", "validation-warning");

        if (scanThread != null) {
            return;
        }

        inputPane.visibleProperty().setValue(false);
        tabs.visibleProperty().setValue(true);
        tabs.getSelectionModel().select(debugTab);
        pane.getScene().getWindow().sizeToScene();

        amUrlField.setDisable(true);
        urnField.setDisable(true);
        typeComboBox.setDisable(true);
        scanButton.setDisable(true);
        cancelButton.setDisable(true);
        scanLogBox.setVisible(true);
        progress.setVisible(true);
        postScanButtons.setVisible(false);

        Runnable scanner = () -> {
            try {
                performScan();
            } catch (final Throwable e) {

                Platform.runLater(new Runnable() {
                    @Override
                    public void run() {
                        logtext.setValue(logtext.getValue() + "\nScanning encountered an error: " + e.getMessage() + "\n" + IOUtils.exceptionToStacktraceString(e) + "\n");
                        amUrlField.setDisable(false);
                        urnField.setDisable(false);
                        typeComboBox.setDisable(false);
                        scanButton.setDisable(false);
                        cancelButton.setDisable(false);
                        scanLogBox.setVisible(true);
                        progress.setVisible(false);
                        postScanButtons.setVisible(false);
                    }
                });
            }
        };
        Thread scanThread = new Thread(scanner);
        scanThread.start();
    }

    private AuthorityScanner scanner;

    private Map<AuthorityScanner.ScannedSingleFact, ScanFactPane> scanFactPaneByType;
    private Map<AuthorityScanner.ScannedSingleFact, ScanFactPane> scanFactPaneByHrn;
    private Map<AuthorityScanner.ScannedSingleFact, ScanFactPane> scanFactPaneByUrn;
    private Map<Pair<ServerType, AuthorityScanner.ScannedSingleFact>, ScanFactPane> scanFactPaneByServertypeAndUrl;

    /* Should be called in seperate Thread */
    private void performScan() {
        scanner = null;
        List<String> inputUrns = new ArrayList<>();
        List<String> inputUrls = new ArrayList<>();
        if (!amUrlField.getText().trim().isEmpty())
            inputUrls.add(amUrlField.getText());
        if (!urnField.getText().trim().isEmpty())
            inputUrns.add(urnField.getText());
        AuthorityType authorityType = null;
        if (!typeComboBox.getSelectionModel().isEmpty())
            authorityType = typeComboBox.getSelectionModel().getSelectedItem();
        List<X509Certificate> trustedCerts = new ArrayList<>();

        List<X509Certificate> userCerts = new ArrayList<>();
        PrivateKey userPrivateKey = null;
        if (geniUserProvider.isUserLoggedIn()) {
            userCerts.addAll(geniUserProvider.getLoggedInGeniUser().getClientCertificateChain());
            userPrivateKey = geniUserProvider.getLoggedInGeniUser().getPrivateKey();
        }

        AuthorityScannerInput input = new AuthorityScannerInput(
                inputUrns, inputUrls, trustedCerts, userCerts, userPrivateKey,
                authorityType, false/*userIsOfScannedAuthority*/);
        ScanFeedback feedback = new ScanFeedback() {
            @Override
            public boolean trustCert(List<X509Certificate> certificateList) {
                //TODO
                return true;
            }

            @Override
            public boolean trustAlias(String urlHost, String certSubject) {
                //TODO
                return true;
            }
        };
        scanner = new AuthorityScanner(logger, input, feedback, jFedPreferences);
        scanner.addResultListener(line -> Platform.runLater(() -> logtext.setValue(logtext.getValue() + line + "\n")));

        scanner.scan();

        //not needed: is added line by line now
//        scanLogTextArea.setText(scanner.getLogText());

        final StringBuilder certs = new StringBuilder();
        for (X509Certificate cert : scanner.getTrustedAuthCerts()) {
            certs.append(KeyUtil.x509certificateToPem(cert)).append('\n');
        }

        Platform.runLater(() -> {
            certTextArea.setText(certs.toString());
            scannedFacts.getChildren().clear();
        });

        scanFactPaneByType = new HashMap<>();
        scanFactPaneByUrn = new HashMap<>();
        scanFactPaneByHrn = new HashMap<>();
        scanFactPaneByServertypeAndUrl = new HashMap<>();

        final List<ScanFactPane> panes = new ArrayList<>();
        for (AuthorityScanner.ScannedSingleFact fact : scanner.getTypes().getFacts()) {
            ScanFactPane pane = new ScanFactPane("type", fact);
            if (fact == scanner.getTypes().getMostTrustedScannedSingleFact())
                pane.setSelected(true);
            else
                pane.setSelected(false);
            panes.add(pane);
            scanFactPaneByType.put(fact, pane);
        }
        for (AuthorityScanner.ScannedSingleFact fact : scanner.getUrns().getFacts()) {
            ScanFactPane pane = new ScanFactPane("urn", fact);
            if (fact == scanner.getUrns().getMostTrustedScannedSingleFact())
                pane.setSelected(true);
            else
                pane.setSelected(false);
            panes.add(pane);
            scanFactPaneByUrn.put(fact, pane);
        }
        for (AuthorityScanner.ScannedSingleFact fact : scanner.getHrns().getFacts()) {
            ScanFactPane pane = new ScanFactPane("hrn", fact);
            if (fact == scanner.getHrns().getMostTrustedScannedSingleFact())
                pane.setSelected(true);
            else
                pane.setSelected(false);
            panes.add(pane);
            scanFactPaneByHrn.put(fact, pane);
        }
        for (Map.Entry<ServerType, AuthorityScanner.ScannedInformation<String>> e : scanner.getUrlsByServerType().entrySet()) {
            ServerType st = e.getKey();
            AuthorityScanner.ScannedInformation<String> urls = e.getValue();
            for (AuthorityScanner.ScannedSingleFact fact : urls.getFacts()) {
                ScanFactPane pane = new ScanFactPane(st.getRole() + " v" + st.getVersion() + " URL", fact);

                //only select URLs if they are NOT a guess
                if (fact == urls.getMostTrustedScannedSingleFact()&& fact != null && fact.getCertainty().value > AuthorityScanner.ScanCertainty.GOOD_GUESS.value)
                    pane.setSelected(true);
                else
                    pane.setSelected(false);
                panes.add(pane);
                scanFactPaneByServertypeAndUrl.put(new Pair<>(st, fact), pane);
            }
        }
        Platform.runLater(() -> scannedFacts.getChildren().setAll(panes));


        //success, let user press OK
        Platform.runLater(() -> {
            logtext.setValue(logtext.getValue() + "\nScanning finished successfully. You can now review the scan results.\n");
            amUrlField.setDisable(false);
            urnField.setDisable(false);
            typeComboBox.setDisable(false);
            scanButton.setDisable(false);
            cancelButton.setDisable(false);
            scanLogBox.setVisible(true);
            progress.setVisible(false);
            postScanButtons.setVisible(true);
            tabs.getSelectionModel().select(outputTab);
        });

        scanThread = null;
    }

    SfaAuthority createAuthority = null;

    public void ok() {
        if (scanner == null)
            return;

        if (scanner != null && !scanner.getTrustedAuthCerts().isEmpty() && !trustCertsCheckbox.isSelected()) {
            trustCertsCheckbox.getStyleClass().add("validation-error");
            tabs.getSelectionModel().select(securityTab);
            trustCertsCheckbox.requestFocus();
            return;
        }
        trustCertsCheckbox.getStyleClass().removeAll("validation-warning", "validation-error");


        AuthorityScanner.ScannedInformation<AuthorityType> types = new AuthorityScanner.ScannedInformation<>();
        Map<ServerType, AuthorityScanner.ScannedInformation<String>> urlsByServerType = new HashMap<>();
        AuthorityScanner.ScannedInformation<String> hrns = new AuthorityScanner.ScannedInformation<>();
        AuthorityScanner.ScannedInformation<String> urns = new AuthorityScanner.ScannedInformation<>();
        Set<X509Certificate> trustedAuthCerts = scanner.getTrustedAuthCerts();
        Set<String> trustedAliases = scanner.getTrustedAliases();

        for (AuthorityScanner.ScannedSingleFact fact : scanner.getTypes().getFacts()) {
            ScanFactPane pane = scanFactPaneByType.get(fact);
            if (pane.isSelected()) types.add(fact);
        }
        for (AuthorityScanner.ScannedSingleFact fact : scanner.getUrns().getFacts()) {
            ScanFactPane pane = scanFactPaneByUrn.get(fact);
            if (pane.isSelected()) urns.add(fact);
        }
        for (AuthorityScanner.ScannedSingleFact fact : scanner.getHrns().getFacts()) {
            ScanFactPane pane = scanFactPaneByHrn.get(fact);
            if (pane.isSelected()) hrns.add(fact);
        }
        for (Map.Entry<ServerType, AuthorityScanner.ScannedInformation<String>> e : scanner.getUrlsByServerType().entrySet()) {
            ServerType st = e.getKey();
            AuthorityScanner.ScannedInformation<String> urls = e.getValue();
            for (AuthorityScanner.ScannedSingleFact fact : urls.getFacts()) {
                ScanFactPane pane = scanFactPaneByServertypeAndUrl.get(new Pair<>(st, fact));
                if (pane.isSelected()) {
                    if (!urlsByServerType.containsKey(st))
                        urlsByServerType.put(st, new AuthorityScanner.ScannedInformation<>());
                    urlsByServerType.get(st).add(fact);
                }
            }
        }

        try {
            createAuthority =
                    AuthorityScanner.createResult(types, urlsByServerType, hrns, urns, trustedAuthCerts, trustedAliases);
        } catch (JFedException e) {
            throw new RuntimeException("Error creating result: " + e.getMessage(), e);
        }

        if (stage != null)
            stage.close();
    }

    /**
     * Check if valid URL
     */
    public void validate() {
        amUrlField.getStyleClass().removeAll("validation-error", "validation-warning");
        try {
            URL test = new URL(amUrlField.getText());
        } catch (MalformedURLException e) {
            amUrlField.getStyleClass().add("validation-error");
        }
    }

    public String getAMUrl() {
        if (amUrlField.getText().equals("")) return null;
        return amUrlField.getText();
    }

    public SfaAuthority getCreatedSfaAuthority() {
        return createAuthority;
    }
}
