package be.iminds.ilabt.jfed.experimenter_gui.bugreporting;

import be.iminds.ilabt.jfed.call_log_output.CallReportReaderFactory;
import be.iminds.ilabt.jfed.call_log_output.SerializableApiCallDetails;
import be.iminds.ilabt.jfed.highlevel.stitcher.StitchingReport;
import be.iminds.ilabt.jfed.log.cache.ApiCallDetailsCache;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.ui.javafx.log_gui.LogHistoryPanel;
import be.iminds.ilabt.jfed.util.IOUtils;
import javafx.application.Platform;
import javafx.beans.property.BooleanProperty;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.collections.transformation.FilteredList;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.fxml.FXML;
import javafx.scene.control.*;
import javafx.scene.control.cell.TextFieldListCell;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.web.WebEngine;
import javafx.scene.web.WebView;
import javafx.stage.FileChooser;
import javafx.util.StringConverter;
import org.apache.commons.codec.binary.Base64;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.io.StringReader;
import java.net.URL;
import java.util.List;
import java.util.Optional;

/**
 * BugReportViewerPanel
 */
public class BugReportViewerPanel {
    private static final Logger LOG = LoggerFactory.getLogger(BugReportViewerPanel.class);
    public static final String BUG_REPORT_VIEWER_FXML = "BugReportViewerPanel.fxml";

    @FXML
    private TextField reportChooserTextField;

    @FXML
    private TextArea descriptionTextArea;
    @FXML
    private TextField versionTextField;
    @FXML
    private TextField environmentTextField;
    @FXML
    private TextField credentialsTextField;
    @FXML
    private TextField emailTextField;
    @FXML
    private Label includedCallsLabel;

    @FXML
    private TextArea loggedInUserTextArea;
    @FXML
    private TextArea preferencesTextArea;
    @FXML
    private TextArea slicesTextArea;

    @FXML
    private LogHistoryPanel logHistoryPanel;

    @FXML
    private CheckBox showThreadOverviewCheckBox;
    @FXML
    private LogLineTable debugLogs;
    private FilteredList<LogLine> filteredLogLines;
    @FXML
    private TextField searchDebugField;
    @FXML
    private ConnectivityTestTable connectivityTestTable;

    @FXML
    private ComboBox<StitchingReport> stitchingReportComboBox;
    @FXML
    private WebView stitchingReportWebView;

    @FXML
    private ImageView screenshotImageView;
    @FXML
    private Tab screenshotTab;


    private BooleanProperty showThreadOverview = new SimpleBooleanProperty(false);

    private final ApiCallDetailsCache cache;
    private final CallReportReaderFactory callReportReaderFactory;

    @Inject
    public BugReportViewerPanel(ApiCallDetailsCache cache, CallReportReaderFactory callReportReaderFactory) {
        this.cache = cache;
        this.callReportReaderFactory = callReportReaderFactory;
    }

    @FXML
    private void initialize() {
        LOG.debug("Bug Report Viewer initialize");

        showThreadOverviewCheckBox.selectedProperty().bindBidirectional(showThreadOverview);
        logHistoryPanel.setAutoShowLast(false);
        logHistoryPanel.setAuthShowLastVisible(false);

        if (startReport != null)
            try {
                fetchReport(startReport);
            } catch (IOException e) {
                LOG.error("Failed to fetch startReport \"" + startReport + "\"", e);
            }

        stitchingReportComboBox.setCellFactory(TextFieldListCell.forListView(new StringConverter<StitchingReport>() {
            @Override
            public String toString(StitchingReport r) {
                if (r == null) return "";
                if (r == null || r.getStates().isEmpty()) return "empty StitchingReport";
                return "StitchingReport @ " + r.getStates().get(0).getStateStart();
            }

            @Override
            public StitchingReport fromString(String s) {
                throw new RuntimeException("not needed");
            }
        }));

        stitchingReportComboBox.setOnAction(new EventHandler<ActionEvent>() {
            @Override
            public void handle(ActionEvent event) {
                StitchingReport r = stitchingReportComboBox.getSelectionModel().getSelectedItem();
                if (r == null) {
                    WebEngine webEngine = stitchingReportWebView.getEngine();
                    webEngine.loadContent("<html><body><p>No Report</p></body></html>");
                } else {
                    try {
                        File tmpFile = File.createTempFile("stitchingReport", ".html");
                        r.toHtml(tmpFile);
                        String reportAsHtml = IOUtils.fileToString(tmpFile);

                        WebEngine webEngine = stitchingReportWebView.getEngine();
                        webEngine.loadContent(reportAsHtml);
                    } catch (IOException e) {
                        LOG.error("IOException showing stitching report", e);
                    }
                }
            }
        });

        searchDebugField.textProperty().addListener(new ChangeListener<String>() {
            @Override
            public void changed(ObservableValue<? extends String> observable, String oldValue, String newValue) {
                if (!newValue.equals(oldValue))
                    debugLogs.getSelectionModel().clearSelection();
                search(newValue);
            }
        });
        searchDebugField.setOnAction(new EventHandler<ActionEvent>() {
            @Override
            public void handle(ActionEvent event) {
                searchNext();
            }
        });
    }

    private void search(String query) {
        if (query == null) return;

        int selectedLine = debugLogs.getSelectionModel().getSelectedIndex();
        if (selectedLine < 0)
            selectedLine = -1;

        int foundIndex = -1;
        Optional<LogLine> found = Optional.empty();
        for (int i = selectedLine + 1; i != selectedLine; ) {
            LogLine ll = filteredLogLines.get(i);
            if (
                    (ll.getMessage() != null && ll.getMessage().toLowerCase().contains(query.toLowerCase())) ||
                            (ll.getException() != null && ll.getException().toLowerCase().contains(query.toLowerCase()))
                    ) {
                foundIndex = i;
                found = Optional.of(ll);
                break;
            }

            i++;
            if (i >= filteredLogLines.size())
                i = 0;
        }

        final int foundIndexFinal = foundIndex;

        if (found.isPresent()) {
            Platform.runLater(new Runnable() {
                @Override
                public void run() {
                    debugLogs.getSelectionModel().select(foundIndexFinal);
                    debugLogs.getFocusModel().focus(foundIndexFinal);
                    debugLogs.scrollTo(foundIndexFinal);
                    LOG.debug("search(" + query + ") focusses " + foundIndexFinal);
                }
            });
        } else {
            Platform.runLater(new Runnable() {
                @Override
                public void run() {
                    debugLogs.getSelectionModel().selectFirst();
                    debugLogs.getFocusModel().focus(-1);
                    debugLogs.scrollTo(0);
                    LOG.debug("search(" + query + ") focusses -1");
                }
            });
        }
    }

    private void searchNext() {
//        LOG.debug("TOOD: searchNext()");
        search(searchDebugField.getText());
    }

    private String startReport;

    public void setStartReport(String urlString) throws IOException {
        this.startReport = urlString;

        if (reportChooserTextField != null)
            fetchReport(startReport);
    }

    public void fetchReport(String urlString) throws IOException {
        reportChooserTextField.setText(urlString);
        URL testUrl = new URL(urlString);
        String bugReportContent = IOUtils.streamToString(testUrl.openStream(), "UTF-8");

        showReport(bugReportContent);
    }

    @FXML
    public void fetchReport() throws IOException {
        URL testUrl = new URL(reportChooserTextField.getText());
        String bugReportContent = IOUtils.streamToString(testUrl.openStream(), "UTF-8");

        showReport(bugReportContent);
    }

    @FXML
    public void openReport() throws IOException {
        FileChooser fileChooser = new FileChooser();
        fileChooser.setTitle("Open BugReport file");
        File file = fileChooser.showOpenDialog(reportChooserTextField.getScene().getWindow());
        URL fileUrl = file.toURI().toURL();

        reportChooserTextField.setText(fileUrl + "");
        fetchReport();
    }

    @FXML
    public void showReport(String bugReportContent) throws IOException {

        BugReportReader bugReportReader =
                new BugReportReader(new StringReader(bugReportContent), cache, callReportReaderFactory);

        BugReport bugReport = null;
        try {
            bugReport = bugReportReader.call();
        } catch (Exception e) {
            LOG.error("Error while reading bugreport", e);
            JFDialogs.create().message("Could not parse bugreport").showException(e);
            return;
        }
        List<SerializableApiCallDetails> apiCallDetails = bugReport.getCalls();
        // for (SerializableApiCallDetails apiCallDetail : apiCallDetails)
        //apiCallDetail.reconstructApiCallReply(); //HACK, because reply is not in XML but can be partially recreated

        if (apiCallDetails != null) {
            ObservableList<SerializableApiCallDetails> observableApiCallDetails =
                    FXCollections.observableArrayList(apiCallDetails);
            logHistoryPanel.setApiCallHistory(observableApiCallDetails);
        }

        descriptionTextArea.setText(bugReport.getDescription());
        versionTextField.setText(bugReport.getVersion());
        environmentTextField.setText(bugReport.getEnvironment());
        credentialsTextField.setText(bugReport.getReporterCredential());
        emailTextField.setText(bugReport.getMail());
        includedCallsLabel.setText(bugReport.getCalls().size() + "");

        StringBuilder sliceText = new StringBuilder();
        for (BRSlice slice : bugReport.getSlices()) {
            sliceText.append("Slice \"" + slice.getSliceUrn() + "\"" + "\n");
            if (slice.getSlivers() != null)
                for (BRSliver sliver : slice.getSlivers()) {
                    sliceText.append("   Sliver \"" + sliver.getSliverUrn() + "\"" + "\n");
                    sliceText.append("       sliverStatus=" + sliver.getSliverStatus() + "" + "\n");
                    sliceText.append("       currentStatus=" + sliver.getCurrentStatus() + "" + "\n");
                    if (sliver.getComponents() != null && !sliver.getComponents().isEmpty())
                        sliceText.append("       component(s)=" + sliver.getComponents() + "" + "\n");
//                    if (sliver.getNodesConnectivity() != null)
//                        for (Map.Entry<String, ConnectivityTest.Status> e : sliver.getNodesConnectivity().entrySet())
//                            sliceText.append("       node \""+e.getKey()+"\" status="+e.getValue()+"\n");
                }
            sliceText.append("\n");
        }
        slicesTextArea.setText(sliceText.toString());
        loggedInUserTextArea.setText(bugReport.getLoggedInUserInfo());
        preferencesTextArea.setText(bugReport.getPreferences());

        List<LogLine> logLines = bugReport.getLogLines();
        filteredLogLines = new FilteredList<>(FXCollections.observableArrayList(logLines));
        if (logLines != null)
            debugLogs.setItems(filteredLogLines);

        filteredLogLines.setPredicate(logLine -> logLine.getMessage() != null && !logLine.getMessage().startsWith("Thread overview:"));

        showThreadOverview.addListener(new ChangeListener<Boolean>() {
            @Override
            public void changed(ObservableValue<? extends Boolean> observableValue, Boolean oldValue, Boolean newValue) {
                if (showThreadOverview.get())
                    filteredLogLines.setPredicate(logLine -> logLine.getMessage() != null && !logLine.getMessage().startsWith("Thread Overview"));
                else
                    filteredLogLines.setPredicate(logLine -> true);
            }
        });

        List<BRConnectivityTest> connectivityTests = bugReport.getConnectivityTests();
        if (connectivityTests != null)
            connectivityTestTable.setItems(FXCollections.observableArrayList(connectivityTests));

        if (bugReport.getScreenshot() != null) {
            byte[] decodedImage = Base64.decodeBase64(bugReport.getScreenshot());

            if (decodedImage != null) {
                ByteArrayInputStream imageInputStream = new ByteArrayInputStream(decodedImage);
                Image image = new Image(imageInputStream);

                screenshotImageView.setImage(image);
            } else {
                screenshotImageView.setImage(null);
            }
        } else {
            screenshotImageView.setImage(null);
        }

        screenshotTab.setDisable(screenshotImageView.getImage() == null);


        assert stitchingReportComboBox != null;
        assert bugReport != null;
        assert bugReport.getStitchingReports() != null;
        stitchingReportComboBox.setItems(FXCollections.observableArrayList(bugReport.getStitchingReports()));

    }
}
