package be.iminds.ilabt.jfed.ui.cli;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.AnyCredential;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.UserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AMv2Wrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AMv3Wrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticAggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticUserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.rspec.model.StringRspec;
import be.iminds.ilabt.jfed.ui.commandline.BaseCli;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.IOUtils;
import com.google.inject.Injector;
import org.apache.commons.cli.*;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;

/**
 * CreateSliverCli
 */
public class DeleteSliverCli {
    private final static org.slf4j.Logger LOG = LoggerFactory.getLogger(DeleteSliverCli.class);

    public static void help(Options options) {
        HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp(200, "jfed-create-sliver-cli [options ... ]", "Options:", options,
                "\n\nTemplate Context Properties File:\n  username = <username>\n" +
                        "  passwordFilename = <filename of file containing password>\n" +
                        "  pemKeyAndCertFilename = <filename of file containing user certificate and private key in PEM format>\n" +
                        "  userAuthorityUrn = <URN of test user's authority>\n"
        );
    }

    public static void main(String[] args) throws IOException, ClassNotFoundException {
        int retval = main_helper(args, System.out, System.err, System.in, new BaseCli(), true);
        System.exit(retval);
    }
    public static int main_helper(String[] args,
                                  final PrintStream out, final PrintStream err, final InputStream in,
                                  BaseCli baseCli,
                                  boolean temporarilyRedirectSystemStreams) throws IOException, ClassNotFoundException {
        final PrintStream orig_out = System.out;
        final PrintStream orig_err = System.err;
        final InputStream orig_in = System.in;
        if (temporarilyRedirectSystemStreams) {
            System.setOut(out);
            System.setErr(err);
            System.setIn(in);
        }

        int retval = -1;
        try {
            retval = main_internal(args, out, err, in, baseCli);
        } finally {
            if (temporarilyRedirectSystemStreams) {
                System.setOut(orig_out);
                System.setErr(orig_err);
                System.setIn(orig_in);
            }
        }
        return retval;
    }

    public static int main_internal(String[] args,
                                    final PrintStream out, final PrintStream err, final InputStream in,
                                    BaseCli baseCli) throws IOException, ClassNotFoundException {
        CommandLineParser parser = new BasicParser();

        // create the Options
        Options options = new Options();

        BaseCli.addUserConfig(options, "The properties file containing context details (only the login info in the context file is used by this tool)", false);

        BaseCli.addAuthoritiesConfig(options);

        //TODO make this "slice", and allow slice name as well (we can make slice URN using the user authority)
        options.addOption(OptionBuilder.withLongOpt("slice")
                .withDescription("The URN or name of the slice to use.")
                .hasArg()
                .withArgName("SLICE URN OR NAME")
                .isRequired()
                .create("s"));
        options.addOption(OptionBuilder.withLongOpt("rspec")
                .withDescription("The rspec file will be used to determine which AM's to contact (default: use info from SA, which may be incomplete)")
                .hasArg()
                .withArgName("RSPEC XML FILE")
                .create("r"));
        options.addOption(OptionBuilder.withLongOpt("am-api")
                .withDescription("The AM Api version to use (\"2\" and \"3\" supported, default: choose automatically)")
                .hasArg()
                .withArgName("AM API VERSION")
                .create("am"));
        options.addOption("q", "quiet", false, "less output");
        options.addOption("d", "debug", false, "extra debugging output");

        CommandLine line = null;
        try {
            // parse the command line arguments
            line = parser.parse(options, args);
        } catch (ParseException exp) {
            err.println("Command line argument Syntax error: " + exp.getMessage());
            help(options);
            return 3;
        }

        String sliceIdArg = line.getOptionValue("slice");
        String rspecArg = line.getOptionValue("rspec");
        String amApiVersionArg = line.getOptionValue("am-api");
        boolean debug = line.hasOption("debug");
        final boolean silent;

        Logger logger = new Logger();

        if (debug)
            silent = false;
        else
            silent = line.hasOption("quiet");

        if (amApiVersionArg != null && !amApiVersionArg.equals("2") && !amApiVersionArg.equals("3")) {
            err.println("Unsupported AM api version: \"" + amApiVersionArg + "\"");
            help(options);
            return 3;
        }

        baseCli.initAuthoritiesModule(line);
        try {
            baseCli.initUserModule(line, out, err, in);
        } catch (BaseCli.CliArgumentException e) {
            LOG.debug("Got CliArgumentException", e);
            //User has been informed already
            return 1;
        }
        Injector injector = baseCli.getInjector(err);
        GeniUser user = injector.getInstance(GeniUser.class);
        GeniUserProvider userProvider = injector.getInstance(GeniUserProvider.class);
        AuthorityListModel authorityListModel = injector.getInstance(AuthorityListModel.class);
        AuthorityFinder authorityFinder = injector.getInstance(AuthorityFinder.class);
        JFedPreferences jFedPreferences = injector.getInstance(JFedPreferences.class);


        List<String> sliceComponentManagers;
        if (rspecArg != null) {
            String rspecString = IOUtils.fileToString(rspecArg);
            StringRspec rspec = new StringRspec(rspecString);
            sliceComponentManagers = rspec.getAllComponentManagerUrns();
        } else
            sliceComponentManagers = new ArrayList<String>();

        if (!BaseCli.checkUser(user, line, out, err, in))
            return 3;


        if (!silent) {
            if (debug) {
                out.println("   User:" + user.getUserUrnString());
                out.println("      Authority URN:" + user.getUserAuthority().getUrnString());
            }
        }

        GeniUrn sliceUrn = null;
        if (sliceIdArg.startsWith("urn:"))
            try {
                sliceUrn = new GeniUrn(sliceIdArg);
            } catch (GeniUrn.GeniUrnParseException e) {
                err.println("The specified slice URN is invalid: \"" + sliceIdArg + "\"");
                help(options);
                return 3;
            }
        else
            sliceUrn = GeniUrn.createGeniUrnFromEncodedParts(user.getUserAuthority().getNameForUrn(), "slice", sliceIdArg);
        if (debug)
            out.println("Slice URN : " + sliceUrn);

        JFedConnectionProvider connectionPool = injector.getInstance(JFedConnectionProvider.class);

        AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory automaticUserAndSliceApiWrapperFactory =
                injector.getInstance(AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory.class);
        AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory automaticAggregateManagerWrapperFactory =
                injector.getInstance(AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory.class);

        List<AnyCredential> sliceCredentials;
        try {
            if (!silent)
                out.println("Fetching user and slice info.");
            UserAndSliceApiWrapper saWrapper = automaticUserAndSliceApiWrapperFactory.create();
            List<AnyCredential> userCredentials = saWrapper.getUserCredentials(logger, user.getUserUrn());
            sliceCredentials = saWrapper.getSliceCredentials(logger, sliceUrn);

            if (sliceComponentManagers.isEmpty()) {
                if (!silent)
                    out.println("Searching component managers used in slice.");
                List<String> res = saWrapper.getAggregatesForSlice(logger, sliceCredentials, sliceUrn);
                if (debug)
                    out.println("Found component managers used in slice: " + res);
                if (res != null)
                    sliceComponentManagers.addAll(res);
            }
        } catch (JFedException e) {
            err.println("\nError retrieving slice data: " + e.getMessage());
            e.printStackTrace();
            return 1;
            //throw new RuntimeException("never reached");
        }

        if (sliceComponentManagers.isEmpty()) {
            System.err.println("\nWarning: no known component managers for " + sliceUrn + " (try the --rspec option if this is not what you expected)");
            return 0;
        }

        if (debug)
            out.println("Will now delete slivers on: " + sliceComponentManagers);

        for (String sliceComponentManagerUrnString : sliceComponentManagers) {
            SfaAuthority targetAuth = authorityFinder.findByUrn(sliceComponentManagerUrnString, AuthorityFinder.Purpose.CREATE_SLIVER);
            if (targetAuth == null) {
                err.println("Warning: could not find authority matching \"" + sliceComponentManagerUrnString + "\"");
                continue;
            }
            if (!silent)
                out.println("Deleting sliver(s) on  component manager: " + sliceComponentManagerUrnString);

            AggregateManagerWrapper amWrapper;
            if (amApiVersionArg == null) {
                amWrapper = automaticAggregateManagerWrapperFactory.create(targetAuth);
            } else {
                if (amApiVersionArg.equals("2"))
                    amWrapper = new AMv2Wrapper(logger, userProvider, connectionPool, jFedPreferences, targetAuth);
                else {
                    if (amApiVersionArg.equals("3"))
                        amWrapper = new AMv3Wrapper(logger, userProvider, connectionPool, jFedPreferences, targetAuth);
                    else
                        throw new RuntimeException("Unsupported AM API: \"" + amApiVersionArg + "\"");
                }
            }

            try {
                amWrapper.deleteSliver(sliceUrn, sliceCredentials);
                out.println("  sliver deleted.");
            } catch (JFedException e) {
                err.println("\nError deleting sliver on " + targetAuth.getUrn() + ": " + e.getMessage());
            }
        }

        if (debug)
            out.println("All done. Exiting.");

        return 0;
    }
}
