package be.iminds.ilabt.jfed.experimenter_gui.bugreporting;

import be.iminds.ilabt.jfed.call_log_output.CallReportWriter;
import be.iminds.ilabt.jfed.call_log_output.SerializableApiCallDetails;
import be.iminds.ilabt.jfed.call_log_output.SerializableApiCallDetailsWriter;
import be.iminds.ilabt.jfed.connectivity_tester.ConnectivityTest;
import be.iminds.ilabt.jfed.experimenter_gui.connectivity_tester.ConnectivityTester;
import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentViewController;
import be.iminds.ilabt.jfed.highlevel.controller.JavaFXTaskThread;
import be.iminds.ilabt.jfed.highlevel.history.StitchingReportHistory;
import be.iminds.ilabt.jfed.highlevel.model.EasyModel;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.highlevel.stitcher.StitchingReport;
import be.iminds.ilabt.jfed.logging.BugReportAppender;
import be.iminds.ilabt.jfed.logging.JFedLoggerManager;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.ErrorDetails;
import be.iminds.ilabt.jfed.lowlevel.userloginmodel.UserLoginModelManager;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.KeyUtil;
import ch.qos.logback.classic.spi.LoggingEvent;
import javafx.beans.property.*;
import javafx.collections.ObservableMap;
import javafx.scene.image.Image;
import javanet.staxutils.IndentingXMLEventWriter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.xml.bind.JAXBException;
import javax.xml.stream.XMLEventFactory;
import javax.xml.stream.XMLEventWriter;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.security.cert.CertificateParsingException;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPublicKey;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.Properties;

import static be.iminds.ilabt.jfed.connectivity_tester.ConnectivityTest.ConnectivityTestResult;
import static be.iminds.ilabt.jfed.experimenter_gui.bugreporting.BugReportConstants.*;
import static be.iminds.ilabt.jfed.highlevel.model.Sliver.ComponentInfo;
import static ch.qos.logback.classic.spi.ThrowableProxyUtil.asString;
import static java.nio.charset.Charset.defaultCharset;
import static java.nio.file.Files.readAllLines;
import static java.time.Instant.ofEpochMilli;
import static java.time.ZoneId.systemDefault;
import static java.time.ZonedDateTime.ofInstant;
import static java.time.format.DateTimeFormatter.ISO_OFFSET_DATE_TIME;
import static java.util.Map.Entry;
import static javafx.embed.swing.SwingFXUtils.fromFXImage;
import static javax.imageio.ImageIO.write;
import static javax.xml.stream.XMLEventFactory.newInstance;
import static org.apache.commons.codec.binary.Base64.encodeBase64;

/**
 * This class contains references to all the information that must be included in a bugreport.
 * <p/>
 * It also extends the Task-class, which must be used to write the info to an outputstream. By extending from Task,
 * we are able to display the overall progress to the user.
 * <p/>
 * User: twalcari
 * Date: 6/13/2014
 * Time: 10:35
 */
public class BugReportComposer {
    private final static Logger LOG = LoggerFactory.getLogger(BugReportComposer.class);

    private static final String VERSION_PROPERTIES_FILE = "/version.properties";
    private static final String PROP_VERSION = "version";

    private static final String PROP_GIT_COMMIT = "build.git_commit";
    private static final String PROP_GIT_BRANCH = "build.git_branch";

    private static final String PROP_BUILD_JENKINS = "build.jenkins_build";

    private static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat("yyyy-MM-dd");
    private static final int SCREENSHOT_PROGRESS = 100;

    private final JavaFXTaskThread javaFXTaskThread;
    private final UserLoginModelManager userLoginModelManager;
    private final EasyModel easyModel;
    private final JFedGuiPreferences jFedPreferences;
    private final JFedLoggerManager jFedLoggerManager;
    private final ConnectivityTester connectivityTester;
    private final ObservableMap<GeniUrn, ExperimentViewController> experimentViewControllers;
    private final CallReportWriter callReportWriter;
    private final StitchingReportHistory stitchingReportHistory;


    private final StringProperty description = new SimpleStringProperty();
    private final StringProperty email = new SimpleStringProperty();
    private final ReadOnlyStringWrapper version = new ReadOnlyStringWrapper();
    private final ReadOnlyStringWrapper environment = new ReadOnlyStringWrapper();
    private final ReadOnlyStringWrapper credentials = new ReadOnlyStringWrapper();
    private final ReadOnlyStringWrapper certificateExpiration = new ReadOnlyStringWrapper();

    private final ObjectProperty<Image> screenshot = new SimpleObjectProperty<>();


    @Inject
    public BugReportComposer(JavaFXTaskThread javaFXTaskThread, UserLoginModelManager userLoginModelManager,
                             EasyModel easyModel, ConnectivityTester connectivityTester,
                             ObservableMap<GeniUrn, ExperimentViewController> experimentViewControllers,
                             JFedGuiPreferences jFedPreferences, JFedLoggerManager jFedLoggerManager,
                             CallReportWriter callReportWriter, StitchingReportHistory stitchingReportHistory) {
        this.javaFXTaskThread = javaFXTaskThread;
        this.userLoginModelManager = userLoginModelManager;
        this.easyModel = easyModel;
        this.connectivityTester = connectivityTester;
        this.experimentViewControllers = experimentViewControllers;
        this.jFedPreferences = jFedPreferences;
        this.jFedLoggerManager = jFedLoggerManager;
        this.callReportWriter = callReportWriter;
        this.stitchingReportHistory = stitchingReportHistory;

        assert javaFXTaskThread != null;
        assert connectivityTester != null;
        assert connectivityTester.getTestResults() != null;
        assert experimentViewControllers != null;

        //start connectivity tests if this hasn't happened yet
        this.connectivityTester.startTests();


        Properties prop = new Properties();
        //now load the version-properties file
        try {
            prop.load(getClass().getResourceAsStream(VERSION_PROPERTIES_FILE));
        } catch (IOException e) {
            throw new RuntimeException("Something went wrong showing the Bug reporting dialog: " + e.getMessage(), e);

        }

        //now load the info into the disabled textfield
        StringBuilder version = new StringBuilder();
        version.append(prop.getProperty(PROP_VERSION, "Unknown"));

        final String jenkinsBuildNr = prop.getProperty(PROP_BUILD_JENKINS);
        if (jenkinsBuildNr != null && !jenkinsBuildNr.isEmpty())
            version.append("  - build #").append(jenkinsBuildNr);

        final String gitCommit = prop.getProperty(PROP_GIT_COMMIT);
        if (gitCommit != null && !gitCommit.isEmpty())
            version.append("  - git commit #").append(gitCommit);

        final String gitBranch = prop.getProperty(PROP_GIT_BRANCH);
        if (gitBranch != null && !gitBranch.isEmpty())
            version.append(" on ").append(gitBranch);

        this.version.set(version.toString());

        this.environment.set(
                System.getProperty("os.name") + " " + System.getProperty("os.version") + " " + System.getProperty("os.arch")
                        + " - Java " + System.getProperty("java.version")
                        + " (" + System.getProperty("java.vendor") + ")");

        this.credentials.set(this.userLoginModelManager.getUserUrnString());

        //try to get user e-mailaddress
        if (this.userLoginModelManager.getUserLoginModelType() ==
                UserLoginModelManager.UserLoginModelType.KEY_CERT_INTERNAL_INFO) {

            try {
                for (List<?> list : this.userLoginModelManager.getClientCertificateChain().get(0)
                        .getSubjectAlternativeNames()) {
                    if (list.get(0).equals(1)) {  //if 1, then this is a rfc822Name-field
                        this.email.set(list.get(1).toString());
                        break;
                    }
                }

                Date expirationDate = getEarliestExpire(this.userLoginModelManager.getClientCertificateChain());
                certificateExpiration.set(DATE_FORMAT.format(expirationDate));

            } catch (CertificateParsingException e) {
                e.printStackTrace();
            }
        }


    }

    private Date getEarliestExpire(List<X509Certificate> certs) {
        Date res = null;
        for (X509Certificate cert : certs) {
            Date expire = cert.getNotAfter();
            if (res == null || res.after(expire))
                res = expire;
        }
        return res;
    }


    private static final int PROGRESS_OVERALL_INFO = 7;

    private int totalProgress;
    private int progress;
    private boolean finished = false;

    /**
     * Writes a bug-report to the output stream set in the field outputStream. It uses StaX instead of JAXB to limit memory usage.
     * <p/>
     * This function expects the connectivityTests to have completed.
     *
     * @throws java.io.IOException
     */
    public void writeToOutputStream(OutputStream outputStream) throws IOException {
        if (finished)
            throw new IllegalStateException("Composer has already finished");


        if (!isPreparationFinished())
            throw new IllegalStateException("Preparation has not finished!");

        if (outputStream == null)
            throw new IllegalStateException("OutputStream must be set before starting the task!");


        //initialize logreader
        BugReportAppender bugReportAppender = jFedLoggerManager.getBugReportMemoryAppender();

        if (bugReportAppender == null)
            LOG.error("bugReportAppender \"BUGREPORT\" not found! Will not be able to add log entries to bug report.");

        progress = 0;
        //determine the total progress to be made
        totalProgress = PROGRESS_OVERALL_INFO
                + javaFXTaskThread.getCallsCount() * 10
                + (bugReportAppender == null ? 0 : bugReportAppender.getLength())
                + connectivityTester.getTestResults().size()
                + experimentViewControllers.size()
                + stitchingReportHistory.size() * 50
                + (screenshot.get() != null ? SCREENSHOT_PROGRESS : 0);

        XMLOutputFactory xof = XMLOutputFactory.newInstance();

        try {
            XMLEventWriter xtw = new IndentingXMLEventWriter(xof.createXMLEventWriter(outputStream, "UTF-8"));
            XMLEventFactory eventFactory = newInstance();

            xtw.add(eventFactory.createStartDocument());

            xtw.add(eventFactory.createStartElement(Q_BUGREPORT, null, null));

            //write overall bugreport info to the stream
            xtw.add(eventFactory.createStartElement(Q_DESCRIPTION, null, null));
            xtw.add(eventFactory.createCharacters(description.get()));
            xtw.add(eventFactory.createEndElement(Q_DESCRIPTION, null)); //description

            xtw.add(eventFactory.createStartElement(Q_EMAIL, null, null));
            xtw.add(eventFactory.createCharacters(email.get()));
            xtw.add(eventFactory.createEndElement(Q_EMAIL, null));

            xtw.add(eventFactory.createStartElement(Q_VERSION, null, null));
            xtw.add(eventFactory.createCharacters(version.get()));
            xtw.add(eventFactory.createEndElement(Q_VERSION, null));

            xtw.add(eventFactory.createStartElement(Q_ENVIRONMENT, null, null));
            xtw.add(eventFactory.createCharacters(environment.get()));
            xtw.add(eventFactory.createEndElement(Q_ENVIRONMENT, null));

            xtw.add(eventFactory.createStartElement(Q_REPORTER_CREDENTIAL, null, null));
            xtw.add(eventFactory.createCharacters(credentials.get()));
            xtw.add(eventFactory.createEndElement(Q_REPORTER_CREDENTIAL, null));

            xtw.add(eventFactory.createStartElement(Q_LOGGEDIN_USER_INFO, null, null));
            xtw.add(eventFactory.createCharacters(getLoggedInUserInfo()));
            xtw.add(eventFactory.createEndElement(Q_LOGGEDIN_USER_INFO, null));


            xtw.add(eventFactory.createStartElement(Q_PREFERENCES, null, null));

            List<String> lines = readAllLines(jFedPreferences.getPropertiesFile().toPath(),
                    defaultCharset());
            for (String line : lines) {
                xtw.add(eventFactory.createCharacters(line));
                xtw.add(eventFactory.createCharacters("\n"));
            }

            xtw.add(eventFactory.createEndElement(Q_PREFERENCES, null)); //preferences

            progress += PROGRESS_OVERALL_INFO;

            xtw.flush();


            //write all the calls to the stream
            xtw.add(eventFactory.createStartElement(Q_CALLS, null, null));

            synchronized (javaFXTaskThread) {
                for (JavaFXTaskThread.FXTaskExecution taskExecution : javaFXTaskThread.getAllTasks()) {
                    for (SerializableApiCallDetails details : taskExecution.getApiCallHistory()) {

                        SerializableApiCallDetailsWriter.writeApiCallDetails(details, xtw, eventFactory);
                        xtw.flush();
                    }

                    progress += 10;
                }
            }

            xtw.add(eventFactory.createEndElement(Q_CALLS, null)); //calls

            xtw.add(eventFactory.createStartElement(Q_STITCHING_REPORTS, null, null));

            synchronized (stitchingReportHistory) {
                for (StitchingReport stitchingReport : stitchingReportHistory.getStitchingReports()) {
                    callReportWriter.writeCallReport(stitchingReport, xtw, eventFactory, true);
                    xtw.flush();

                    progress += 50;
                }
            }
            xtw.add(eventFactory.createEndElement(Q_STITCHING_REPORTS, null)); //stitchingreports


            //write all log-lines to the stream
            xtw.add(eventFactory.createStartElement(Q_LOG, null, null));

            if (bugReportAppender != null) {
                List<LoggingEvent> loggingEvents = bugReportAppender.getListCopy();
                for (LoggingEvent loggingEvent : loggingEvents) {

                    writeLogLine(loggingEvent, xtw, eventFactory);
                    progress++;

                }
            }


            xtw.add(eventFactory.createEndElement(Q_LOG, null)); //log
            xtw.flush();

            //write all connectivity tests to the stream
            xtw.add(eventFactory.createStartElement(Q_CONNECTIVITY_TESTS, null, null));
            for (Entry<ConnectivityTest, ConnectivityTestResult> entry : connectivityTester.getTestResults().entrySet()) {
                writeConnectivityTest(entry, xtw, eventFactory);

                progress++;
            }

            xtw.add(eventFactory.createEndElement(Q_CONNECTIVITY_TESTS, null));  //connectivity-tests
            xtw.flush();

            //write current slice information to the stream

            xtw.add(eventFactory.createStartElement(Q_SLICES, null, null));


            for (Slice slice : easyModel.getSlices()) {
                writeSlice(slice, xtw, eventFactory);

                progress++;
            }


            xtw.add(eventFactory.createEndElement(Q_SLICES, null)); //slices

            xtw.add(eventFactory.createStartElement(Q_SCREENSHOT, null, null));

            if (screenshot.get() != null) {

                try {
                    ByteArrayOutputStream screenshotBuffer = new ByteArrayOutputStream();
                    write(fromFXImage(screenshot.get(), null), "png", screenshotBuffer);
                    byte[] encoded = encodeBase64(screenshotBuffer.toByteArray());
                    xtw.add(eventFactory.createCharacters(new String(encoded)));
                } catch (IOException ex) {
                    LOG.warn("Error while writing bugreport screenshot: ignoring!", ex);
                }

                progress += SCREENSHOT_PROGRESS;
            }


            xtw.add(eventFactory.createEndElement(Q_SCREENSHOT, null)); //screenshot

            xtw.add(eventFactory.createEndElement(Q_BUGREPORT, null)); //bugreport
            xtw.flush();

        } catch (XMLStreamException | JAXBException ex) {
            throw new IOException("Error while writing bug report to stream", ex);
        }

        finished = true;
    }

    private void writeSlice(Slice slice, XMLEventWriter xtw, XMLEventFactory eventFactory) throws XMLStreamException {
        xtw.add(eventFactory.createStartElement(Q_SLICE, null, null));

        xtw.add(eventFactory.createStartElement(Q_SLICE_URN, null, null));
        xtw.add(eventFactory.createCharacters(slice.getUrnString()));
        xtw.add(eventFactory.createEndElement(Q_SLICE_URN, null));

        for (Sliver sliver : slice.getSlivers()) {
            xtw.add(eventFactory.createStartElement(Q_SLIVER, null, null));

            xtw.add(eventFactory.createStartElement(Q_SLIVER_URN, null, null));
            String sliverUrn = sliver.getUrn() != null ?
                    sliver.getUrn().toString() :
                    "SLIVER-ON: " + sliver.getAuthority().getUrnString();

            xtw.add(eventFactory.createCharacters(sliverUrn));
            xtw.add(eventFactory.createEndElement(Q_SLIVER_URN, null)); //sliverUrn

            xtw.add(eventFactory.createStartElement(Q_CURRENT_STATUS, null, null));
            xtw.add(eventFactory.createCharacters(sliver.getStatusString()));
            xtw.add(eventFactory.createEndElement(Q_CURRENT_STATUS, null));

            xtw.add(eventFactory.createStartElement(Q_COMPONENTS, null, null));
            for (ComponentInfo ci : sliver.getComponentInfo()) {
                if (ci.getClientId() != null || ci.getComponentId() != null) {
                    xtw.add(eventFactory.createStartElement(Q_COMPONENT, null, null));
                    if (ci.getClientId() == null)
                        xtw.add(eventFactory.createCharacters(ci.getClientId()));
                    else
                        xtw.add(eventFactory.createCharacters(ci.getComponentId()));
                    xtw.add(eventFactory.createEndElement(Q_COMPONENT, null)); //component
                }
            }
            xtw.add(eventFactory.createEndElement(Q_COMPONENTS, null)); //components

            if (sliver.getStatus() != null) {
                xtw.add(eventFactory.createStartElement(Q_SLIVER_STATUS, null, null));
                xtw.add(eventFactory.createCharacters(sliver.getStatus().getGlobalStatus().toString()));
                xtw.add(eventFactory.createEndElement(Q_SLIVER_STATUS, null));

                for (ErrorDetails errorDetails : sliver.getStatus().getAllErrors()) {
                    xtw.add(eventFactory.createStartElement(Q_ERROR, null, null));
                    xtw.add(eventFactory.createAttribute(Q_TIMESTAMP,
                            ISO_OFFSET_DATE_TIME.format(ofInstant(
                                    ofEpochMilli(errorDetails.getTimestamp()), systemDefault()))));
                    if (errorDetails.getSliverUrn() != null)
                        xtw.add(eventFactory.createAttribute(Q_SLIVER_URN, errorDetails.getSliverUrn().toString()));
                    if (errorDetails.getComponentUrn() != null)
                        xtw.add(eventFactory.createAttribute(Q_COMPONENT_URN, errorDetails.getComponentUrn().toString()));
                    xtw.add(eventFactory.createCharacters(errorDetails.getError()));
                    xtw.add(eventFactory.createEndElement(Q_ERROR, null)); //error
                }
            }

            xtw.add(eventFactory.createEndElement(Q_SLIVER, null)); //sliver
        }

        xtw.add(eventFactory.createEndElement(Q_SLICE, null));//slice
    }

    private void writeConnectivityTest(Entry<ConnectivityTest, ConnectivityTestResult> connectivityTestEntry, XMLEventWriter xtw, XMLEventFactory eventFactory) throws XMLStreamException {
        xtw.add(eventFactory.createStartElement(Q_CONNECTIVITY_TEST, null, null));

        xtw.add(eventFactory.createStartElement(Q_NAME, null, null));
        xtw.add(eventFactory.createCharacters(connectivityTestEntry.getKey().getName()));
        xtw.add(eventFactory.createEndElement(Q_NAME, null));

        if (connectivityTestEntry.getValue() != null) {
            xtw.add(eventFactory.createStartElement(Q_STATUS, null, null));
            xtw.add(eventFactory.createCharacters(connectivityTestEntry.getValue().getStatus().toString()));
            xtw.add(eventFactory.createEndElement(Q_STATUS, null));

            xtw.add(eventFactory.createStartElement(Q_MESSAGE, null, null));
            xtw.add(eventFactory.createCharacters(connectivityTestEntry.getValue().getMessage()));
            xtw.add(eventFactory.createEndElement(Q_MESSAGE, null));

            if (connectivityTestEntry.getValue().getExceptionString() != null) {
                xtw.add(eventFactory.createStartElement(Q_EXCEPTION, null, null));
                xtw.add(eventFactory.createCharacters(connectivityTestEntry.getValue().getExceptionString()));
                xtw.add(eventFactory.createEndElement(Q_EXCEPTION, null));
            }
        }

        xtw.add(eventFactory.createEndElement(Q_CONNECTIVITY_TEST, null)); //connectivityTests
    }

    private void writeLogLine(LoggingEvent loggingEvent, XMLEventWriter xtw, XMLEventFactory eventFactory) throws XMLStreamException {
        xtw.add(eventFactory.createStartElement(Q_LOGLINE, null, null));

        xtw.add(eventFactory.createStartElement(Q_LEVEL, null, null));
        xtw.add(eventFactory.createCharacters(loggingEvent.getLevel().toString()));
        xtw.add(eventFactory.createEndElement(Q_LEVEL, null));

        xtw.add(eventFactory.createStartElement(Q_CLASSNAME, null, null));
        xtw.add(eventFactory.createCharacters(loggingEvent.getLoggerName()));
        xtw.add(eventFactory.createEndElement(Q_CLASSNAME, null));

        xtw.add(eventFactory.createStartElement(Q_DATETIME, null, null));
        xtw.add(eventFactory.createCharacters(ISO_OFFSET_DATE_TIME.format(ofInstant(ofEpochMilli(loggingEvent.getTimeStamp()), systemDefault()))));
        xtw.add(eventFactory.createEndElement(Q_DATETIME, null));


        xtw.add(eventFactory.createStartElement(Q_TIME, null, null));
        xtw.add(eventFactory.createCharacters(Long.toString(loggingEvent.getTimeStamp())));
        xtw.add(eventFactory.createEndElement(Q_TIME, null));

        xtw.add(eventFactory.createStartElement(Q_THREADNAME, null, null));
        xtw.add(eventFactory.createCharacters(loggingEvent.getThreadName()));
        xtw.add(eventFactory.createEndElement(Q_THREADNAME, null));

        xtw.add(eventFactory.createStartElement(Q_MESSAGE, null, null));
        xtw.add(eventFactory.createCharacters(loggingEvent.getFormattedMessage()));
        xtw.add(eventFactory.createEndElement(Q_MESSAGE, null));

        if (loggingEvent.getThrowableProxy() != null) {
            xtw.add(eventFactory.createStartElement(Q_EXCEPTION, null, null));
            xtw.add(eventFactory.createCharacters(asString(loggingEvent.getThrowableProxy())));
            xtw.add(eventFactory.createEndElement(Q_EXCEPTION, null));
        }

        xtw.add(eventFactory.createEndElement(Q_LOGLINE, null)); //logline
    }

    public void reset() {
        progress = 0;
        totalProgress = 0;
        finished = false;
    }

    public boolean isFinished() {
        return finished;
    }

    public int getTotalProgress() {
        return totalProgress;
    }

    public int getProgress() {
        return progress;
    }

    public boolean isPreparationFinished() {
        return connectivityTester.getTestsFinished();

    }

    private String getLoggedInUserInfo() {
        String userInfoStr = null;
        try {
            if (userLoginModelManager.isUserLoggedIn()) {
                GeniUser u = userLoginModelManager.getLoggedInGeniUser();
                //TODO: certInfo would better contain info about the certificate such as all data that can be extracted.
                //      PEM does contain all that info but is not human readable.
                userInfoStr = u.getUserUrnString() + " @ " + u.getUserAuthority().getUrn();
                if (u.getPrivateKey() == null)
                    userInfoStr += " WARNING, no private key!";
                try {
                    String pubkey =
                            KeyUtil.rsaPublicKeyToOpenSshAuthorizedKeysFormatData((RSAPublicKey) u.getClientCertificateChain().get(0).getPublicKey());
                    userInfoStr += " with pubkey " + pubkey;
                } catch (Throwable e) {
                    userInfoStr += " ERROR getting pubkey " + e.getMessage();
                }
                try {
                    String certInfo = KeyUtil.x509certificateChainToPem(u.getClientCertificateChain());
                    userInfoStr += " with cert " + certInfo;
                } catch (Throwable e) {
                    userInfoStr += " ERROR getting cert " + e.getMessage();
                }
            } else
                userInfoStr = "No user logged in";
        } catch (Throwable e) {
            if (userInfoStr == null)
                userInfoStr = "NULL then ";
            userInfoStr += " ERROR getting user info " + e.getMessage();
        }
        return userInfoStr;
    }

    public String getDescription() {
        return description.get();
    }

    public StringProperty descriptionProperty() {
        return description;
    }

    public void setDescription(String description) {
        this.description.set(description);
    }

    public String getEmail() {
        return email.get();
    }

    public StringProperty emailProperty() {
        return email;
    }

    public void setEmail(String email) {
        this.email.set(email);
    }

    public String getVersion() {
        return version.get();
    }

    public ReadOnlyStringProperty versionProperty() {
        return version.getReadOnlyProperty();
    }

    public String getEnvironment() {
        return environment.get();
    }

    public ReadOnlyStringProperty environmentProperty() {
        return environment.getReadOnlyProperty();
    }


    public String getCredentials() {
        return credentials.get();
    }

    public ReadOnlyStringProperty credentialsProperty() {
        return credentials.getReadOnlyProperty();
    }

    public String getCertificateExpiration() {
        return certificateExpiration.get();
    }

    public ReadOnlyStringProperty certificateExpirationProperty() {
        return certificateExpiration.getReadOnlyProperty();
    }

    public ReadOnlyIntegerProperty callsCountProperty() {
        return javaFXTaskThread.callsCountProperty();
    }

    public int stitchingReportCount() {
        return stitchingReportHistory.size();
    }

    public Image getScreenshot() {
        return screenshot.get();
    }

    public ObjectProperty<Image> screenshotProperty() {
        return screenshot;
    }

    public void setScreenshot(Image screenshot) {
        this.screenshot.set(screenshot);
    }
}
