package be.iminds.ilabt.jfed.experimenter_gui.dialogs;

import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentController;
import be.iminds.ilabt.jfed.experimenter_gui.slice.tasks.ExperimentTasksFactory;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticAggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.ui.javafx.util.InformationLabel;
import com.google.common.base.Joiner;
import com.google.common.collect.Collections2;
import javafx.scene.control.*;
import javafx.scene.control.cell.CheckBoxTreeCell;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.stage.Window;
import javafx.util.Callback;
import org.controlsfx.control.action.Action;
import org.controlsfx.dialog.Dialog;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

/**
 */
public class RebootDialog extends BorderPane {
    private final static Logger LOG = LoggerFactory.getLogger(RebootDialog.class);

    private final Window parentStage;
    private final ExperimentController experimentController;
    private final ExperimentTasksFactory experimentTasksFactory;
    private final JFedGuiPreferences jFedPreferences;
    private final AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory
            automaticAggregateManagerWrapperFactory;

    private final Slice slice;

    private final TreeView<Rebootable> treeView;

    private final boolean hasRebootableItems;

    public RebootDialog(Window parent, ExperimentController experimentController, ExperimentTasksFactory experimentTasksFactory, JFedGuiPreferences jFedPreferences, AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory automaticAggregateManagerWrapperFactory) {
        this.parentStage = parent;
        this.experimentController = experimentController;
        this.experimentTasksFactory = experimentTasksFactory;
        this.jFedPreferences = jFedPreferences;
        this.automaticAggregateManagerWrapperFactory = automaticAggregateManagerWrapperFactory;

        this.slice = experimentController.getExperiment().getSlice();

        VBox center = new VBox(10);

        setCenter(center);

        List<String> informations = new ArrayList<>();

        boolean hasRebootableItems = false;

        treeView = new TreeView<>();

        treeView.setEditable(true);
        treeView.setCellFactory(new Callback<TreeView<Rebootable>, TreeCell<Rebootable>>() {
            @Override
            public TreeCell<Rebootable> call(TreeView<Rebootable> rebootableTreeView) {
                return new CheckBoxTreeCell<Rebootable>() {
                    @Override
                    public void updateItem(Rebootable rebootable, boolean empty) {
                        super.updateItem(rebootable, empty);
                        if (rebootable != null && !empty) {
                            setEditable(rebootable.isRebootable());

                        }
                    }
                };
            }
        });

        CheckBoxTreeItem<Rebootable> rootItem = new CheckBoxTreeItem<>(new RebootableExperiment());
        rootItem.setExpanded(true);
        treeView.setRoot(rootItem);
        int itemsCount = 1;

        Set<SfaAuthority> authorities = slice.getSlivers().stream()
                .map(Sliver::getAuthority)
                .filter(SfaAuthority::isReal)
                .collect(Collectors.toSet());

        for (SfaAuthority authority : authorities) {
            CheckBoxTreeItem<Rebootable> authorityTreeItem =
                    new CheckBoxTreeItem<>(new RebootableAuthority(authority));
            authorityTreeItem.setExpanded(true);

            List<Sliver> slivers = slice.findSlivers(authority);

            for (Sliver sliver : slivers) {
                Collection<RspecNode> nodes = sliver.getNodes();
                if (nodes != null && !nodes.isEmpty()) {
                    RebootableSliver rebootableSliver = new RebootableSliver(sliver, nodes);
                    if (rebootableSliver.isRebootable()) {
                        authorityTreeItem.getChildren().add(new CheckBoxTreeItem<>(rebootableSliver));
                    }
                }
            }

            if (!authorityTreeItem.getChildren().isEmpty()) {
                rootItem.getChildren().add(authorityTreeItem);
                hasRebootableItems = true;
                itemsCount += authorityTreeItem.getChildren().size() + 1;
            } else {
                informations.add(String.format("%s does not have rebootable resources.", authority.getHrn()));
            }

        }
        rootItem.setSelected(true);
        treeView.setPrefWidth(100);
        treeView.setPrefHeight(Math.min(1 + (itemsCount * 25), 250));

        center.getChildren().add(treeView);

        if (!informations.isEmpty()) {
            InformationLabel label = new InformationLabel(Joiner.on("\n").join(informations));
            center.getChildren().add(label);
            //label.prefWidthProperty().bind(center.widthProperty()); //causes height-bug  FEDIBBTDEV-747
        }

        this.hasRebootableItems = hasRebootableItems;
    }

    public void showDialog() {
        if (!hasRebootableItems) {
            JFDialogs.create()
                    .owner(parentStage)
                    .message("This experiment doesn't contain any resources that support rebooting.")
                    .masthead("Rebooting is unavailable")
                    .showWarning();
            return;
        }

        Dialog dialog = new Dialog(parentStage, "Reboot experiment nodes");
        dialog.setMasthead("Please select the testbeds on which the nodes must be rebooted");
        dialog.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.POWER_OFF, Color.RED));
        dialog.setContent(this);
        dialog.getActions().setAll(Dialog.ACTION_OK, Dialog.ACTION_CANCEL);

        Action response = dialog.show();

        if (response != Dialog.ACTION_OK)
            return;

        long numberOfSliversSelected =
                treeView.getRoot().getChildren().stream().flatMap(item -> item.getChildren().stream())
                        .filter(item -> ((CheckBoxTreeItem<Rebootable>) item).isSelected())
                        .count();

        if (numberOfSliversSelected > 0) {

            Alert confirmAlert = new Alert(Alert.AlertType.WARNING);
            confirmAlert.setContentText(String.format("You are about to reboot %d sliver(s). All data in memory WILL BE LOST. Are you sure that you want to continue?", numberOfSliversSelected));
            confirmAlert.setTitle("Rebooting slivers");

            ButtonType rebootButtonType = new ButtonType("Reboot slivers", ButtonBar.ButtonData.OTHER);

            confirmAlert.getButtonTypes().setAll(rebootButtonType, ButtonType.CANCEL);

            ((Button)confirmAlert.getDialogPane().lookupButton(ButtonType.CANCEL)).setDefaultButton(true);


            if (confirmAlert.showAndWait().orElse(ButtonType.CANCEL) == ButtonType.CANCEL)
                return;

        }

        //we do not process the root
        for (TreeItem<Rebootable> child : treeView.getRoot().getChildren()) {
            processTreeItem((CheckBoxTreeItem<Rebootable>) child);
        }
    }

    private void processTreeItem(CheckBoxTreeItem<Rebootable> treeItem) {
        if (!treeItem.isIndeterminate() && treeItem.isSelected()) {
            treeItem.getValue().reboot();
        } else {
            for (TreeItem<Rebootable> child : treeItem.getChildren()) {
                processTreeItem((CheckBoxTreeItem<Rebootable>) child);
            }
        }
    }

    private static interface Rebootable {
        public void reboot();

        public boolean isRebootable();
    }

    private static class RebootableExperiment implements Rebootable {

        @Override
        public void reboot() {
            //do nothing
        }

        @Override
        public boolean isRebootable() {
            return true;
        }

        @Override
        public String toString() {
            return "All nodes";
        }
    }

    private class RebootableSliver implements Rebootable {
        private Sliver sliver;
        private boolean canReboot;
        private Collection<RspecNode> nodes;

        private RebootableSliver(Sliver sliver, Collection<RspecNode> nodes) {
            this.sliver = sliver;
            this.nodes = nodes;

            canReboot =
                    automaticAggregateManagerWrapperFactory.create(sliver.getAuthority()).isRestartSupported();
        }

        public Sliver getSliver() {
            return sliver;
        }

        public boolean isRebootable() {
            return canReboot;
        }

        public void reboot() {
            experimentController.submitExperimentTask(experimentTasksFactory.createRebootSliverTask(experimentController, sliver));
        }

        @Override
        public String toString() {
            return Joiner.on(", ").join(Collections2.transform(nodes, rspecNode -> rspecNode.getClientId() != null
                    ? rspecNode.getClientId() : "<NO ID>"));
        }
    }

    private class RebootableAuthority implements Rebootable {
        private SfaAuthority auth;
        private boolean canReboot;

        private RebootableAuthority(SfaAuthority auth) {
            this.auth = auth;

            canReboot = automaticAggregateManagerWrapperFactory.create(auth).isRestartSupported();

        }

        public SfaAuthority getAuthority() {
            return auth;
        }

        public boolean isRebootable() {
            return canReboot;
        }

        public void reboot() {
            experimentController.submitExperimentTask(experimentTasksFactory.createRebootSliverTask(experimentController, auth));
        }

        @Override
        public String toString() {
            return auth.getHrn();
        }
    }
}

