package be.iminds.ilabt.jfed.experimenter_gui.slice;

import be.iminds.ilabt.jfed.experimenter_gui.slice.tasks.CheckExpirationTaskGenerator;
import be.iminds.ilabt.jfed.experimenter_gui.slice.tasks.ExperimentTask;
import be.iminds.ilabt.jfed.experimenter_gui.slice.tasks.ExperimentTaskGenerator;
import be.iminds.ilabt.jfed.experimenter_gui.slice.tasks.ExperimentTasksFactory;
import be.iminds.ilabt.jfed.highlevel.model.AuthorityList;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import javafx.stage.Window;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.time.Duration;
import java.time.Instant;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

/**
 * This class will take care of an Experiment-object. It makes sure that the appropriate actions are taken when
 * the state of the Slice, or one of it's slivers changes.
 * <p/>
 * User: twalcari
 * Date: 1/30/2015
 * Time: 14:04
 */
public class ExperimentController implements ExperimentChangeListener {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimentController.class);
    private final Experiment experiment;
    private final ExperimentTasksFactory experimentTasksFactory;
    private final AuthorityList authorityList;

    private static final ExecutorService executorService = Executors.newCachedThreadPool();
    private final ScheduledExecutorService scheduledExecutorService = Executors.newScheduledThreadPool(2);

    private Window parentWindow = null;
    //helper variable for showing dialog messages. Ideally, this should be refactored away

    private List<ExperimentControllerListener> listeners = new ArrayList<>();
    private final ExpirationDetector expirationDetector = new ExpirationDetector();

    public ExperimentController(Experiment experiment, ExperimentTasksFactory experimentTasksFactory, AuthorityList authorityList) {
        this.experiment = experiment;
        this.experimentTasksFactory = experimentTasksFactory;
        this.authorityList = authorityList;

        experiment.addExperimentChangeListener(this);

    }

    /**
     * This method must be called for the experimentController to start taking actions.
     */
    public void start() {
        //bootstrap state machine
        LOG.debug("Bootstrapping experiment controller of '{}' in state {}", experiment.getName(), experiment.getSliceState());
        onSliceStateChange(experiment.getSliceState());
    }


    @Override
    public void onSliceStateChange(SliceState newSliceState) {
        switch (newSliceState) {
            case PENDING:
                createSlice();
                break;
            case RESTORING:
                restoreSlice();
                break;
            case CREATED:
                startSliceExpirationDetection();
                if (experiment.getNewRequestRspecSource() != null) {
                    if (experiment.getRequestedStartTime() != null) {
                        if (!experiment.isReservationMade()) {
                            experiment.setSliceState(SliceState.RESERVING);
                        } else if (experiment.getRequestedStartTime().isAfter(Instant.now())) {
                            experiment.setSliceState(SliceState.FUTURE_RESERVATION);
                        } else {
                            experiment.setSliceState(SliceState.PROVISIONING);
                        }
                    } else {
                        experiment.setSliceState(SliceState.PROVISIONING);
                    }
                } else {
                    if (experiment.getRequestedStartTime() != null
                            && experiment.getRequestedStartTime().isAfter(Instant.now())) {
                        experiment.setSliceState(SliceState.FUTURE_RESERVATION);
                    } else {
                        experiment.setSliceState(SliceState.TESTING);
                    }
                }
                break;
            case RESERVING:
                createReservation();
                break;
            case FUTURE_RESERVATION:
                waitForReservationStart();
                break;
            case PROVISIONING:
                initializeSlivers();
                break;
            case TESTING:
                testSliversConnectivity();
                break;
            case READY:
                break;
            case FAILED:
                break;
            case RESERVATION_FAILED:
                break;
            case EXPIRING:
                checkForExpiration();
                break;

        }


    }

    private void checkForExpiration() {
        final CheckExpirationTaskGenerator checkExpirationTaskGenerator =
                experimentTasksFactory.createCheckExpirationTaskGenerator(this);

        checkExpirationTaskGenerator.addTaskFinishedCallback(status -> {
            if (experiment.getSliceState() != SliceState.EXPIRED) {

                //the experiment hasn't expired completely after completing our check. We need to schedule the expiration detector again.
                startSliceExpirationDetection();
            }

        });

        executorService.submit(checkExpirationTaskGenerator);
    }

    private void testSliversConnectivity() {
        executorService.submit(experimentTasksFactory.createTestConnectivityTaskGenerator(this));
    }

    public void createSlice() {
        //we must initiate the creation of the slice for this experiment
        assert (experiment.getRequestedEndTime() != null);

        submitExperimentTask(experimentTasksFactory.createRegisterExperimentTask(experiment));
    }

    private void restoreSlice() {
        executorService.submit(experimentTasksFactory.createRestoreSliversTaskGenerator(this));
    }

    private void createReservation() {
        executorService.submit(experimentTasksFactory.createReserveSliversTaskGenerator(this));
    }


    private void waitForReservationStart() {
        assert (experiment.getRequestedStartTime() != null);

        Duration durationToStartTime = Duration.between(Instant.now(), experiment.getRequestedStartTime());

        Runnable startProvisioning = () -> {
            experiment.setSliceState(SliceState.PROVISIONING);
        };

        if (!durationToStartTime.isNegative()) {


            scheduledExecutorService.schedule(startProvisioning, durationToStartTime.getSeconds() + 1, TimeUnit.SECONDS);
        } else {
            startProvisioning.run();
        }
    }


    private void startSliceExpirationDetection() {
        assert (experiment.getSlice() != null);

        expirationDetector.scheduleForNextExpiration();
    }


    private class ExpirationDetector implements Runnable {
        @Override
        public void run() {
            final Instant firstExpiration = findFirstExpiration();
            if (firstExpiration != null && firstExpiration.isBefore(Instant.now())) {
                LOG.trace("First expiration time {} has been exceeded.", firstExpiration);
                //(a part of) the experiment seems to have expired!
                experiment.setSliceState(SliceState.EXPIRING);
            } else {
                scheduleForNextExpiration();
            }
        }

        public void scheduleForNextExpiration() {
            final Instant firstExpiration = findFirstExpiration();

            if(firstExpiration != null) {
                Duration durationToFirstExpiration =
                        Duration.between(Instant.now(), firstExpiration);

                assert !durationToFirstExpiration.isNegative();

                LOG.trace("Scheduling next expiration check for {} in {} seconds", experiment.getName(), durationToFirstExpiration.getSeconds() + 1);

                scheduledExecutorService.schedule(this, durationToFirstExpiration.getSeconds() + 1, TimeUnit.SECONDS);
            }else{

                LOG.trace("Could not find an expiration time. Trying again in 10 seconds");
                scheduledExecutorService.schedule(this, 10, TimeUnit.SECONDS);
            }
        }

        private Instant findFirstExpiration() {

            Instant firstExpiration = experiment.getSlice().getExpirationDate();

            for (Sliver sliver : experiment.getSlice().getSlivers()) {
                if (firstExpiration == null || sliver.getExpirationDate() != null && sliver.getExpirationDate().isBefore(firstExpiration)) {
                    firstExpiration = sliver.getExpirationDate();
                }
            }
            return firstExpiration;
        }
    }

    private void initializeSlivers() {
        assert (experiment.getNewRequestRspecSource() != null);

        experiment.getSlice().setRequestRspec(experiment.getNewRequestRspecSource());

        if (experiment.getNewRequestRspecSource().getStringRspec().isStitching(authorityList.getAuthorityListModel(), authorityList.finder())) {
            executorService.submit(experimentTasksFactory.createInitializeSliversWithStitchTaskGenerator(this));
        } else {
            executorService.submit(experimentTasksFactory.createInitializeSliversWithoutStitchTaskGenerator(this));
        }
    }


    public void submitExperimentTask(ExperimentTask experimentTask) {
        assert (experimentTask.getExperiment() == experiment);

        LOG.trace("Received ExperimentTask {} for {}", experimentTask, experiment.getName());

        listeners.forEach(listener -> listener.onExperimentTaskSubmitted(experimentTask));

        executorService.submit(experimentTask);
    }

    public void submitExperimentTaskGenerator(ExperimentTaskGenerator experimentTaskGenerator) {
        assert (experimentTaskGenerator.getExperimentController() == this);
        LOG.trace("Received ExperimentTaskGenerator {} for {}", experimentTaskGenerator , experiment.getName());
        executorService.submit(experimentTaskGenerator);
    }

    public Experiment getExperiment() {
        return experiment;
    }

    public Window getParentWindow() {
        return parentWindow;
    }

    public void setParentWindow(Window parentWindow) {
        this.parentWindow = parentWindow;
    }

    public void addListener(ExperimentControllerListener listener) {
        this.listeners.add(listener);
    }

    public void removeListener(ExperimentControllerListener listener) {
        this.listeners.remove(listener);
    }

}
