package be.iminds.ilabt.jfed.experimenter_gui.slice.tasks;

import be.iminds.ilabt.jfed.connectivity_tester.ConnectivityTest;
import be.iminds.ilabt.jfed.experimenter_gui.slice.Experiment;
import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentController;
import be.iminds.ilabt.jfed.experimenter_gui.slice.SliceState;
import be.iminds.ilabt.jfed.experimenter_gui.ui.status.TaskStatusIndicator;
import be.iminds.ilabt.jfed.highlevel.controller.HighLevelController;
import be.iminds.ilabt.jfed.highlevel.controller.Task;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import javafx.application.Platform;
import javafx.scene.control.*;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collection;
import java.util.List;
import java.util.concurrent.CountDownLatch;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 2/3/2015
 * Time: 07:40
 */
public class InitializeSliversWithoutStitchingTaskGenerator extends InitializeSliversTaskGenerator {
    private static final Logger LOG = LoggerFactory.getLogger(InitializeSliversWithoutStitchingTaskGenerator.class);

    private final ExperimentTasksFactory experimentTasksFactory;
    private final JFedPreferences jFedPreferences;

    private CountDownLatch authoritiesReadyCountdownLatch;

    InitializeSliversWithoutStitchingTaskGenerator(ExperimentController experimentController, HighLevelController hlc,
                                                   ExperimentTasksFactory experimentTasksFactory, JFedPreferences jFedPreferences) {
        super(experimentController, hlc);
        this.experimentTasksFactory = experimentTasksFactory;
        this.jFedPreferences = jFedPreferences;
    }

    @Override
    public TaskStatusIndicator.Status generateExperimentTasks() {
        assert (experiment.getRequestedEndTime() != null);
        List<HighLevelController.CreateSliverTask> csInitTasks =
                hlc.createSliversNoStitching(experiment.getSlice(), experiment.getRequestedEndTime(), experiment.isIncludeProjectMemberSshKeys());

        HighLevelController.RenewSliceTask renewTask = createRenewSliceTask();
        if (renewTask != null) {
            for (HighLevelController.CreateSliverTask csInitTask : csInitTasks)
                csInitTask.addExtraDependency(renewTask);
        }

        authoritiesReadyCountdownLatch = new CountDownLatch(csInitTasks.size());
        for (final HighLevelController.CreateSliverTask initTask : csInitTasks) {
            experimentController.submitExperimentTask(new InitializeSliverTask(experiment, initTask, hlc));
        }

        try {
            authoritiesReadyCountdownLatch.await();

            if (experiment.getSliceState() == SliceState.PROVISIONING) {

                boolean sliverFailed =
                        experiment.getSlice().getSlivers().stream()
                                .filter(sliver -> sliver.getStatus().getGlobalStatus() == StatusDetails.SliverStatus.FAIL)
                                .findAny().isPresent();

                if (!sliverFailed) {
                    experiment.setSliceState(SliceState.READY);
                } else {
                    experiment.setSliceState(SliceState.FAILED);
                }


            } else
                LOG.warn("Cannot transfer to SliceState READY as current state is {} instead of PROVISIONING", experiment.getSliceState());
        } catch (InterruptedException ignored) {
        }

        return TaskStatusIndicator.Status.SUCCESS;

    }

    public final class InitializeSliverTask extends SingleCallExperimentTask {
        private final HighLevelController.CreateSliverTask createSliverTask;

        private final SfaAuthority authority;

        public InitializeSliverTask(Experiment experiment, HighLevelController.CreateSliverTask createSliverTask, HighLevelController hlc) {
            super(experiment, hlc);
            this.createSliverTask = createSliverTask;
            this.authority = createSliverTask.getSfaAuthority();

            updateTitle(String.format("Initialize nodes at %s", authority.getName()));
        }

        @Override
        public Task createTask() {
            return createSliverTask;
        }

        @Override
        public void onTaskFinished(TaskExecution taskExecution) {
            //first check if sliver creation was a success

            if (taskExecution.getState() != TaskExecution.TaskState.SUCCESS) {
                if (createSliverTask.hasFailedDueToNotEnoughFreeResources()) {
                    Platform.runLater(() -> {
                        updateMessage("Could not map to resources (resource shortage or unsupported topology))");
                        JFDialogs.create()
                                .owner(experimentController.getParentWindow())
                                .message(String.format("Initializing nodes at %s failed, because the request could not be mapped on the resources available on the testbed. " +
                                        "This is probably caused by a resource shortage, or because the requested topology is not supported", createSliverTask.getSfaAuthority().getName()))
                                .masthead("Initializing nodes failed")
                                .showError();
                    });
                }

                authoritiesReadyCountdownLatch.countDown();
                return;
            }

            StatusUntilReadyTask statusUntilReadyTask =
                    new StatusUntilReadyTask(experiment, createSliverTask.getGetStatusUntilReadyOrFailTask());

            statusUntilReadyTask.addTaskFinishedCallback(status -> {
                if (status == TaskStatusIndicator.Status.SUCCESS) {
                    tryStartConnectivityTest();

                } else {
                    Platform.runLater(() -> {
                        JFDialogs.create()
                                .owner(experimentController.getParentWindow())
                                .message(String.format(String.format("One or more resources on %s have permanently failed. " +
                                        "Consider terminating this experiment and trying again. " +
                                        "If the problem continues, please contact support to investigate the issue.", authority.getHrn())))
                                .masthead("Permanent failure on " + authority.getHrn())
                                .title("Permanent failure on " + authority.getHrn())
                                .showError();
                    });


                    authoritiesReadyCountdownLatch.countDown();
                }
            });

            //now wait for the sliver(s) of the authority to become ready
            experimentController.submitExperimentTask(statusUntilReadyTask);
        }

        private void tryStartConnectivityTest() {
            Collection<FXRspecNode> nodesToTest =
                    AuthorityConnectivityTestTask.getNodesToTest(experiment.getSlice(), authority);
            if (nodesToTest != null) {
                final AuthorityConnectivityTestTask authorityConnectivityTestTask =
                        experimentTasksFactory.createAuthorityConnectivityTestTask(experiment, authority, nodesToTest);
                authorityConnectivityTestTask.addTaskFinishedCallback(status -> {
                    authoritiesReadyCountdownLatch.countDown();

                    //check if we should show a warning about connectivity issues
                    final List<String> failedNodeNames =
                            authorityConnectivityTestTask.getNodesConnectivity().entrySet().stream()
                                    .filter(entry -> entry.getValue() == ConnectivityTest.Status.FAILED)
                                    .map(entry -> entry.getKey().getClientId())
                                    .collect(Collectors.toList());


                    Platform.runLater(() -> {
                        if (!failedNodeNames.isEmpty()) {
                            //show a (mutable) notification to mac-users that the terminal opens in the background
                            if (jFedPreferences.getBoolean(GuiPreferenceKey.PREF_SHOW_CONNECTIVITY_WARNING_DIALOG, true)) {
                                LOG.debug("Showing connectivity warning dialog");


                                VBox pane = new VBox();
                                pane.setPrefWidth(400);
                                pane.setPrefHeight(250);
                                pane.setSpacing(5);

                                Label textLabel =
                                        new Label(String.format("jFed detected that node(s) %s cannot be reached. " +
                                                "You can take the following steps to resolve the problem:\n\n" +
                                                "1. Try enabling the proxy\n" +
                                                "2. Check your network settings. Are you behind a firewall? Do you have IPv6-connectivity?\n" +
                                                "3. Check the result of the API-calls\n" +
                                                "4. Check if there is an error in the requested experiment\n" +
                                                "5. Check if the requested resources are available\n" +
                                                "6. Contact First Level Support", failedNodeNames.stream().collect(Collectors.joining(", "))));
                                textLabel.setPrefHeight(180); //work around bug FEDIBBTDEV-1008
                                textLabel.setWrapText(true);
                                pane.getChildren().add(textLabel);
                                final CheckBox checkBox = new CheckBox("Do not show this dialog again");
                                pane.getChildren().add(checkBox);

                                DialogPane dialogPane = new DialogPane();
                                dialogPane.setHeaderText("Experiment connectivity issues");
                                dialogPane.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.PLAY, Color.GREEN));
                                dialogPane.setContent(pane);
                                dialogPane.getButtonTypes().setAll(ButtonType.OK);


                                Dialog<ButtonType> dialog = new Dialog<>();
                                dialog.setTitle("Experiment connectivity issues");
                                dialog.setDialogPane(dialogPane);

                                dialog.showAndWait();

                                jFedPreferences.setBoolean(GuiPreferenceKey.PREF_SHOW_CONNECTIVITY_WARNING_DIALOG,
                                        !checkBox.isSelected());
                            } else {
                                LOG.debug("Skipping connectivity issues dialog because user muted it before");
                            }
                        }
                    });

                });

                experimentController.submitExperimentTask(authorityConnectivityTestTask);
            } else {
                authoritiesReadyCountdownLatch.countDown();
            }
        }


    }
}
