package be.iminds.ilabt.jfed.experimenter_gui.util;

import be.iminds.ilabt.jfed.experimenter_gui.preferences.PreferencesDialogFactory;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.rspec.model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.ssh_terminal_tool.ExternalSshTerminal;
import be.iminds.ilabt.jfed.ssh_terminal_tool.ssh_key_info.UserSshKeyInfo;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import com.google.common.base.Preconditions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.inject.Singleton;
import java.util.List;

/**
 * User: twalcari
 * Date: 1/29/2015
 * Time: 16:13
 */
@Singleton
public class SSHTerminalUtil {
    private static final Logger LOG = LoggerFactory.getLogger(SSHTerminalUtil.class);


    private final ExternalSshTerminal externalSshTerminal;
    private final PreferencesDialogFactory preferencesDialogFactory;
    private final GeniUserProvider geniUserProvider;


    @Inject
    public SSHTerminalUtil(ExternalSshTerminal externalSshTerminal, PreferencesDialogFactory preferencesDialogFactory, GeniUserProvider geniUserProvider) {
        this.externalSshTerminal = externalSshTerminal;
        this.preferencesDialogFactory = preferencesDialogFactory;
        this.geniUserProvider = geniUserProvider;
    }

    public void launchSSHTerminalForcedLocalUser(FXRspecNode rspecNode, ManifestRspecSource manifestRspecSource) {
        String username =
                geniUserProvider.getLoggedInGeniUser().getUserUrn().getEncodedResourceName();
        launchSSHTerminal(rspecNode, manifestRspecSource, username);
    }


    public void launchSSHTerminalForcedLocalUser(FXRspecNode rspecNode, Slice slice) {
        String username =
                geniUserProvider.getLoggedInGeniUser().getUserUrn().getEncodedResourceName();
        launchSSHTerminal(rspecNode, slice, username);
    }

    public void launchSSHTerminal(FXRspecNode rspecNode, Slice slice) {
        launchSSHTerminal(rspecNode, slice, null);
    }

    public void launchSSHTerminal(@Nonnull FXRspecNode rspecNode, @Nullable Slice slice, @Nullable String overrideUserName){
        //get the associated manifest RSpec-node
        if (slice == null) {
            LOG.warn("Could not open SSH terminal: manifestRspecSource=null");
            JFDialogs.create()
                    .owner(this)
                    .message("There is currently insufficient information available to login.")
                    .showError();
            return;
        }
        else{
            launchSSHTerminal(rspecNode, slice.getManifestRspec(), overrideUserName);
        }
    }



    public void launchSSHTerminal(FXRspecNode rspecNode, ManifestRspecSource manifestRspecSource) {
        launchSSHTerminal(rspecNode, manifestRspecSource, null);
    }

    /**
     * The rspecNode is NOT necessarily a part of the provided manifestRspecSource. It can be that the rspecNode is from the RequestRspec.
     * <p/>
     * We first have to match the rspecNode to the corresponding node in the manifstRspecSource!
     *
     * @param manifestRspecSource
     * @param rspecNode
     * @param overrideUsername
     */
    public void launchSSHTerminal(@Nonnull FXRspecNode rspecNode, @Nullable ManifestRspecSource manifestRspecSource, @Nullable String overrideUsername) {


        //get the associated manifest RSpec-node
        if (manifestRspecSource == null) {
            LOG.warn("Could not open SSH terminal: manifestRspecSource=null");
            JFDialogs.create()
                    .owner(this)
                    .message("There is currently insufficient information available to login.")
                    .showError();
            return;
        }

        final List<BasicStringRspec.LoginService> loginServices =
                manifestRspecSource.getNodeLoginInfo(rspecNode);


        if (loginServices == null || loginServices.isEmpty()) {
            JFDialogs.create()
                    .owner(this)
                    .message("The selected node doesn't provide any information to perform an interactive login.")
                    .masthead("Could not open SSH terminal")
                    .title("Could not open SSH terminal")
                    .showException(new RuntimeException(
                            "node id=\"" + rspecNode.getUniqueId() + "\" "
                                    + manifestRspecSource.getStringRspec().getXmlString()
                    ));

            return;
        }

        //try to find a loginService with a login-name equal to that of the logged-in user
        GeniUser geniUser = geniUserProvider.getLoggedInGeniUser();

        BasicStringRspec.LoginService chosenLoginService = loginServices.get(0);
        for (int i = 1; i < loginServices.size(); i++) {
            BasicStringRspec.LoginService otherLoginService = loginServices.get(i);
            if (otherLoginService.getUsername().equals(geniUser.getUserUrn().getEncodedResourceName()))
                chosenLoginService = otherLoginService;
        }

        launchSSHTerminal(chosenLoginService, overrideUsername);
    }

    public void launchSSHTerminal(BasicStringRspec.LoginService loginService) {
        launchSSHTerminal(loginService, null);
    }


    public void launchSSHTerminal(BasicStringRspec.LoginService loginService, String overrideUsername) {
        Preconditions.checkNotNull(loginService);

        if (!externalSshTerminal.isConfigured()) {
            JFDialogs.create()
                    .owner(this)
                    .message("jFed must be configured with the location of the terminal application and a valid Private Key.")
                    .showWarning();
            preferencesDialogFactory.showDialog();
        }
        if (!externalSshTerminal.isConfigured()) {
            JFDialogs.create()
                    .owner(this)
                    .message("jFed was not properly configured. Please configure jFed first, and try again.")
                    .showError();
            return;
        }

        GeniUser geniUser = geniUserProvider.getLoggedInGeniUser();


        UserSshKeyInfo userSshKeyInfo = new UserSshKeyInfo(geniUser);

        externalSshTerminal.launch(overrideUsername != null ? overrideUsername : loginService.getUsername(),
                loginService.getHostname(),
                loginService.getPort(),

                //info on the authority proxy. Note that this might not be used, if the preferences specify another proxy!
                geniUser.getUserAuthority().getProxies(),
                geniUser/* user */,
                userSshKeyInfo/* this is always the SSH key from log file. Even if other key is used, this is the info needed. The externalSshterminal will get key from preferences if needed. */);
    }

}
