package be.iminds.ilabt.jfed.highlevel.controller;

import be.iminds.ilabt.jfed.highlevel.history.StitchingReportHistory;
import be.iminds.ilabt.jfed.highlevel.model.*;
import be.iminds.ilabt.jfed.highlevel.stitcher.ParallelStitcher;
import be.iminds.ilabt.jfed.highlevel.stitcher.StitchingReportFactory;
import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.log.ResultListener;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.api.AbstractFederationApi;
import be.iminds.ilabt.jfed.lowlevel.api.FederationRegistryApi1;
import be.iminds.ilabt.jfed.lowlevel.api.FederationSliceAuthorityApi2;
import be.iminds.ilabt.jfed.lowlevel.api.OCCI;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.UserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticAggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticUserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.*;
import be.iminds.ilabt.jfed.lowlevel.fedmon.Fedmon;
import be.iminds.ilabt.jfed.lowlevel.fedmon.TestResult;
import be.iminds.ilabt.jfed.preferences.CorePreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecFactory;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.ManifestRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.JFedUtils;
import be.iminds.ilabt.jfed.util.JavaFXLogger;
import javafx.application.Platform;
import javafx.util.Pair;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.inject.Singleton;
import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.time.Instant;
import java.util.*;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

/**
 * HighLevelController
 * <p/>
 * The HighLevelController provides access to various typical gui tasks with simple calls.
 * It will select API's as needed.
 * Dependant calls will be performed if needed, and post completion, timeout or failure operations will be run.
 * <p/>
 * It also stores the connection pool.
 */
@Singleton
public class HighLevelController {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(HighLevelController.class);

    private final EasyModel easyModel;
    private final AuthorityList authorityList;
    private final JFedConnectionProvider connectionProvider;
    private final GeniUserProvider geniUserProvider;
    private final TaskThread tt;

    private final Map<Slice, Task> sliceCredentialCalls = new HashMap<>();
    private final Map<Slice, Task> sliceManifestCalls = new HashMap<>();
    private final Map<Sliver, Task> sliverManifestCalls = new HashMap<>();
    private final Map<Sliver, Task> sliverStatusCalls = new HashMap<>();
    //global call                 name
    //call per slice              name + slice
    //call per slice + auth pair  name + slice + auth
    //===> unique name?
    private final Map<Pair<Slice, SfaAuthority>, SliceStatusTask> sliceStatusOnAuthCalls =
            new HashMap<>();
    private final Map<Pair<Slice, SfaAuthority>, GetSliceManifestTask> sliceManifestOnAuthCalls =
            new HashMap<>();
    private final Map<Sliver, Task> deleteSliverCalls = new HashMap<>();
    private final Map<Slice, GetAggregatesForSliceTask> getAggregatesForSliceCalls = new HashMap<>();
    private final Map<SfaAuthority, Task> getVersionCalls = new HashMap<>();

    private final Map<SfaAuthority, GetAdvertisementTask> availableAdvertisementCalls = new HashMap<>();
    private final Map<SfaAuthority, GetAdvertisementTask> allAdvertisementCalls = new HashMap<>();

    private final Map<Slice, Task> sliceStatusCalls = new HashMap<>();
    private Task userSSHKeyCall = null;
    private Map<String, Task> projectMemberSSHKeysCallByProjectName = new HashMap<>();
    private Map<GeniUrn, SliceMembersSshKeysTask> sliceMemberSSHKeysCallBySliceUrnName = new HashMap<>();
    private Task userCredentialCall = null;
    private ListSlicesTask getSlicesForUserCall = null;
    private ListSubAuthoritiesTask findSubAuthCall = null;
    private Task resolveUserCallAndSlices = null;
    private AutomaticUserAndSliceApiWrapper wrapper;
    private final JavaFXLogger logger;
    private final StitchingReportHistory stitchingReportHistory;
    private final StitchingReportFactory stitchingReportFactory;

    private final JFedPreferences jFedPreferences;
    private final AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory
            automaticUserAndSliceApiWrapperFactory;
    private final AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory
            automaticAggregateManagerWrapperFactory;

    private final RspecFactory rspecFactory = FXRspecFactory.getInstance();

    @Inject
    public HighLevelController(EasyModel easyModel, AuthorityList authorityList, JFedConnectionProvider connectionProvider,
                               GeniUserProvider geniUserProvider, JavaFXLogger logger, TaskThread tt,
                               AutomaticUserAndSliceApiWrapper.AutomaticUserAndSliceApiWrapperFactory automaticUserAndSliceApiWrapperFactory,
                               JFedPreferences jFedPreferences, AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory automaticAggregateManagerWrapperFactory,
                               StitchingReportHistory stitchingReportHistory, StitchingReportFactory stitchingReportFactory) {
        assert easyModel != null;
        assert authorityList != null;
        assert connectionProvider != null;
        assert geniUserProvider != null;
        assert logger != null;
        assert tt != null;
        assert jFedPreferences != null;
        assert stitchingReportHistory != null;
        assert stitchingReportFactory != null;

        this.easyModel = easyModel;
        this.authorityList = authorityList;
        this.connectionProvider = connectionProvider;
        this.geniUserProvider = geniUserProvider;
        this.logger = logger;
        this.tt = tt;
        this.jFedPreferences = jFedPreferences;
        this.automaticUserAndSliceApiWrapperFactory = automaticUserAndSliceApiWrapperFactory;
        this.automaticAggregateManagerWrapperFactory = automaticAggregateManagerWrapperFactory;
        this.stitchingReportHistory = stitchingReportHistory;
        this.stitchingReportFactory = stitchingReportFactory;
    }

    /**
     * UserAndSliceApiWrapper stores state! (user credentials)
     * <p/>
     * So it needs to be created only once.
     * <p/>
     * this is public, anyone who wants to use this directly is welcome.
     */
    public synchronized UserAndSliceApiWrapper getUserAndSliceApiWrapper() {
        if (wrapper == null)
            wrapper = automaticUserAndSliceApiWrapperFactory.create();

        return wrapper;
    }

    /**
     * this is public, anyone who wants to use this directly is welcome.
     *
     * @param amAuth        the Aggregate Manager to connect to
     * @param taskExecution the TaskExecution object that needs to receive copies of all calls. May be null.
     */
    public AggregateManagerWrapper getAggregateManagerWrapper(TaskExecution taskExecution, SfaAuthority amAuth) {
        assert jFedPreferences != null;
        return automaticAggregateManagerWrapperFactory.create(
                taskExecution == null ? logger : JavaFXLogger.wrap(logger, taskExecution),
                amAuth);
    }

    public Task getSliceCredential(@Nonnull final Slice slice) {

        Task call = sliceCredentialCalls.get(slice);
        if (call != null) return call;

        call = new Task("Get Slice Credential " + slice.getUrnString()) {
            @Override
            public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
                UserAndSliceApiWrapper w = getUserAndSliceApiWrapper();

                w.getSliceCredentials(JavaFXLogger.wrap(logger, taskExecution), GeniUrn.parse(slice.getUrnString()));

                //assert slice.hasCredential() : "resulting credential was NOT registered with EasyModel!";
                if (!slice.hasCredential())
                    throw new RuntimeException("Error processing reply: No valid slice credential was found");

                if (slice.isExpired())
                    throw new RuntimeException("Error processing reply: got credential for expired slice");
            }

            @Override
            public List<Task> initDependsOn() {
                List<Task> res = new ArrayList<>();

                LOG.debug("getSliceCredential Task#initDependsOn -> wrapper.needSubAuthCredential()=" +
                        "  user auth urn=" + geniUserProvider.getLoggedInGeniUser().getUserAuthority().getUrnString());
                res.add(getUserCredential());
                return res;
            }


//            @Override
//            public List<Task> initAlwaysDependsOn() {
//                List<Task> res = new ArrayList<Task>();
//
//                //if slice is expired, not bad to resolve it first (call might fail if it doesn't exist)
//                if (slice.isExpired())
//                    res.add(resolveSlice(slice));
//
//                return res;
//            }

            @Override
            public boolean newExecutionNeeded() {
                if (slice.isExpired()) {
                    LOG.warn("Slice " + slice.getUrnString() + " has expired, so a new execution of Get Slice Credential is needed.");
                    return true;
                }

                //TODO also check whether slice credential hasn't expired  (somewhat pointless: slice credential expires is same as slice expire...)

                return !slice.hasCredential();
            }
        };
        sliceCredentialCalls.put(slice, call);
        return call;
    }

    public Task getManifest(@Nonnull final Slice slice) {
        Task call = sliceManifestCalls.get(slice);
        if (call != null) return call;

        call = new Task("Get Slice Manifest") {
            @Override
            public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
                LOG.debug("getManifest cm count: " + slice.manifestComponentManagersProperty().size());
                for (SfaAuthority auth : slice.getManifestComponentManagers())
                    if (auth.hasAmUrl()) {
                        //TODO: see getStatus, here as well we need to block ourself and add these as deps instead of just adding them
                        Task subCall = getSliceManifest(slice, auth);
                        tt.addTask(subCall);
                    } else
                        LOG.error("Tried to get manifest for slice \"" + slice.getUrnString() + "\", at authority \"" + auth.getUrnString() + "\", but that authority has no AM URL: will ignore slice on this authority.");
            }

            @Override
            public List<Task> initDependsOn() {
                List<Task> res = new ArrayList<>();
                res.add(getSliceCredential(slice));
                return res;
            }

            @Override
            public List<Task> initAlwaysDependsOn() {
                List<Task> res = new ArrayList<>();
                res.add(getAggregatesForSlice(slice));
                return res;
            }
        };
        sliceManifestCalls.put(slice, call);
        return call;
    }

    public Task getManifest(@Nonnull final Sliver sliver) {
        Task call = sliverManifestCalls.get(sliver);
        if (call != null) return call;

        SfaAuthority auth = sliver.getAuthority();
        if (!auth.hasAmUrl()) {
            LOG.error("Tried to get manifest for sliver \"" + sliver.getUrnString() + "\", at authority \"" + auth.getUrnString() + "\", but that authority has no AM URL: cannot continue.");
            return null;
        }

        call = new Task("Get Sliver Manifest") {
            @Override
            public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
                SfaAuthority auth = sliver.getAuthority();
                assert auth != null;

//                if (auth.getUrl(AuthorityType.GeniServerRole.AM, 2) != null) {
                AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(taskExecution, auth);
                amWrapper.describe(sliver.getSlice().getUrn(), sliver.getSlice().getCredentials());
//                    EasyAggregateManager2 am2 =
//                            new EasyAggregateManager2(JavaFXLogger.wrap(logger, taskExecution), easyModel, auth);
//                    am2.listSliceResources(sliver.getSlice());
//                }
            }

            @Override
            public List<Task> initDependsOn() {
                List<Task> res = new ArrayList<>();
                res.add(getSliceCredential(sliver.getSlice()));
                return res;
            }
        };
        sliverManifestCalls.put(sliver, call);
        return call;
    }

    public Task getSliverStatus(@Nonnull final Sliver sliver) {
        Task call = sliverStatusCalls.get(sliver);
        if (call != null) return call;

        SfaAuthority auth = sliver.getAuthority();
        if (!auth.hasAmUrl()) {
            LOG.error("Tried to get Status for sliver \"" + sliver.getUrnString() + "\", at authority \"" + auth.getUrnString() + "\", but that authority has no AM URL: cannot continue.");
            return null;
        }

        call = new Task("Get Sliver Status @ " + sliver.getAuthority().getUrnString()) {
            @Override
            public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
                SfaAuthority auth = sliver.getAuthority();
                assert auth != null;

//                if (auth.getUrl(AuthorityType.GeniServerRole.AM, 2) != null) {
                AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(taskExecution, auth);
                amWrapper.status(sliver.getSlice().getUrn(), sliver.getSlice().getCredentials());

//                    EasyAggregateManager2 am2 =
//                            new EasyAggregateManager2(JavaFXLogger.wrap(logger, taskExecution), easyModel, auth);
//                    am2.sliverStatus(sliver.getSlice());
//                }
            }

            @Override
            public List<Task> initDependsOn() {
                List<Task> res = new ArrayList<>();
                res.add(getSliceCredential(sliver.getSlice()));
                return res;
            }
        };
        sliverStatusCalls.put(sliver, call);
        return call;
    }

    /**
     * get a Status for a specific AM. We don't assume there is a sliver, as getSliverManifest does
     */
    public Task getSliceStatus(@Nonnull Slice slice, @Nonnull SfaAuthority auth) {
        Pair<Slice, SfaAuthority> pair = new Pair<>(slice, auth);
        Task call = sliceStatusOnAuthCalls.get(pair);
        if (call != null) return call;

        if (!auth.hasAmUrl()) {
            LOG.error("Tried to get Status for slice \"" + slice.getUrnString() + "\", at authority \"" + auth.getUrnString() + "\", but that authority has no AM URL: cannot continue.");
            return null;
        }

        SliceStatusTask newCall = new SliceStatusTask(slice, auth);
        sliceStatusOnAuthCalls.put(pair, newCall);
        return newCall;
    }

    /**
     * get a manifest for a specific AM. We don't assume there is a sliver, as getSliverManifest does
     */
    public GetSliceManifestTask getSliceManifest(@Nonnull final Slice slice, @Nonnull final SfaAuthority auth) {
        Pair<Slice, SfaAuthority> pair = new Pair<>(slice, auth);
        GetSliceManifestTask call = sliceManifestOnAuthCalls.get(pair);
        if (call != null) return call;

        if (!auth.hasAmUrl()) {
            LOG.error("Tried to get manifest for slice \"" + slice.getUrnString() + "\", at authority \"" + auth.getUrnString() + "\", but that authority has no AM URL: cannot continue.");
            return null;
        }

        GetSliceManifestTask newCall = new GetSliceManifestTask(auth, slice);
        sliceManifestOnAuthCalls.put(pair, newCall);
        return newCall;
    }

    public class GetSliceManifestTask extends Task {
        private final SfaAuthority auth;
        private final Slice slice;
        private String manifestRspec = null;
        private StringRspec manifestStringRspec = null;

        public GetSliceManifestTask(SfaAuthority auth, Slice slice) {
            super("get Manifest of Slice @ " + auth.getName());
            this.auth = auth;
            this.slice = slice;
        }

        @Override
        public void doTask(TaskExecution singleTask) throws JFedException, InterruptedException {
//                if (auth.getUrl(AuthorityType.GeniServerRole.AM, 2) != null) {
            AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(singleTask, auth);
            manifestRspec = amWrapper.describe(slice.getUrn(), slice.getCredentials());
        }

        @Override
        public List<Task> initDependsOn() {
            List<Task> res = new ArrayList<>();
            res.add(getSliceCredential(slice));
            return res;
        }

        public String getManifestRspec() {
            return manifestRspec;
        }

        public StringRspec getManifestStringRspec() {
            if (manifestRspec == null)
                return null;
            if (manifestStringRspec == null)
                manifestStringRspec = new StringRspec(manifestRspec);
            return manifestStringRspec;
        }
    }

    public Task deleteSliver(@Nonnull final Sliver sliver) {
        Task call = deleteSliverCalls.get(sliver);
        if (call != null) return call;

        SfaAuthority auth = sliver.getAuthority();
        if (!auth.hasAmUrl()) {
            LOG.error("Tried to delete sliver \"" + sliver.getUrnString() + "\", but the Component Manager for that sliver, \"" + auth.getUrnString() + "\" has no AM URL: cannot delete!");
            return null;
        }

        call = new Task("Delete Sliver " + sliver.getUrn()) {
            @Override
            public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
                SfaAuthority auth = sliver.getAuthority();
                assert auth != null;

                if (auth.isFake())
                    return;

                Slice slice = sliver.getSlice();
                AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(taskExecution, auth);
                amWrapper.deleteSliver(slice.getUrn(), slice.getCredentials());

                List<GeniUrn> sliverUrns = new ArrayList<>();
                if (sliver.getUrn() != null)
                    sliverUrns.add(sliver.getUrn());
                if (!sliverUrns.isEmpty())
                    unregisterSliversAtSATaskExecution(slice, auth, sliverUrns, this);

                //old method: do the calls here instead of in task depending on this one
//                UserAndSliceApiWrapper samaWrapper = getUserAndSliceApiWrapper();
//                if (samaWrapper.isRegisterAggregatesForSliceSupported()) {
//                    List<GeniUrn> sliverUrns = new ArrayList<>();
//
//                    if (sliver.getUrn() != null) {
//                        sliverUrns.add(sliver.getUrn());
//                    }
//                    if (!sliverUrns.isEmpty()) {
//                        Logger logger = JavaFXLogger.wrap(logger, taskExecution);
//                        samaWrapper.unregisterAggregatesForSlice(logger,
//                                slice.getCredentials(), slice.getUrn(),
//                                auth.getUrn(), sliverUrns);
//                    }
//                }
            }

            @Override
            public List<Task> initDependsOn() {
                List<Task> res = new ArrayList<>();
                res.add(getSliceCredential(sliver.getSlice()));
                return res;
            }
        };
        deleteSliverCalls.put(sliver, call);
        return call;
    }

    public GetAggregatesForSliceTask getAggregatesForSlice(@Nonnull final Slice slice) {
        GetAggregatesForSliceTask call = getAggregatesForSliceCalls.get(slice);
        if (call == null) {
            call = new GetAggregatesForSliceTask(slice);
            getAggregatesForSliceCalls.put(slice, call);

        }
        return call;
    }

    public Task getUserSSHKey() {
        if (userSSHKeyCall != null) return userSSHKeyCall;

        userSSHKeyCall = new Task("Get User SSH Keys") {
            @Override
            public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
                UserAndSliceApiWrapper w = getUserAndSliceApiWrapper();
                w.getSshKeysForUser(JavaFXLogger.wrap(logger, taskExecution), GeniUrn.parse(geniUserProvider.getLoggedInGeniUser().getUserUrnString()));
//                EasySliceAuthority easySliceAuthority = new EasySliceAuthority(JavaFXLogger.wrap(logger, taskExecution), easyModel, auth);
//                easySliceAuthority.getSshKeys();

            }
            // @Override public boolean needed() { return easyModel.getUserKeys() == null || easyModel.getUserKeys().isEmpty(); };

            @Override
            public List<Task> initDependsOn() {
                List<Task> res = new ArrayList<>();
                res.add(getUserCredential());
                return res;
            }
        };
        return userSSHKeyCall;
    }

    private final static String projectNameForGlobalProject = "projectNameForGlobalProjectInsteadOfNull";

    public Task getProjectMemberSSHKeys(final String projectName) {
        Task task = projectMemberSSHKeysCallByProjectName.get(
                projectName == null ? projectNameForGlobalProject : projectName);
        if (task != null) return task;

        task = new Task(projectName == null ? "Get all User SSH Keys"
                : "Get User SSH Keys For users in project " + projectName) {
            @Override
            public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
                UserAndSliceApiWrapper w = getUserAndSliceApiWrapper();
                List<GeniUrn> userUrns =
                        w.getUsersForSubAuthority(JavaFXLogger.wrap(logger, taskExecution), projectName);
                if (userUrns == null) userUrns = Collections.emptyList();
                for (GeniUrn userUrn : userUrns) {
                    try {
                        w.getSshKeysForUser(JavaFXLogger.wrap(logger, taskExecution), userUrn);
                    } catch (JFedException e) {
                        LOG.warn("Could not get user key for " + userUrn, e);
                    }
                }

            }
            // @Override public boolean needed() { return easyModel.getUserKeys() == null || easyModel.getUserKeys().isEmpty(); };

            @Override
            public List<Task> initDependsOn() {
                List<Task> res = new ArrayList<>();
                res.add(getUserCredential());
                return res;
            }
        };
        projectMemberSSHKeysCallByProjectName.put(
                projectName == null ? projectNameForGlobalProject : projectName,
                task);
        return task;
    }

    public SliceMembersSshKeysTask getSliceMemberSSHKeys(final Slice slice) {
        assert slice != null;
        final GeniUrn sliceUrn = slice.getUrn();
        assert sliceUrn != null;

        SliceMembersSshKeysTask task = sliceMemberSSHKeysCallBySliceUrnName.get(sliceUrn);
        if (task != null) return task;

        task = new SliceMembersSshKeysTask(slice);
        sliceMemberSSHKeysCallBySliceUrnName.put(sliceUrn, task);
        return task;
    }

    public void scheduleTask(Task task, int initialWait) {
        tt.scheduleTask(task, initialWait);
    }

    public class GetOtherUserSshKeysTask extends Task {
        private final GeniUrn userUrn;
        private List<String> keys;
        private List<Task> extraTaskDeps;

        public GetOtherUserSshKeysTask(@Nonnull GeniUrn userUrn, @Nullable List<Task> extraTaskDeps) {
            super("Get SSH keys for " + userUrn);
            this.userUrn = userUrn;
            this.extraTaskDeps = extraTaskDeps;
        }

        @Override
        public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
            UserAndSliceApiWrapper w = getUserAndSliceApiWrapper();
            keys = w.getSshKeysForUser(JavaFXLogger.wrap(logger, taskExecution), userUrn);
        }

        public List<String> getKeys() {
            return keys;
        }

        @Override
        public List<Task> initDependsOn() {
            List<Task> res = new ArrayList<>();
            res.add(getUserCredential());
            if (extraTaskDeps != null)
                res.addAll(extraTaskDeps);
            return res;
        }
    }

    private Map<GeniUrn, GetOtherUserSshKeysTask> getOtherUserSSHKeyTasksByUser = new HashMap<>();

    public GetOtherUserSshKeysTask getOtherUserSSHKey(@Nonnull GeniUrn userUrn) {
        return getOtherUserSSHKey(userUrn, (List<Task>) null);
    }

    public GetOtherUserSshKeysTask getOtherUserSSHKey(@Nonnull GeniUrn userUrn, @Nullable Task extraTaskDep) {
        List<Task> extraTaskDeps = new ArrayList<>();
        if (extraTaskDep != null)
            extraTaskDeps.add(extraTaskDep);
        return getOtherUserSSHKey(userUrn, extraTaskDeps);
    }

    public GetOtherUserSshKeysTask getOtherUserSSHKey(@Nonnull GeniUrn userUrn, @Nullable List<Task> extraTaskDeps) {
        GetOtherUserSshKeysTask res = getOtherUserSSHKeyTasksByUser.get(userUrn);
        if (res == null)
            res = new GetOtherUserSshKeysTask(userUrn, extraTaskDeps);
        return res;
    }

    public synchronized Task getUserCredential() {
        if (userCredentialCall != null) return userCredentialCall;

        this.userCredentialCall = new Task("Get User Credential") {
            @Override
            public void doTask(TaskExecution singleTask) throws JFedException, InterruptedException {
                LOG.debug("Get User Credential started");

                UserAndSliceApiWrapper w = getUserAndSliceApiWrapper();

                List<AnyCredential> res =
                        w.getLoggedinUserCredentials(JavaFXLogger.wrap(logger, singleTask));

                try {
                    List<Date> expireDatesRecv =
                            res.stream().map(AnyCredential::getExpiresDate).collect(Collectors.toList());
                    List<Date> expireDatesStored =
                            w.getCachedUserCredentialsForAM().stream().map(AnyCredential::getExpiresDate).collect(Collectors.toList());
                    boolean hasUserCredentials = w.hasUserCredentials();

                    LOG.debug("GetCredential call finished. Expire dates of received creds: " + expireDatesRecv + ". Expire dates of stored credentials: " + expireDatesStored + "  hasUserCredentials=" + hasUserCredentials);
                } catch (Throwable e) {
                    LOG.warn("Exception while creating log output", e);
                }

                assert res != null;
                assert !res.isEmpty();
            }

            @Override
            public boolean newExecutionNeeded() {
                UserAndSliceApiWrapper w = getUserAndSliceApiWrapper();
                boolean hasUserCredentials = w.hasUserCredentials();
                try {
                    if (w.getCachedUserCredentialsForAM() != null) {
                        List<Date> expireDates =
                                w.getCachedUserCredentialsForAM().stream().map(AnyCredential::getExpiresDate).collect(Collectors.toList());
                        LOG.info("getUserCredential newExecutionNeeded() called -> hasUserCredentials=" + hasUserCredentials + " now=" + new Date() + " expireDates=" + expireDates);
                    } else {
                        LOG.info("getUserCredential newExecutionNeeded() called -> hasUserCredentials=" + hasUserCredentials + " now=" + new Date() + " expireDates=null");

                    }
                } catch (Throwable t) {
                    LOG.warn("Exception in logging ignored. hasUserCredentials=" + hasUserCredentials, t);
                }
                return !hasUserCredentials;
            }

            @Override
            public List<Task> initDependsOn() {
                List<Task> res = new ArrayList<>();
                return res;
            }
        };
        return userCredentialCall;
    }

    public ListSlicesTask getSlicesForUser() {
        if (getSlicesForUserCall != null) return getSlicesForUserCall;

        getSlicesForUserCall = new ListSlicesTask();
        return getSlicesForUserCall;
    }


    public ListSubAuthoritiesTask findSubAuthorities() {
        if (findSubAuthCall != null) return findSubAuthCall;
        findSubAuthCall = new ListSubAuthoritiesTask();
        return findSubAuthCall;
    }

    public Task resolveUserAndSlices() {
        if (resolveUserCallAndSlices != null) return resolveUserCallAndSlices;

        resolveUserCallAndSlices = new Task("Resolve User & Slices") {
            @Override
            public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
                UserAndSliceApiWrapper w = getUserAndSliceApiWrapper();

                List<GeniUrn> sliceUrns =
                        w.getSlicesForUser(JavaFXLogger.wrap(logger, taskExecution), GeniUrn.parse(geniUserProvider.getLoggedInGeniUser().getUserUrnString()));
//                EasySliceAuthority easySliceAuthority = new EasySliceAuthority(JavaFXLogger.wrap(logger, taskExecution), easyModel, auth);
//                UserInfo res = easySliceAuthority.resolveUser(geniUserProvider.getLoggedInGeniUser().getUserUrn());

                //note: resolveUser call result might not have triggered listener yet, so easymodel might not yet know of it!

                for (final GeniUrn sliceUrn : sliceUrns) {
                    //TODO very dirty, fix! (see TODOs below for info)
                    //TODO how to wait for result to be processed so we can request slice resolve? One way is to run this in JavaFX thread,
                    //TODO     it is a dirty method, as it relies on much implementation info.
                    //TODO     a better method would be to have a method that waits for all pending updates in EasyModel or the EasyModelListeners.
                    Platform.runLater(new Runnable() {
                        @Override
                        public void run() {
                            Slice slice = easyModel.getSlice(sliceUrn.toString());
                            assert slice != null;
                            Task resolveSliceCall = getAggregatesForSlice(slice);

                            //TODO reblock self and add these as dependencies instead
                            //Note: this to do requires that the other is done first
                            tt.addTask(resolveSliceCall);
                        }
                    });
                }
            }

            @Override
            public List<Task> initDependsOn() {
                List<Task> res = new ArrayList<>();
                res.add(getUserCredential());
                return res;
            }
        };
        return resolveUserCallAndSlices;
    }

    public Task getVersion(@Nonnull final SfaAuthority auth) {
        Task call = getVersionCalls.get(auth);
        if (call != null) return call;

        call = new Task("GetVersion on " + auth.getName()) {
            @Override
            public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
                AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(taskExecution, auth);
                amWrapper.getVersion();
//                EasyAggregateManager2 am2 =
//                        new EasyAggregateManager2(JavaFXLogger.wrap(logger, taskExecution), easyModel, auth);
//                am2.getVersion();
            }

            @Override
            public boolean newExecutionNeeded() {
                //TODO: Store cached GetVersion result?
//                AuthorityInfo ai = easyModel.getAuthorityList().get(auth);
//                if (ai.hetCachehedGetVersionResult() != null) return false;
                return super.newExecutionNeeded();
            }

        };
        getVersionCalls.put(auth, call);
        return call;
    }

    public Task emulabRegistry(@Nonnull final SfaAuthority auth) {
        Task call = new Task("Emulab Registry " + auth.getName()) {
            @Override
            public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
                FederationRegistryApi1 registryApi =
                        new FederationRegistryApi1(JavaFXLogger.wrap(logger, taskExecution), jFedPreferences);

                GeniUser user = geniUserProvider.getLoggedInGeniUser();
                SfaConnection con = (SfaConnection)
                        connectionProvider.getConnectionByAuthority(user, auth, new ServerType(ServerType.GeniServerRole.GENI_CH, 1));
                if (con == null)
                    throw new RuntimeException("No connection can be created. Perhaps the authority \"" + auth.getUrnString() + "\" does not have a federation registry?");

                AbstractFederationApi.FederationApiReply<List<FederationRegistryApi1.ServiceDetails>> reply =
                        registryApi.lookupAggregates(con, null, null, null);

                if (!reply.getGeniResponseCode().isSuccess()) throw new JFedException("reply not success");
            }
        };
        return call;
    }

    public GetAdvertisementTask getAdvertisement(@Nonnull final SfaAuthority auth, boolean available) {
        GetAdvertisementTask call = available ? availableAdvertisementCalls.get(auth) : allAdvertisementCalls.get(auth);
        if (call != null) return call;

        call = new GetAdvertisementTask(auth, available);
        if (available)
            availableAdvertisementCalls.put(auth, call);
        else
            allAdvertisementCalls.put(auth, call);
        return call;
    }

    public class FedmonResultsTask extends Task {
        public final String testbedFilter;
        public final String testDefinitionFilter;
        public final boolean successOnly;

        public List<TestResult> filteredResults;
        public List<TestResult> results;

        public FedmonResultsTask(@Nullable String testbedFilter, @Nullable String testDefinitionFilter, boolean successOnly) {
            super("Get Femon Results");
            this.testbedFilter = testbedFilter;
            this.testDefinitionFilter = testDefinitionFilter;
            this.successOnly = successOnly;
        }

        @Override
        protected void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
            Fedmon fedmon = new Fedmon(logger, jFedPreferences);
            SfaAuthority fedmonAuth =
                    authorityList.getByUrnExact("urn:publicid:IDN+flsmonitor.fed4fire.eu+authority+cm").getSfaAuthority();
            GeniUser user = geniUserProvider.getLoggedInGeniUser();
            HttpConnection con =
                    (HttpConnection) connectionProvider.getConnectionByAuthority(user, fedmonAuth, new ServerType(ServerType.GeniServerRole.Fedmon, 1));
            Fedmon.FedmonReply<List<TestResult>> reply =
                    fedmon.getTestResults(con, testbedFilter, testDefinitionFilter);
            if (reply.getGeniResponseCode().isSuccess())
                results = reply.getValue();
            else
                results = new ArrayList<>();
        }

        public List<TestResult> getResults() {
            if (filteredResults == null && successOnly)
                filteredResults = results.stream().filter(TestResult::isSuccessSummary).collect(Collectors.toList());

            if (successOnly)
                return filteredResults;
            else
                return results;
        }
    }

    private FedmonResultsTask fedmonResultsTask = null;

    public FedmonResultsTask getFedmonResultsTask(String testbedFilter, String testDefinitionfilter, boolean successOnly) {
        if (fedmonResultsTask != null) return fedmonResultsTask;

        fedmonResultsTask = new FedmonResultsTask(testbedFilter, testDefinitionfilter, successOnly);
        return fedmonResultsTask;
    }

    //always returns new task!
    public GetOcciNetworksForLocationTask getOcciNetworksForLocation(@Nonnull final SfaAuthority auth, @Nonnull String locationId) {
        return new GetOcciNetworksForLocationTask(auth, locationId);
    }

    public GetOcciStoragesForLocationTask getOcciStoragesForLocation(@Nonnull final SfaAuthority auth, @Nonnull String locationId) {
        return new GetOcciStoragesForLocationTask(auth, locationId);
    }

    /**
     * @param sliceName      name of the slice to be created. (not urn)
     * @param subAuthName    may be null if no subAuthority is needed
     * @param expirationDate time at which the slice should expire. Note that no call offers hard guarantees: so slice might expire later or earlier than this time.
     * @return
     */
    public CreateSliceTask createSlice(@Nonnull String sliceName, @Nullable String subAuthName, @Nullable Instant expirationDate) {
        return new CreateSliceTask(sliceName, subAuthName, expirationDate);
    }

    public List<CreateSliverReservationTask> createSliverReservations(final Slice slice, final Instant reservationStartTime, final Instant reservationEndTime) {
        assert (slice.getRequestRspec() != null);

        List<CreateSliverReservationTask> calls = new ArrayList<>();

        List<GeniUrn> componentManagerUrns = slice.getRequestRspec().getAllComponentManagerUrns();

        //authorityInfos might contain less items than componentManagerUrns, if some URNs are actually the same authority
        Set<AuthorityInfo> authorityInfos = new HashSet<>();
        for (GeniUrn cmUrn : componentManagerUrns) {
            AuthorityInfo ai =
                    authorityList.get(authorityList.finder().findByUrn(cmUrn, AuthorityFinder.Purpose.CREATE_SLIVER));
            if (ai == null) {
                LOG.warn("Could not retrieve authority with URN '" + cmUrn + "'");
                continue;
            }

            SfaAuthority auth = ai.getSfaAuthority();
            if (!(auth.hasSpecialCase(SfaAuthority.SpecialCase.SUPPORTS_RESERVATIONS_NITOS) ||
                    auth.hasSpecialCase(SfaAuthority.SpecialCase.SUPPORTS_RESERVATIONS_GENI))) {
                LOG.warn(String.format("Authority %s does not support reservations!", auth.getHrn()));
                continue;
            }

            authorityInfos.add(ai);
        }
        calls.addAll(authorityInfos.stream().map(ai -> new CreateSliverReservationTask(slice, reservationStartTime, reservationEndTime, ai)).collect(Collectors.toList()));

        return calls;
    }

    /**
     * This assumes the request is stored in the slice, and it will create the slivers on all component managers
     * that are mentioned in the Rspec
     */
    public List<CreateSliverTask> createSliversNoStitching(final Slice slice, final Instant expirationDate, boolean includeProjectMemberSshKeys) {
        List<CreateSliverTask> calls = new ArrayList<>();

        List<GeniUrn> componentManagerUrns = slice.getRequestRspec().getAllComponentManagerUrns();
        LOG.debug("  createSlivers start  componentManagerUrns=" + componentManagerUrns);


        //authorityInfos might contain less items than componentManagerUrns, if some URNs are actually the same authority
        String prefenceForceExoSM =
                jFedPreferences.getString(CorePreferenceKey.PREF_TESTBEDSPECIFIC_EXOGENI_FORCE_EXOSM).trim().toLowerCase(); //AUTO, YES or NO
        Set<AuthorityInfo> authorityInfos = new HashSet<>();
        for (final GeniUrn componentManagerUrn : componentManagerUrns) {
            Boolean forceExoSM;
            if (prefenceForceExoSM.equals("auto"))
                forceExoSM = null;
            else
                forceExoSM = prefenceForceExoSM.equals("yes") ? true : false;

            final AuthorityInfo ai =
                    authorityList.get(authorityList.finder().findByUrn(componentManagerUrn, AuthorityFinder.Purpose.CREATE_SLIVER));
            if (ai == null) {
                //TODO handle better
                LOG.warn("INPUT ERROR: Component Manager Urn in Request RSpec unknown: \"" + componentManagerUrn + "\". Will ignore!");
                continue;
            }
            final SfaAuthority auth = ai.getSfaAuthority();
            assert auth != null;

            authorityInfos.add(ai);

        }

        //temporary fix for nitos reservations: ignore an authority when a lease is detected on one of the nodes
        if (slice.getManifestRspec() != null) {
            FXModelRspec modelRspec = (FXModelRspec) slice.getManifestRspec().getModelRspec(ModelRspecType.FX);

            for (FXRspecNode rspecNode : modelRspec.getNodes()) {
                if (!rspecNode.getLeaseIdRefs().isEmpty()) {
                    final AuthorityInfo ai =
                            authorityList.get(authorityList.finder().findByUrn(rspecNode.getComponentManagerId(), AuthorityFinder.Purpose.CREATE_SLIVER));
                    if (ai == null) {
                        //TODO handle better
                        LOG.warn("INPUT ERROR: Component Manager Urn in Request RSpec unknown: \"" + rspecNode.getComponentManagerId() + "\". Will ignore!");
                        continue;
                    }

                    LOG.info("Removing authority {} from authorities to be called for sliver creation, as a lease has been detected on node {}." +
                                    " This means that the CreateSliver-command has already been called for this authority, and no further actions are necessary.",
                            ai.getSfaAuthority().getName(), rspecNode.getUniqueId());
                    authorityInfos.remove(ai);

                }
            }
        } else {
            LOG.debug("Skipping lease detection, as no manifest is available.");
        }


        for (AuthorityInfo ai : authorityInfos) {
            CreateSliverTask call =
                    new CreateSliverTask(slice, expirationDate, ai, null/*overrideRspec*/, includeProjectMemberSshKeys);

            calls.add(call);
        }

        return calls;
    }

    public List<Task> getSliceStatusAndManifest(final Slice slice) {
        List<Task> res = new ArrayList<>();
        res.add(getSliceStatus(slice));
        res.add(getManifest(slice));
        return res;
    }

    /**
     * call Status on all slivers in the slice. Needs to be sure resolve has been done first, so that all AM involved are known.
     */
    public Task getSliceStatus(final Slice slice) {
        Task call = sliceStatusCalls.get(slice);
        if (call != null) return call;

        call = new Task("Status of Slice") {
            @Override
            public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {

                //create a copy to prevent deadlocks because of a changing number of authorities
                Set<SfaAuthority> manifestAuthorities = new HashSet<>(slice.getManifestComponentManagers());

                //with this latch, we wait for all child-calls to finish before returning ourselves.
                CountDownLatch callsDoneLatch = new CountDownLatch(manifestAuthorities.size());

                for (SfaAuthority auth : manifestAuthorities) {
                    if (auth.hasAmUrl()) {
                        Task subCall = getSliceStatus(slice, auth);

                        tt.addTask(subCall, (task, taskExecution1, state) -> callsDoneLatch.countDown());

                    } else {
                        LOG.error("Tried to get Status for slice \"" + slice.getUrnString() + "\", at authority \"" + auth.getUrnString() + "\", but that authority has no AM URL: will ignore slice on this authority.");
                        callsDoneLatch.countDown();
                    }
                }

                callsDoneLatch.await();
            }

            @Override
            public List<Task> initDependsOn() {
                List<Task> res = new ArrayList<>();
                res.add(getSliceCredential(slice));
                return res;
            }

            @Override
            public List<Task> initAlwaysDependsOn() {
                List<Task> res = new ArrayList<>();
                res.add(getAggregatesForSlice(slice));
                return res;
            }
        };
        sliceStatusCalls.put(slice, call);
        return call;
    }

    /**
     * call delete all slivers in the slice. Needs to be sure resolve has been done first, so that all AM involved are known.
     */
    public Task deleteSlice(final Slice slice) {
        List<Task> calls = new ArrayList<>();

        //TODO this should use multiple tasks, to delete in parallell! (but since we need to resolve first, that is not easy to implement right now)

        Task call = new Task("Delete All Slivers on slice " + slice.getUrnString()) {
            @Override
            public void doTask(TaskExecution singleTask) throws JFedException, InterruptedException {
                if (slice.isExpired()) {
                    throw new RuntimeException("Slice has expired.");
                }
                if (!slice.hasCredential()) {
                    throw new RuntimeException("No credentials for slice. Has it expired? slice expiration date = " + slice.getExpirationDate() + " now=" + new Date());
                }

                Set<SfaAuthority> componentManagers = new HashSet<>();
                componentManagers.addAll(slice.getManifestComponentManagers());

                List<JFedException> exceptions = new ArrayList<>();
                for (SfaAuthority auth : componentManagers)
                    if (!auth.isFake() && auth.hasAmUrl()) {
                        try {
                            for (Sliver sliver : slice.findSlivers(auth)) {
                                sliver.setDeleteRequested(true);
                            }

                            AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(singleTask, auth);
                            amWrapper.deleteSliver(slice.getUrn(), slice.getCredentials());

                            unregisterSliversAtSATaskExecution(slice, auth, this);
                        } catch (JFedException e) {
                            LOG.error("Exception while calling Delete at " + auth.getUrnString() + ": " + e.getMessage(), e);
                            exceptions.add(e);
                        }
                    } else {
                        LOG.warn("Slice has component manager \"" + auth.getUrnString() + "\", but that authority has no known AM URL: cannot delete there.");
                    }

                if (!exceptions.isEmpty())
                    throw new JFedException("There where " + exceptions.size() + " exceptions" +
                            " while calling Delete. (first exception added as cause of this one)",
                            exceptions.get(0));
            }

            @Override
            public List<Task> initDependsOn() {
                List<Task> res = new ArrayList<>();
                res.add(getSliceCredential(slice));
                return res;
            }

            @Override
            public List<Task> initAlwaysDependsOn() {
                List<Task> res = new ArrayList<>();
                res.add(getAggregatesForSlice(slice));
                return res;
            }
        };

        return call;
    }

    public TaskExecution unregisterSliversAtSATaskExecution(Slice slice, SfaAuthority auth, Task extraTaskToDependOn, ResultListener... extraResultListeners) {
        assert slice != null;
        assert auth != null;
        UserAndSliceApiWrapper samaWrapper = getUserAndSliceApiWrapper();
        if (samaWrapper.isRegisterAggregatesForSliceSupported()) {
            List<GeniUrn> localSliverUrns = new ArrayList<>();
            for (Sliver s : slice.getSlivers())
                if (s.getUrn() != null) {
                    if (s.getUrn().getEncodedTopLevelAuthority_withoutSubAuth().equals(auth.getUrn().getEncodedTopLevelAuthority_withoutSubAuth()))
                        localSliverUrns.add(s.getUrn());
                }
            if (!localSliverUrns.isEmpty()) {
                return unregisterSliversAtSATaskExecution(slice, auth, localSliverUrns, extraTaskToDependOn, extraResultListeners);
            }
        }

        return null;
    }

    public TaskExecution unregisterSliversAtSATaskExecution(Slice slice, SfaAuthority auth, Collection<GeniUrn> sliversToUnregister, Task extraTaskToDependOn, ResultListener... extraResultListeners) {
        assert slice != null;
        assert auth != null;

        if (!sliversToUnregister.isEmpty()) {
            UnregisterSliversTask unregisterSliverTask =
                    new UnregisterSliversTask(slice, auth, sliversToUnregister, extraTaskToDependOn, extraResultListeners);
            return tt.addTask(unregisterSliverTask);
        }

        return null;
    }


    public RenewSliceTask renewSlice(Slice slice, Instant newExpirationDate, boolean renewAllSlivers) {
        return new RenewSliceTask(slice, newExpirationDate, renewAllSlivers);
    }

    public RenewSliversTask renewSliversAtAuthority(Slice slice, Instant newExpirationDate, SfaAuthority authority) {
        return new RenewSliversTask(slice, newExpirationDate, authority, null /*renew all slivers*/, false);
    }

    public RenewSliversTask renewSliversAtAuthority(Slice slice, Instant newExpirationDate, SfaAuthority authority, Collection<Sliver> slivers, boolean onlyIfNeeded) {
        return new RenewSliversTask(slice, newExpirationDate, authority, slivers, onlyIfNeeded);
    }

    public GetStatusUntilReadyOrFailTask getGetStatusUntilReadyOrFailTask(SfaAuthority auth, Slice slice) {
        GetStatusUntilReadyOrFailTask res = new GetStatusUntilReadyOrFailTask(auth, slice);
        res.addTaskToScheduleWhenReady(new ListSliceResourcesTask(auth, slice));
        return res;
    }

    /**
     * This assumes the request is stored in the slice, and it will create the slivers on all component managers
     * that are mentioned in the Rspec
     */
    public StitchTask createSliversWithStitching(final Slice slice, final Instant expirationDate, SfaAuthority scsAuth, boolean includeProjectMemberSshKeys, ParallelStitcher.GiveUpStitchingCallback giveUpStitchingCallback) {
        StitchTask call =
                new StitchTask(slice, expirationDate, scsAuth, includeProjectMemberSshKeys, giveUpStitchingCallback);
        return call;
    }

    public Task getShareLanTask(final Slice slice, final SfaAuthority auth, final String lanClientId, final String newSharedLanName) {
        Task shareLanTask = new Task("Share Lan " + lanClientId + " as " + newSharedLanName) {
            @Override
            public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
                AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(taskExecution, auth);
                amWrapper.shareLan(slice.getUrn(), slice.getCredentials(), lanClientId, newSharedLanName);
            }

            @Override
            public List<Task> initDependsOn() {
                List<Task> res = new ArrayList<>();
                res.add(getSliceCredential(slice));
                return res;
            }
        };
        return shareLanTask;
    }

    public Task getUnshareLanTask(final Slice slice, final SfaAuthority auth, final String lanClientId, final String sharedLanName) {
        Task unshareLanTask = new Task("Unshare Lan " + lanClientId + " as " + sharedLanName) {
            @Override
            public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
                AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(taskExecution, auth);
                amWrapper.unshareLan(slice.getUrn(), slice.getCredentials(), lanClientId, sharedLanName);
            }

            @Override
            public List<Task> initDependsOn() {
                List<Task> res = new ArrayList<>();
                res.add(getSliceCredential(slice));
                return res;
            }
        };
        return unshareLanTask;
    }

    public Task getRestartTask(final Slice slice, final GeniUrn restartTarget, final SfaAuthority auth) {
        if (restartTarget == null) return null;

        Task restartLanTask = new Task("Restart " + restartTarget) {
            @Override
            public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
                AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(taskExecution, auth);
                if (!amWrapper.isRestartSupported())
                    throw new JFedHighLevelException("Authority " + auth.getUrnString() + " does not support restart.");
                amWrapper.restart(restartTarget, slice.getCredentials());
            }

            @Override
            public List<Task> initDependsOn() {
                List<Task> res = new ArrayList<>();
                res.add(getSliceCredential(slice));
                return res;
            }
        };
        return restartLanTask;
    }

    public Task getReloadOSTask(final Slice slice, final GeniUrn reloadOSTarget, final SfaAuthority auth) {
        if (reloadOSTarget == null) return null;

        Task reloadOSLanTask = new Task("Reload OS on " + reloadOSTarget) {
            @Override
            public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
                AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(taskExecution, auth);
                if (!amWrapper.isReloadOSSupported())
                    throw new JFedHighLevelException("Authority " + auth.getUrnString() + " does not support reloadOS.");
                amWrapper.reloadOS(reloadOSTarget, slice.getCredentials());
            }

            @Override
            public List<Task> initDependsOn() {
                List<Task> res = new ArrayList<>();
                res.add(getSliceCredential(slice));
                return res;
            }
        };
        return reloadOSLanTask;
    }

    public Task getRestartSliceTask(final Slice slice, final SfaAuthority auth) {
        return getRestartTask(slice, slice.getUrn(), auth);
    }

    public Task getRestartSliverTask(final Sliver sliver) {
        if (sliver.getUrn() == null)
            return null;
        return getRestartTask(sliver.getSlice(), sliver.getUrn(), sliver.getAuthority());
    }

    public Task getReloadOSTask(final Sliver sliver) {
        if (sliver.getUrn() == null)
            return null;
        return getReloadOSTask(sliver.getSlice(), sliver.getUrn(), sliver.getAuthority());
    }

    public class CreateImageTask extends Task {
        private final Slice slice;
        private final GeniUrn sliverUrn;
        private final String imageName;

        private URL resultDiskImageUrl = null;
        private GeniUrn resultDiskImageUrn = null;

        public CreateImageTask(Slice slice, GeniUrn sliverUrn, String imageName) {
            super("Create Image on node " + sliverUrn);
            this.slice = slice;
            this.sliverUrn = sliverUrn;
            this.imageName = imageName;
        }

        @Override
        public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
            Sliver sliver = slice.findSliver(sliverUrn);
            assert (sliver != null);

            AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(taskExecution, sliver.getAuthority());
            if (!amWrapper.areUserDiskImageFunctionsSupported())
                throw new RuntimeException("Authority " + sliver.getAuthority().getUrnString() + " does not support create image.");
            AggregateManagerWrapper.ImageInfo imageInfo =
                    amWrapper.createImage(slice.getUrn(), slice.getCredentials(), sliverUrn, imageName);
            if (imageInfo != null) {
                try {
                    resultDiskImageUrl = new URL(imageInfo.getUrl());
                } catch (MalformedURLException e) {
                    LOG.warn("CreateImageTask createImage returned invalid URL: \"" + imageInfo.getUrl() + "\"", e);
                    resultDiskImageUrl = null;
                }
                resultDiskImageUrn = GeniUrn.parse(imageInfo.getUrn());

            } else {
                throw new JFedHighLevelException("Did not receive a valid answer on CreateImage-call");
            }
        }

        @Override
        public List<Task> initDependsOn() {
            List<Task> res = new ArrayList<>();
            res.add(getSliceCredential(slice));
            return res;
        }

        public URL getResultDiskImageUrl() {
            return resultDiskImageUrl;
        }

        public GeniUrn getResultDiskImageUrn() {
            return resultDiskImageUrn;
        }
    }

    public CreateImageTask getCreateImageTask(final Slice slice, final GeniUrn sliverUrn, String imageName) {
        CreateImageTask createImageTask = new CreateImageTask(slice, sliverUrn, imageName);
        return createImageTask;
    }

    public Task getDeleteImageTask(final SfaAuthority auth, final GeniUrn imageUrn) {
        return new Task("Delete Image " + imageUrn + " on " + auth.getUrnString()) {

            @Override
            protected void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
                AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(taskExecution, auth);
                if (!amWrapper.areUserDiskImageFunctionsSupported())
                    throw new RuntimeException("Authority " + auth.getUrnString() + " does not support list image.");

                UserAndSliceApiWrapper w = getUserAndSliceApiWrapper();
                if (!w.hasUserCredentials())
                    throw new JFedException("UserAndSliceApiWrapper did not have user Credentials.");

                amWrapper.deleteImage(w.getCachedUserCredentialsForAM(), imageUrn);
            }

            @Override
            public List<Task> initDependsOn() {
                List<Task> res = new ArrayList<Task>();
                res.add(getUserCredential());
                return res;
            }
        };
    }


    public class ListDiskImagesTask extends Task {
        private final SfaAuthority auth;
        private List<AggregateManagerWrapper.ImageInfo> diskImageInfos = null;

        public ListDiskImagesTask(SfaAuthority auth) {
            super("List User Images on " + auth.getName());
            this.auth = auth;
        }

        @Override
        protected void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
            AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(taskExecution, auth);
            if (!amWrapper.areUserDiskImageFunctionsSupported())
                throw new RuntimeException("Authority " + auth.getUrnString() + " does not support list image.");


            UserAndSliceApiWrapper w = getUserAndSliceApiWrapper();
            if (!w.hasUserCredentials())
                throw new JFedException("UserAndSliceApiWrapper did not have user Credentials.");

            List<AggregateManagerWrapper.ImageInfo> imageInfos =
                    amWrapper.listImages(w.getCachedUserCredentialsForAM());
            diskImageInfos = Collections.unmodifiableList(imageInfos);
        }

        @Override
        public List<Task> initDependsOn() {
            List<Task> res = new ArrayList<Task>();
            res.add(getUserCredential());
            return res;
        }

        public List<AggregateManagerWrapper.ImageInfo> getDiskImageInfos() {
            return diskImageInfos;
        }
    }

    public ListDiskImagesTask getListDiskImagesTask(final SfaAuthority auth) {
        ListDiskImagesTask listImageTask = new ListDiskImagesTask(auth);
        return listImageTask;

    }


    public class OpenConsoleTask extends Task {
        private final Sliver sliver;

        private URL url = null;

        public OpenConsoleTask(Sliver sliver) {
            super("Get Console URL for sliver " + sliver);
            this.sliver = sliver;
        }

        @Override
        public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
            SfaAuthority auth = sliver.getAuthority();
            Slice slice = sliver.getSlice();
            AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(taskExecution, auth);
            if (!amWrapper.isOpenConsoleSupported())
                throw new RuntimeException("Authority " + auth.getUrnString() + " does not support open console.");
            this.url = amWrapper.openConsole(sliver.getUrn(), slice.getCredentials());
        }

        @Override
        public List<Task> initDependsOn() {
            List<Task> res = new ArrayList<>();
            res.add(getSliceCredential(sliver.getSlice()));
            return res;
        }

        public URL getUrl() {
            return url;
        }
    }

    public OpenConsoleTask getOpenConsoleTask(final Sliver sliver) {
        OpenConsoleTask oenConsoleTask = new OpenConsoleTask(sliver);
        return oenConsoleTask;
    }

    /**
     * Note: One task per user added
     */
    public Task getShareSliceTask(final Slice slice, final GeniUrn userUrn) {
        Task shareSliceTask =
                new Task("Share slice \"" + slice.getUrn().getEncodedResourceName() + "\" with " + userUrn.getEncodedResourceName()) {
                    @Override
                    public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
                        UserAndSliceApiWrapper saWrapper = getUserAndSliceApiWrapper();
                        saWrapper.shareSlice(
                                JavaFXLogger.wrap(logger, taskExecution),
                                slice.getUrn(),
                                slice.getCredentials(),
                                userUrn);
                    }

                    @Override
                    public List<Task> initDependsOn() {
                        List<Task> res = new ArrayList<>();
                        res.add(getSliceCredential(slice));
                        return res;
                    }
                };
        return shareSliceTask;
    }

    /**
     * @param slice
     * @param userSpecs as in http://groups.geni.net/geni/wiki/GAPI_AM_API_DRAFT/Adopted#ChangeSetQ:Supportchangingusersandkeysonexistingcomputeslivers
     *                  - New users may be added
     *                  - Omitting a user means there will be no change to the keys installed for that user. It does not mean the user should be removed.
     *                  - Existing users will have all SSH keys replaced by the new set. Note that a user may have an empty list of SSH keys specified, effectively preventing the user from accessing the node.
     * @return
     */
    public Task getEditSshKeysTask(final Slice slice, final SfaAuthority auth, final List<UserSpec> userSpecs) {
        Task editSshKeysTask = new Task("Edit SSH Keys") {
            @Override
            public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
                AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(taskExecution, auth);
                if (!amWrapper.isEditSshKeysSupported())
                    throw new JFedException("Authority " + auth.getUrnString() + " does not support edit of SSH keys.");
                amWrapper.editSshKeys(slice.getUrn(), slice.getCredentials(), userSpecs);
            }

            @Override
            public List<Task> initDependsOn() {
                List<Task> res = new ArrayList<>();
                res.add(getSliceCredential(slice));
                return res;
            }
        };
        return editSshKeysTask;
    }

    private UserAuthorityGetVersionTask userAuthorityGetVersionTask = null;

    public UserAuthorityGetVersionTask getUserAuthorityGetVersionTask() {
        if (userAuthorityGetVersionTask == null) {
            userAuthorityGetVersionTask =
                    new UserAuthorityGetVersionTask(geniUserProvider.getLoggedInGeniUser().getUserAuthority());
        }
        return userAuthorityGetVersionTask;
    }

    public final class UserAuthorityGetVersionTask extends Task {
        private boolean supportSliverRegistration;
        private SfaAuthority userAuthority;

        private boolean known;
        private boolean hasFedMAApiv2;
        private boolean hasFedSAApiv2;

        public UserAuthorityGetVersionTask(SfaAuthority userAuthority) {
            super("GetVersion Info @ " + userAuthority.getName());

            this.userAuthority = userAuthority;
            supportSliverRegistration = false; //initial value
            known = false;

            this.hasFedMAApiv2 = userAuthority.getUrl(ServerType.GeniServerRole.GENI_CH_MA, 2) != null;
            this.hasFedSAApiv2 = userAuthority.getUrl(ServerType.GeniServerRole.GENI_CH_SA, 2) != null;

            if (!hasFedSAApiv2) {
                known = true;
                supportSliverRegistration = false;
            }

            LOG.debug("UserAuthorityGetVersionTask: init done known=" + known
                    + " supportSliverRegistration=" + supportSliverRegistration +
                    " hasFedMAApiv2=" + hasFedMAApiv2 + " hasFedSAApiv2=" + hasFedSAApiv2);
        }

        @Override
        public void doTask(TaskExecution singleTask) throws JFedException, InterruptedException {
            FederationSliceAuthorityApi2 sa =
                    new FederationSliceAuthorityApi2(JavaFXLogger.wrap(logger, singleTask), jFedPreferences);

            JFedConnection con =
                    connectionProvider.getConnectionByAuthority(geniUserProvider.getLoggedInGeniUser(), userAuthority, FederationSliceAuthorityApi2.class);

            AbstractFederationApi.FederationApiReply<FederationSliceAuthorityApi2.GetVersionSAResult> gv =
                    sa.getVersion((SfaConnection) con);
            if (!gv.getGeniResponseCode().isSuccess()) {
                known = true;
                supportSliverRegistration = false;
                throw new JFedException("GetVersion call failed", gv.getXmlRpcCallDetailsWithCodeValueError(), gv.getGeniResponseCode());
            }
            FederationSliceAuthorityApi2.GetVersionSAResult r = gv.getValue();
            if (r == null) {
                known = true;
                supportSliverRegistration = false;
                throw new JFedException("null reply for GetVersion call", gv.getXmlRpcCallDetailsWithCodeValueError(), gv.getGeniResponseCode());
            }
            known = true;
            supportSliverRegistration = r.getServices().contains("SLIVER_INFO");
        }

        public boolean isKnown() {
            return known;
        }

        public boolean isSupportSliverRegistration() {
            return supportSliverRegistration;
        }
    }

    public final class SliceStatusTask extends Task {
        private final Slice slice;
        private final SfaAuthority auth;

        private StatusDetails status = null;

        private SliceStatusTask(Slice slice, SfaAuthority auth) {
            super("Status of Slice @ " + auth.getName());
            this.slice = slice;
            this.auth = auth;
        }

        @Override
        public void doTask(TaskExecution singleTask) throws JFedException, InterruptedException {
            AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(singleTask, auth);
            status = amWrapper.status(slice.getUrn(), slice.getCredentials());
            if (status == null)
                throw new JFedHighLevelException("Sliver status update failed");
        }

        @Override
        public List<Task> initDependsOn() {
            return Collections.singletonList(getSliceCredential(slice));
        }

        public StatusDetails getStatus() {
            return status;
        }
    }

    public class GetAggregatesForSliceTask extends Task {

        private final Slice slice;

        private List<GeniUrn> aggregates;

        public GetAggregatesForSliceTask(Slice slice) {
            super("Resolve Slice");
            this.slice = slice;
        }

        @Override
        protected void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
            UserAndSliceApiWrapper w = getUserAndSliceApiWrapper();

            List<String> aggregateStrings = w.getAggregatesForSlice(JavaFXLogger.wrap(logger, taskExecution),
                    slice.getCredentials(),
                    GeniUrn.parse(slice.getUrnString()));

            if (aggregateStrings != null) {
                this.aggregates = new ArrayList<>();
                for (String aggregateString : aggregateStrings) {
                    GeniUrn aggregateUrn = GeniUrn.parse(aggregateString);

                    if (aggregateUrn == null) {
                        LOG.warn("Could not parse '{}' into an URN. IGNORING", aggregateString);
                    }
                    this.aggregates.add(aggregateUrn);
                }

            } else {
                this.aggregates = null;
            }
        }

        @Override
        public List<Task> initDependsOn() {
            List<Task> res = new ArrayList<>();

            LOG.debug("resolveSlice Task#initDependsOn -> wrapper.needSubAuthCredential()=" +
                    "  user auth urn=" + geniUserProvider.getLoggedInGeniUser().getUserAuthority().getUrnString());

            res.add(getUserCredential());
            res.add(getSliceCredential(slice));
            return res;
        }

        public List<GeniUrn> getAggregates() {
            return aggregates;
        }

        // if the slice did not receive any aggregate managers in the first call, it is likely empty.
//        @Override
//        public boolean newExecutionNeeded() {
//            //resolve if no component managers
//            return slice.getManifestComponentManagers().isEmpty();
//        }
    }

    public class ListSubAuthoritiesTask extends Task {
        private UserAndSliceApiWrapper.SubAuthoritySupport subAuthoritySupport;
        private List<String> subAuthNames;

        public ListSubAuthoritiesTask() {
            super("List Sub Authorities");
        }

        @Override
        public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
            UserAndSliceApiWrapper w = getUserAndSliceApiWrapper();

            subAuthoritySupport = w.getSubAuthoritySupport(JavaFXLogger.wrap(logger, taskExecution));
            easyModel.setSubAuthoritySupport(subAuthoritySupport); //easymodel cannot always see this itself, as this does not always require a call!
            subAuthNames =
                    w.getSubAuthorityNames(JavaFXLogger.wrap(logger, taskExecution), geniUserProvider.getLoggedInGeniUser().getUserUrn());
        }

        public List<String> getSubAuthNames() {
            return subAuthNames;
        }

        public UserAndSliceApiWrapper.SubAuthoritySupport getSubAuthoritySupport() {
            return subAuthoritySupport;
        }

        @Override
        public List<Task> initDependsOn() {
            List<Task> res = new ArrayList<>();
            res.add(getUserCredential());
            return res;
        }
    }

    public final class GetAdvertisementTask extends Task {

        private final SfaAuthority auth;
        private final boolean available;
        private String advertisementRspecString;

        public GetAdvertisementTask(SfaAuthority auth, boolean available) {
            super("Fetch Advertisement RSpec on " + auth.getName() + " (" + (available ? "available)" : "all)"));
            this.auth = auth;
            this.available = available;
        }

        @Override
        public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
            AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(taskExecution, auth);

            UserAndSliceApiWrapper w = getUserAndSliceApiWrapper();
            if (!w.hasUserCredentials())
                throw new JFedException("UserAndSliceApiWrapper did not have user Credentials.");

            advertisementRspecString = amWrapper.listResources(w.getCachedUserCredentialsForAM(), available);

            if (advertisementRspecString == null)
                throw new JFedHighLevelException("Got null when asked " + auth.getUrnString() + " for advertisement.");
        }

        @Override
        public List<Task> initDependsOn() {
            return Collections.singletonList(getUserCredential());
        }

        public String getAdvertisementRspecString() {
            return advertisementRspecString;
        }
    }

    public final class GetOcciNetworksForLocationTask extends Task {
        private final SfaAuthority auth;
        private final String locationId;

        public GetOcciNetworksForLocationTask(SfaAuthority auth, String locationId) {
            super("Fetch OCCI networks for location " + locationId);
            this.auth = auth;
            this.locationId = locationId;
        }

        @Override
        public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
            OCCI occiApi = new OCCI(logger, jFedPreferences);

            GeniUser user = geniUserProvider.getLoggedInGeniUser();
            RestSslPasswordConnection con = (RestSslPasswordConnection)
                    connectionProvider.getConnectionByAuthority(user, auth, new ServerType(ServerType.GeniServerRole.OCCI, 1));
            if (con == null)
                throw new RuntimeException("No connection can be created. Perhaps the authority \"" + auth.getUrnString() + "\" does not have an OCCI interface?");

            OCCI.OCCIReply<String> reply = occiApi.getLocationNetworks(con, locationId);

            if (!reply.getGeniResponseCode().isSuccess())
                throw new JFedException("OCCI Reply is not success. request URL=" +
                        (reply.getHttpCallDetails() == null ? "null" : reply.getHttpCallDetails().getServerUrl())
                        + "\n\nHTTP reply=" +
                        (reply.getHttpCallDetails() == null ? "null" :
                                (reply.getHttpCallDetails().getResultHttpStatusLine() + "\n" +
                                        reply.getHttpCallDetails().getResultHttpHeaders() + "\n\n" +
                                        reply.getHttpCallDetails().getResultHttpContent())
                        )
                );
        }
    }

    public final class GetOcciStoragesForLocationTask extends Task {
        private final SfaAuthority auth;
        private final String locationId;

        public GetOcciStoragesForLocationTask(SfaAuthority auth, String locationId) {
            super("Fetch OCCI storages for location " + locationId);
            this.auth = auth;
            this.locationId = locationId;
        }

        @Override
        public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
            OCCI occiApi = new OCCI(logger, jFedPreferences);

            GeniUser user = geniUserProvider.getLoggedInGeniUser();
            RestSslPasswordConnection con = (RestSslPasswordConnection)
                    connectionProvider.getConnectionByAuthority(user, auth, new ServerType(ServerType.GeniServerRole.OCCI, 1));
            if (con == null)
                throw new RuntimeException("No connection can be created. Perhaps the authority \"" + auth.getUrnString() + "\" does not have an OCCI interface?");

            OCCI.OCCIReply<String> reply = occiApi.getLocationStorages(con, locationId);

            if (!reply.getGeniResponseCode().isSuccess()) throw new JFedException("reply not success");
        }
    }

    public final class CreateSliceTask extends Task {

        private final String sliceName;
        private final String subAuthority;
        private final Instant expirationDate;
        private Slice slice;

        /**
         * subAuthority may be null (= no sub authority)
         */
        public CreateSliceTask(@Nonnull String sliceName, @Nullable String subAuthority) {
            this(sliceName, subAuthority, null);
        }

        /**
         * subAuthority may be null (= no sub authority)
         */
        public CreateSliceTask(@Nonnull String sliceName, @Nullable String subAuthority, @Nullable Instant expirationDate) {
            super("Create Slice");
            this.sliceName = sliceName;
            this.subAuthority = subAuthority;
            this.expirationDate = expirationDate;
        }

        @Override
        public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
            UserAndSliceApiWrapper w = getUserAndSliceApiWrapper();
            LOG.debug("CreateSliceTask Requesting slice with expiration time " +
                            (expirationDate == null ? "null" : expirationDate.toString() + "")
            );

            assert w.hasUserCredentials() : "There was no valid user credential available.";
            if (!w.hasUserCredentials())
                LOG.warn("Register call saw that UserAndSliceApiWrapper does not have a valid credential. The call will most lickely fail.");

            UserAndSliceApiWrapper.SliceInfo sliceInfo;

            sliceInfo =
                    w.createSlice(JavaFXLogger.wrap(logger, taskExecution), sliceName,
                            expirationDate != null ? Date.from(expirationDate) : null, subAuthority);

            if (sliceInfo.getCredentials() != null) {
                GeniUrn sliceUrn = sliceInfo.getUrn();

                slice = easyModel.getSlice(sliceUrn.toString());
                if (slice == null) {
                    LOG.error("CreateSliceTask failed: slice was not added to EasyModel sliceUrn=" + sliceUrn);
                    throw new JFedHighLevelException("Creation of slice was successful, but Task FAILED because slice was not added to EasyModel. sliceUrn=" + sliceUrn);
                }
                LOG.debug("CreateSliceTask success: " + sliceUrn.toString() + " ");
            } else {
                LOG.error("CreateSliceTask failed: sliceCredential == null");
                throw new JFedHighLevelException("Creation of slice failed");
            }
        }

        @Override
        public List<Task> initDependsOn() {
            List<Task> res = new ArrayList<>();
            res.add(getUserCredential());
            return res;
        }

        public String getSliceName() {
            return sliceName;
        }

        public Slice getSlice() {
            return slice;
        }
    }

    public final class RenewSliceTask extends Task {

        @Nonnull
        private final Slice slice;
        @Nonnull
        private final Instant newExpirationDate;
        private final boolean renewAllSlivers;

        public RenewSliceTask(@Nonnull Slice slice, @Nonnull Instant newExpirationDate, boolean renewAllSlivers) {
            super("Renew Slice");
            this.slice = slice;
            this.newExpirationDate = newExpirationDate;
            this.renewAllSlivers = renewAllSlivers;
        }

        @Override
        protected void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
            UserAndSliceApiWrapper w = getUserAndSliceApiWrapper();

            if (slice.getExpirationDate() == null || slice.getExpirationDate().isBefore(newExpirationDate)) {
                try {
                    w.renewSlice(JavaFXLogger.wrap(logger, taskExecution), slice.getCredentials(), Date.from(newExpirationDate));
                } catch (Exception e) {
                    LOG.error("Failed to renew slice: " + e.getMessage(), e);

                    //failure to renew slice will normally cause failure to renew slivers
                    if (slice.getExpirationDate().isBefore(newExpirationDate))
                        if (e instanceof JFedException)
                            throw (JFedException) e;
                        else
                            throw new JFedException("failed to renew slice " + slice.getUrn(), e);
                }
            } else
                LOG.info("Not renewing slice: new expiration date (" + newExpirationDate + ") is not after current expiration date (" + slice.getExpirationDate() + ")");

            //continue with trying to renew slivers (slice might already be long enough, which would also cause failure)

            if (this.renewAllSlivers) {
                Set<SfaAuthority> authorities = new HashSet<>();
                for (Sliver sliver : slice.getSlivers()) {
                    authorities.add(sliver.getAuthority());
                }

                //TODO do this in parallel
                for (SfaAuthority auth : authorities) {
                    if (auth.isFake())
                        continue;

                    AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(taskExecution, auth);

                    amWrapper.renewSlivers(
                            slice.getUrn(),
                            null/*sliver URNs (not needed because all slivers in slice need to be renewed) todo: provide them anyway for completeness */,
                            slice.getCredentials(),
                            Date.from(newExpirationDate));
                }

                List<Sliver> slivers = slice.getSlivers();
                List<GeniUrn> sliverUrns = new ArrayList<>();
                if (slivers != null)
                    for (Sliver sliver : slivers) {
                        if (sliver.getUrn() != null &&
                                (sliver.getStatus().getGlobalStatus().equals(StatusDetails.SliverStatus.READY)
                                        || sliver.getStatus().getGlobalStatus().equals(StatusDetails.SliverStatus.CHANGING)
                                        || sliver.getStatus().getGlobalStatus().equals(StatusDetails.SliverStatus.FAIL)
                                ))
                            sliverUrns.add(sliver.getUrn());
                    }
                if (sliverUrns.isEmpty())
                    sliverUrns = null;

                if (sliverUrns != null && !sliverUrns.isEmpty()) {
                    UpdateSliversTask updateSliverTask =
                            new UpdateSliversTask(this, slice, sliverUrns, newExpirationDate);
                    tt.addTask(updateSliverTask);
                }
            }
        }

        @Override
        public List<Task> initDependsOn() {
            List<Task> res = new ArrayList<>();
            res.add(getSliceCredential(slice));
            return res;
        }

        @Override
        public List<Task> initAlwaysDependsOn() {
            List<Task> res = new ArrayList<>();
            res.add(getAggregatesForSlice(slice));
            return res;
        }
    }

    public final class RenewSliversTask extends Task {
        private final Slice slice;
        private final Instant newExpirationDate;
        private final SfaAuthority auth;
        private final Collection<Sliver> slivers;
        private final boolean onlyIfNeeded;

        private Collection<Task> extraDependencies = new ArrayList<>();

        /**
         * Renew one or more or all slivers at the given authority.
         * <p/>
         * This doesn't care about slice expiration date, it just tries to renew.
         *
         * @param slice             the slice the slivers are all in. may ot be null.
         * @param newExpirationDate
         * @param auth              may not be null
         * @param slivers           may be null to renew all slivers at the authority. May be list of selected slivers.
         */
        public RenewSliversTask(Slice slice, Instant newExpirationDate, SfaAuthority auth, Collection<Sliver> slivers, boolean onlyIfNeeded) {
            super("Renew Slivers @ " + auth.getName());
            assert slice != null;
            this.slice = slice;
//            assert slice.getSlivers() != null; //more a warning if this is null
            assert newExpirationDate != null;
            this.newExpirationDate = newExpirationDate;
            this.auth = auth;
            this.slivers = slivers;
            this.onlyIfNeeded = onlyIfNeeded;
        }

        private void addExtraDependency(Task t) {
            extraDependencies.add(t);
        }

        @Override
        protected void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
            if (onlyIfNeeded) {
                assert newExpirationDate != null;
                boolean needed = false;
                Instant newExpirationDateNoSubSecond = Instant.ofEpochSecond(newExpirationDate.getEpochSecond());

                List<Sliver> sliversChecked = new ArrayList<>();
                if (slivers == null) {
                    assert slice.getSlivers() != null;
                    sliversChecked.addAll(slice.getSlivers());
                } else {
                    sliversChecked.addAll(slivers);
                }

                //if no slivers known, fall back to trying renew anyway
                if (sliversChecked.isEmpty())
                    needed = true;

                for (Sliver sliver : sliversChecked) {
                    if (sliver.getExpirationDate() != null && !sliver.getExpirationDate().equals(newExpirationDateNoSubSecond)) {
                        LOG.debug("RenewSliversTask task needs Renew for sliver " + sliver.getUrn() + " expiring at " + sliver.getExpirationDate());
                        needed = true;
                        break;
                    } else {
                        LOG.debug("RenewSliversTask task does not need renew for sliver " + sliver.getUrn() + " expiring at " + sliver.getExpirationDate());
                    }
                }
                if (!needed) {
                    LOG.info("RenewSliversTask task running for date " + newExpirationDate + ", but none of the " + sliversChecked.size() + " slivers need to be renewed: will do nothing.");
                    return;
                } else {
                    LOG.info("RenewSliversTask task running for date " + newExpirationDate + ", and at least one sliver (of " + sliversChecked.size() + ") needs to be renewed: will continue.");
                }
            }

            UserAndSliceApiWrapper w = getUserAndSliceApiWrapper();

            List<GeniUrn> sliverUrns = new ArrayList<>();
            if (slivers != null)
                for (Sliver sliver : slivers) {
                    if (sliver.getUrn() != null &&
                            (sliver.getStatus().getGlobalStatus().equals(StatusDetails.SliverStatus.READY)
                                    || sliver.getStatus().getGlobalStatus().equals(StatusDetails.SliverStatus.CHANGING)
                                    || sliver.getStatus().getGlobalStatus().equals(StatusDetails.SliverStatus.FAIL)
                            ))
                        sliverUrns.add(sliver.getUrn());
                }
            if (sliverUrns.isEmpty())
                sliverUrns = null;

            AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(taskExecution, auth);
            amWrapper.renewSlivers(
                    slice.getUrn(),
                    sliverUrns,
                    slice.getCredentials(),
                    Date.from(newExpirationDate));

            if (sliverUrns != null && !sliverUrns.isEmpty()) {
                UpdateSliversTask updateSliverTask =
                        new UpdateSliversTask(this, slice, sliverUrns, newExpirationDate);
                tt.addTask(updateSliverTask);
            }

        }

        @Override
        public List<Task> initDependsOn() {
            List<Task> res = new ArrayList<>();
            res.add(getSliceCredential(slice));
            res.addAll(extraDependencies);
            return res;
        }
    }

    public final class RegisterSliversTask extends Task {
        @Nonnull
        private final Slice slice;
        @Nonnull
        private final Collection<GeniUrn> createdSliverUrns;
        @Nullable
        private final Task extraTaskToDependOn;
        @Nullable
        private final Instant sliverExpirationDate;
        @Nonnull
        private final SfaAuthority auth;
        @Nonnull
        private final ResultListener[] extraResultListeners;

        public RegisterSliversTask(@Nonnull Slice slice, @Nonnull SfaAuthority auth,
                                   @Nonnull Collection<GeniUrn> createdSliverUrns, @Nullable Instant sliverExpirationDate,
                                   @Nullable Task extraTaskToDependOn,
                                   ResultListener... extraResultListeners) {
            super("Registering slivers created @ " + auth.getUrnString());

            this.slice = slice;
            this.createdSliverUrns = createdSliverUrns;
            this.sliverExpirationDate = sliverExpirationDate;
            this.auth = auth;
            this.extraTaskToDependOn = extraTaskToDependOn;
            this.extraResultListeners = extraResultListeners;
        }

        @Override
        protected void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
            UserAndSliceApiWrapper samaWrapper = getUserAndSliceApiWrapper();
            ResultListener[] resultListeners = Arrays.copyOf(extraResultListeners, extraResultListeners.length + 1);
            resultListeners[extraResultListeners.length] = taskExecution;
            Logger logger = JavaFXLogger.wrap(HighLevelController.this.logger, resultListeners);

            Collection<GeniUrn> sliverUrns = createdSliverUrns;

            LOG.debug("RegisterSliversTask has " + sliverUrns.size() + " sliver Urns to register.");
            if (!sliverUrns.isEmpty())
                samaWrapper.registerAggregatesForSlice(logger,
                        slice.getCredentials(), slice.getUrn(),
                        auth.getUrn(), sliverUrns,
                        sliverExpirationDate != null ? Date.from(sliverExpirationDate) : null);
        }

        @Override
        public List<Task> initDependsOn() {
            List<Task> res = new ArrayList<>();
            if (extraTaskToDependOn != null)
                res.add(extraTaskToDependOn);
            res.add(getSliceCredential(slice));
            return res;
        }
    }

    public final class UnregisterSliversTask extends Task {
        private final Slice slice;
        private final Collection<GeniUrn> deletedSliverUrns;
        private final Task extraTaskToDependOn;
        private final SfaAuthority auth;
        private final ResultListener[] extraResultListeners;

        public UnregisterSliversTask(Slice slice, SfaAuthority auth, Collection<GeniUrn> deletedSliverUrns, Task extraTaskToDependOn,
                                     ResultListener... extraResultListeners) {
            super("UnregisterSliversTask slivers deleted @ " + auth.getUrnString());

            this.slice = slice;
            this.deletedSliverUrns = deletedSliverUrns;
            this.auth = auth;
            this.extraTaskToDependOn = extraTaskToDependOn;
            this.extraResultListeners = extraResultListeners;
        }

        @Override
        protected void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
            UserAndSliceApiWrapper samaWrapper = getUserAndSliceApiWrapper();
            ResultListener[] resultListeners = Arrays.copyOf(extraResultListeners, extraResultListeners.length + 1);
            resultListeners[extraResultListeners.length] = taskExecution;
            Logger logger = JavaFXLogger.wrap(HighLevelController.this.logger, resultListeners);

//            List<Sliver> slivers = createdSlivers;

//            LOG.trace("RegisterSliversTask found "+ slivers.size()+" slivers to register.");
//            List<GeniUrn> sliverUrns = new ArrayList<>();
//            for (Sliver sliver : slivers) {
//                if (sliver.getUrn() != null)
//                    try {
//                        sliverUrns.add(new GeniUrn(sliver.getUrn()));
//                    } catch (GeniUrn.GeniUrnParseException e) {
//                        LOG.warn("GeniUrnParseException for \""+sliver.getUrn()+"\"", e);
//                    }
//            }
            Collection<GeniUrn> sliverUrns = deletedSliverUrns;
            LOG.debug("UnregisterSliversTask has " + sliverUrns.size() + " sliver Urns to unregister.");
            if (!sliverUrns.isEmpty())
                samaWrapper.unregisterAggregatesForSlice(logger,
                        slice.getCredentials(), slice.getUrn(),
                        auth.getUrn(), sliverUrns);
        }

        @Override
        public List<Task> initDependsOn() {
            List<Task> res = new ArrayList<>();
            if (extraTaskToDependOn != null)
                res.add(extraTaskToDependOn);
            res.add(getSliceCredential(slice));
            return res;
        }
    }

    public final class UpdateSliversTask extends Task {
        private final Slice slice;
        private final Collection<GeniUrn> changedSliverUrns;
        private final Task task;
        private final Instant newSliverExpirationDate;

        /**
         * @param task
         * @param slice
         * @param changedSliverUrns
         * @param newSliverExpirationDate
         */
        public UpdateSliversTask(@Nonnull Task task, @Nonnull Slice slice,
                                 @Nonnull Collection<GeniUrn> changedSliverUrns, @Nonnull Instant newSliverExpirationDate) {
            super("Updating registered expiration date of slivers");

            assert task != null;
            assert slice != null;
            assert changedSliverUrns != null;
            assert newSliverExpirationDate != null;

            this.slice = slice;
            this.changedSliverUrns = changedSliverUrns;
            this.newSliverExpirationDate = newSliverExpirationDate;
            this.task = task;
        }

        @Override
        protected void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
            UserAndSliceApiWrapper samaWrapper = getUserAndSliceApiWrapper();

            if (!samaWrapper.isRegisterAggregatesForSliceSupported()) {
                LOG.debug("UpdateSliversTask: This SA does not support updating sliver info. Doing nothing.");
                return;
            }

            Logger logger = JavaFXLogger.wrap(HighLevelController.this.logger, taskExecution);

            Collection<GeniUrn> sliverUrns = changedSliverUrns;
            //Date sliceExpiration = slice.getExpirationDate();
            LOG.debug("UpdateSliversTask has " + sliverUrns.size() + " sliver Urns to update.");
            if (!sliverUrns.isEmpty())
                samaWrapper.updateSliverExpirationDateInfoForSlice(logger,
                        slice.getCredentials(), slice.getUrn(),
                        sliverUrns, Date.from(newSliverExpirationDate));
        }

        @Override
        public List<Task> initDependsOn() {
            List<Task> res = new ArrayList<>();
            res.add(task);
            res.add(getSliceCredential(slice));
            return res;
        }
    }

    public final class CreateSliverReservationTask extends Task {
        private final Slice slice;
        private final Instant reservationStartTime;
        private final Instant reservationEndTime;
        private final AuthorityInfo ai;
        private final SfaAuthority auth;

        private String manifestRspec;
        private String status;

        private List<GeniUrn> failedResources = new ArrayList<>();

        /**
         * Use this constructor to make a future reservation
         */
        public CreateSliverReservationTask(@Nonnull Slice slice, @Nonnull Instant reservationStartTime,
                                           @Nonnull Instant reservationEndTime, AuthorityInfo ai) {
            super("Create Sliver @ " + ai.getSfaAuthority().getUrnString());
            this.slice = slice;
            this.reservationStartTime = reservationStartTime;
            this.reservationEndTime = reservationEndTime;
            this.ai = ai;
            this.auth = ai.getSfaAuthority();
        }


        @Override
        protected void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
            AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(taskExecution, auth);

            if (auth.hasSpecialCase(SfaAuthority.SpecialCase.SUPPORTS_RESERVATIONS_GENI)) {


                manifestRspec = amWrapper.allocate(slice.getUrn(),
                        slice.getCredentials(),
                        slice.getRequestRspec().getRspecXmlString(),
                        Date.from(reservationStartTime),
                        Date.from(reservationEndTime)
                );


            } else if (auth.hasSpecialCase(SfaAuthority.SpecialCase.SUPPORTS_RESERVATIONS_NITOS)) {

                //create a copy of the request rspec
                RequestRspecSource requestCopy =
                        new RequestRspecSource(new StringRspec(slice.getRequestRspec().getRspecXmlString()));

                assert requestCopy.getModelRspec() != null && !requestCopy.getModelRspec().isLosingData();

                ModelRspec requestWithLease = requestCopy.getModelRspec(ModelRspecType.FX);

                String clientLeaseId = UUID.randomUUID().toString();

                //we have to edit the rspec to include the lease request
                Lease newLease = rspecFactory.createLease(requestWithLease, null/*lease id*/,
                        clientLeaseId,
                        Date.from(reservationStartTime),
                        Date.from(reservationEndTime));

                requestWithLease.addLease(newLease);

                requestWithLease.getNodes().stream()
                        .filter(node -> auth.getUrn().equals(node.getComponentManagerId()))
                        .forEach(node -> {
                            node.getLeaseIdRefs().clear();
                            node.addLeaseIdRef(clientLeaseId);
                        });

                requestWithLease.getChannels().stream()
                        .filter(channel -> auth.getUrn().equals(channel.getComponentManagerId()))
                        .forEach(channel -> {
                            channel.getLeaseIdRefs().clear();
                            channel.addLeaseIdRef(clientLeaseId);
                        });

                RequestRspecSource requestWithLeaseSource = new RequestRspecSource(requestWithLease);

                List<UserSpec> userSpecs =
                        UserSpecHelper.getMergedUserSpecs(geniUserProvider, easyModel, requestWithLease);


                manifestRspec = amWrapper.createSliver(slice.getUrn(),
                        slice.getCredentials(),
                        requestWithLeaseSource.getRspecXmlString(),
                        userSpecs,
                        Date.from(reservationEndTime));

                if (manifestRspec != null) {
                    //check if lease was successfull, otherwise set manifestRspec to null to indicate a reservation failure


                    ManifestRspecSource manifestRspecSource = new ManifestRspecSource(new StringRspec(manifestRspec));
                    FXModelRspec modelRspec = (FXModelRspec) manifestRspecSource.getModelRspec(ModelRspecType.FX);

                    failedResources.addAll(
                            modelRspec.getNodes().stream()
                                    .filter(rspecNode -> auth.getUrn().equals(rspecNode.getComponentManagerId()))
                                    .filter(rspecNode -> rspecNode.getLeaseIdRefs().isEmpty())
                                    .map(RspecNode::getComponentId)
                                    .collect(Collectors.toList())
                    );




                    failedResources.addAll(
                            modelRspec.getChannels().stream()
                                    .filter(channel -> auth.getUrn().equals(channel.getComponentManagerId()))
                                    .filter(channel -> channel.getLeaseIdRefs().isEmpty())
                                    .map(Channel::getComponentId)
                                    .collect(Collectors.toList())
                    );

                    //sometimes, nodes are fully removed from the manifest instead of being returned without lease.
                    // Also support this scenario!
                    failedResources.addAll(
                            requestWithLease.getNodes().stream()
                                    .filter(node -> auth.getUrn().equals(node.getComponentManagerId()))
                                    .filter(node -> modelRspec.getNodeByClientId(node.getClientId()) == null)
                                    .map(RspecNode::getComponentId)
                                    .collect(Collectors.toList())
                    );

                    //while we're at it: do the same for channels
                    failedResources.addAll(
                            requestWithLease.getChannels().stream()
                                    .filter(channel -> auth.getUrn().equals(channel.getComponentManagerId()))
                                    .filter(channel -> modelRspec.getChannelByComponentId(channel.getComponentId()) == null)
                                    .map(Channel::getComponentId)
                                    .collect(Collectors.toList())
                    );



                    if (!failedResources.isEmpty()) {
                        status =
                                "Could not make a reservation for resource(s) " + failedResources.stream().map(GeniUrn::toString).collect(Collectors.joining(", "));
                        manifestRspec = null;
                        throw new JFedHighLevelException("Could not make a reservation for node(s) " + failedResources.stream().map(GeniUrn::toString).collect(Collectors.joining(", ")));
                    }
                } else {
                    status = "Reservation call was unsuccessful";
                }
            } else {
                throw new JFedHighLevelException("Making a reservation is not supported on this authority!");
            }


        }

        @Override
        protected List<Task> getDependsOn() {
            return Collections.singletonList(getSliceCredential(slice));
        }

        public String getManifestRspec() {
            return manifestRspec;
        }

        public SfaAuthority getAuthority() {
            return auth;
        }

        public String getStatus() {
            return status;
        }

        public List<GeniUrn> getFailedResources() {
            return failedResources;
        }
    }


    public final class CreateSliverTask extends Task {
        private final Slice slice;
        private final Instant expirationTime;
        private final AuthorityInfo ai;
        private final SfaAuthority auth;
        private final String overrideRspec;
        private final boolean includeProjectMemberSshKeys;
        private Sliver sliver;
        private String manifestRspec;
        private boolean failedDueToNotEnoughFreeResources;
        private GetStatusUntilReadyOrFailTask getStatusUntilReadyOrFailTask;

        private final List<Task> extraDependencies = new ArrayList<>();

        /**
         * Use this constructor to create a sliver that will be used immediately!
         *
         * @param overrideRspec may be null, in that case, the rspec is taken from the slice
         */
        public CreateSliverTask(Slice slice, Instant expirationTime, AuthorityInfo ai, String overrideRspec, boolean includeProjectMemberSshKeys) {
            super("Create Sliver @ " + ai.getSfaAuthority().getUrnString());
            this.slice = slice;
            this.expirationTime = expirationTime;
            this.ai = ai;
            this.auth = ai.getSfaAuthority();
            this.overrideRspec = overrideRspec;
            this.includeProjectMemberSshKeys = includeProjectMemberSshKeys;
            this.getStatusUntilReadyOrFailTask = new GetStatusUntilReadyOrFailTask(auth, slice);
        }

        /**
         * see FEDIBBTDEV-1334: ION and others see "expires" in the request RSpec as the requested sliver expiration
         * so if not yet specified, we add it here
         * <p/>
         * note: this modifies the ModelRspec in the original RequestRspecSource, AND returns a new one (to reset derived rspec forms such as xml string)
         */
        public RequestRspecSource updateExpires(SfaAuthority auth, RequestRspecSource origSource, Instant requestedEndTime) {
            if (origSource != null) {
                if (requestedEndTime != null && auth.hasSpecialCase(SfaAuthority.SpecialCase.ADD_EXPIRES_ATTRIBUTE_TO_REQUEST)) {
                    ModelRspec modelRspec = origSource.getModelRspec();
                    // if (modelRspec.getExpiresString() == null) { //always overwrite see FEDIBBTDEV-1410

                    //don't ever modify the original! The user doesn't expect expires to be added (or deleted)! see FEDIBBTDEV-1448
                    modelRspec = modelRspec.copy();
                    modelRspec.setExpires(Date.from(requestedEndTime));
                    return new RequestRspecSource(modelRspec);
                    // }
                } else {
                    //remove expires
                    ModelRspec modelRspec = origSource.getModelRspec();
                    modelRspec = modelRspec.copy();
                    modelRspec.setExpires((String) null);
                    return new RequestRspecSource(modelRspec);
                }
            }
            return origSource;
        }

        @Override
        public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
            FXModelRspec.RequestRspecSpecialCases rspecSpecialCase = FXModelRspec.RequestRspecSpecialCases.NONE;
            if ((auth.getType() != null && auth.getType().equals("planetlab"))
                    || auth.getUrn().getValue().startsWith("urn:publicid:IDN+ple:ibbtple+authority+")) {
                rspecSpecialCase = FXModelRspec.RequestRspecSpecialCases.PLE;
            }

            AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(taskExecution, auth);

            List<UserSpec> userSpecs = overrideRspec == null ?
                    UserSpecHelper.getMergedUserSpecs(geniUserProvider, easyModel, slice.getRequestRspec().getModelRspec())
                    : UserSpecHelper.getMergedUserSpecs(geniUserProvider, easyModel, overrideRspec);

            if (includeProjectMemberSshKeys)
                userSpecs =
                        UserSpecHelper.mergeUserSpecs(userSpecs, UserSpecHelper.getSliceMemberUserSpecs(slice.getUrn(), easyModel));

            String rspec;
            if (overrideRspec == null) {
                RequestRspecSource requestRspecSource = slice.getRequestRspec(); //ok here
                LOG.trace("HLC.CreateSliverTask request RSpec before modification: expires=" + requestRspecSource.getModelRspec().getExpiresString());
                requestRspecSource = updateExpires(auth, requestRspecSource, expirationTime);
                LOG.debug("HLC.CreateSliverTask request RSpec after modification: expires=" + requestRspecSource.getModelRspec().getExpiresString());
                rspec = requestRspecSource.getRspecXmlString(rspecSpecialCase);
            } else {
                rspec = overrideRspec;
            }

            LOG.info("createSlivers @ {}  with user SSH keys: {}", auth.getUrnString(), easyModel.getUserKeys()); //missing here
            try {
                Instant usedExpirationDate = amWrapper.isExirationdateSetByCreateSliver() ? expirationTime
                        : null; //don't set date here if CreateSliver doesn't set it itself
                if (auth.hasSpecialCase(SfaAuthority.SpecialCase.FORCE_RENEW_AFTER_CREATESLIVER))
                    usedExpirationDate = expirationTime;

                this.manifestRspec = amWrapper.createSliver(
                        slice.getUrn(),
                        slice.getCredentials(),
                        rspec,
                        userSpecs,
                        usedExpirationDate != null ? Date.from(usedExpirationDate) : null
                );
            } catch (AggregateManagerWrapper.NotEnoughFreeResourcesException e) {
                failedDueToNotEnoughFreeResources = true;
                this.manifestRspec = null;
                throw new JFedHighLevelException("Creation of the sliver failed: could not map to resources " +
                        "(resource shortage or unsupported topology)", e);
            }

            //Note: listener that handles createSliver should update the list of component managers, as it will have changed

            //TODO: this might result in multiple slivers! Also we need to get info on them directly, instead of this inderect way
            //      if we do it indirectly like this, we might get slivers that existed before amWrapper.createSliver was called
            sliver = slice.findSliverFromAuthorityUrn(auth.getUrn());

            //either something failed, or the sliver is now being prepared. Either way, we will check the status
            if (this.manifestRspec != null) {
                //try in 10 seconds
                this.getStatusUntilReadyOrFailTask.addTaskToScheduleWhenReady(new ListSliceResourcesTask(auth, slice));

                if (!amWrapper.isExirationdateSetByCreateSliver()) {
                    boolean onlyIfNeeded = true;
                    if (auth.hasSpecialCase(SfaAuthority.SpecialCase.FORCE_RENEW_AFTER_CREATESLIVER))
                        onlyIfNeeded = false; //always call renew
                    RenewSliversTask renewSliversTask =
                            new RenewSliversTask(slice, expirationTime, auth, null/*slivers (null = all)*/, onlyIfNeeded/*only if renew is needed*/);
                    renewSliversTask.addExtraDependency(this);
                    this.getStatusUntilReadyOrFailTask.addTaskToScheduleWhenReady(renewSliversTask);
                }

                tt.scheduleTask(this.getStatusUntilReadyOrFailTask, 10000);

                Set<Sliver> createdSlivers = new HashSet<>();
                Collection<Sliver> manifestSlivers =
                        EasyModelAbstractListener.findSliversInManifest(easyModel, auth, slice.getUrnString(), new ManifestRspecSource(manifestRspec, ModelRspecType.FX));
                //add slivers found by easyModel due to SliverStatus and Status calls. //AL2S is one case where this is neeed, see FEDIBBTDEV-1392
                Set<Sliver> statusSlivers =
                        slice.getSlivers() == null ? Collections.emptySet() :
                                slice.getSlivers().stream().filter(s -> {
                                            if (s.getAuthority() == null || auth == null)
                                                return s.getAuthority() == auth;
                                            else
                                                return s.getAuthority().equals(auth);
                                        }
                                )/*. map(Sliver::getUrn).filter(Objects::nonNull)*/.collect(Collectors.toSet());

                if (manifestSlivers != null) createdSlivers.addAll(manifestSlivers);
                if (statusSlivers != null) createdSlivers.addAll(statusSlivers);

                if (!createdSlivers.isEmpty())
                    registerSliversAtSATaskExecution(createdSlivers, this);

                if (sliver == null)
                    throw new NullPointerException("sliver=null because sliver for authority \"" + auth.getUrn() + "\" could not be found. (Status is being checked despite this error)");
            } else {
                throw new JFedHighLevelException("Creation of the sliver failed.");
            }
        }

        @Override
        public List<Task> initDependsOn() {
            List<Task> res = new ArrayList<>();
            res.add(getSliceCredential(slice));
            res.add(getUserSSHKey());
            if (includeProjectMemberSshKeys)
                res.add(getSliceMemberSSHKeys(slice));
            res.addAll(extraDependencies);
            return res;
        }

        public void addExtraDependency(Task t) {
            extraDependencies.add(t);
        }

        public Slice getSlice() {
            return slice;
        }

        public AuthorityInfo getAuthorityInfo() {
            return ai;
        }

        public SfaAuthority getSfaAuthority() {
            return auth;
        }

        public Sliver getSliver() {
            return sliver;
        }

        public boolean hasFailedDueToNotEnoughFreeResources() {
            return failedDueToNotEnoughFreeResources;
        }

        public GetStatusUntilReadyOrFailTask getGetStatusUntilReadyOrFailTask() {
            return getStatusUntilReadyOrFailTask;
        }
    }

    /**
     * this wil create and execute a Task
     *
     * @param slivers             a collection of slivers, belonging to the same slice, and existing on the same authority
     * @param extraTaskToDependOn a task that the returned task must depend on.
     */
    public TaskExecution registerSliversAtSATaskExecution(Collection<Sliver> slivers, Task extraTaskToDependOn, ResultListener... extraResultListeners) {
        UserAndSliceApiWrapper samaWrapper = getUserAndSliceApiWrapper();
        if (slivers != null && samaWrapper.isRegisterAggregatesForSliceSupported() && !slivers.isEmpty()) {
            Collection<GeniUrn> sliverUrns = Sliver.getUrns(slivers);
            Instant sliverExpirationDate = null;
            Slice slice = null;
            SfaAuthority auth = null;
            for (Sliver sliver : slivers) {
                if (sliver.getExpirationDate() != null)
                    if (sliverExpirationDate == null)
                        sliverExpirationDate = sliver.getExpirationDate();
                    else {
                        if (!sliverExpirationDate.equals(sliver.getExpirationDate()))
                            LOG.warn("The slivers should expire at same time, they were created with single call. " +
                                    "sliverExpirationDate=" + sliverExpirationDate + " sliver.getExpirationDate()=" + sliver.getExpirationDate() + " for " + sliver.getUrnString());

//                        assert sliverExpirationDate.equals(sliver.getExpirationDate()) :
//                                "the slivers should expire at same time, they were created with single call. " +
//                                        "sliverExpirationDate=" + sliverExpirationDate + " sliver.getExpirationDate()=" + sliver.getExpirationDate();
                    }

                if (sliver.getSlice() != null) {
//                    assert slice == sliver.getSlice();
                    assert slice == null || slice.getUrn().equals(sliver.getSlice().getUrn()) :
                            "All slivers provided to this call should be on the same slice";
                    slice = sliver.getSlice();
                }
                if (sliver.getAuthority() != null) {
                    assert auth == null || auth.getUrn().equals(sliver.getAuthority().getUrn()) :
                            "All slivers provided to this call should be on the same authority";
                    auth = sliver.getAuthority();
                }
            }
            assert slice != null;
            assert auth != null;
            RegisterSliversTask registerSliverTask =
                    new RegisterSliversTask(slice, auth, sliverUrns, sliverExpirationDate, extraTaskToDependOn, extraResultListeners);
            return tt.addTask(registerSliverTask);
        }
        return null;
    }

    public final class GetStatusUntilReadyOrFailTask extends Task {
        private static final long PERIOD_MS = 10000l; //10s
        private static final int MAX_FAIL_COUNT = 5;
        private static final int MAX_BUSY_COUNT = 30 * 6; //30 minutes

        private final SfaAuthority auth;
        private final Slice slice;

        private boolean finished = false;
        private int failCount = 0;
        private int busyCount = 0;
        private int triesCount = 0;
        private String status = "";
        private StatusDetails statusDetails;

        private List<Task> tasksToScheduleWhenReady = new ArrayList<>();

        public GetStatusUntilReadyOrFailTask(SfaAuthority auth, Slice slice) {
            super("Status of Sliver @ " + auth.getUrnString());
            this.auth = auth;
            this.slice = slice;
        }

        public void addTaskToScheduleWhenReady(Task t) {
            tasksToScheduleWhenReady.add(t);
        }

        @Override
        public void doTask(TaskExecution singleTask) throws JFedException, InterruptedException {
            triesCount++;
            AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(singleTask, auth);

            boolean busy = false;
            statusDetails = null;
            try {
                statusDetails = amWrapper.status(slice.getUrn(), slice.getCredentials());
            } catch (JFedException ex) {
                if (ex.getGeniResponseCode().isBusy()) {
                    LOG.warn("Got an busy-error while fetching status of sliver. Skipping...");
                    busy = true;
                } else {
                    throw ex;
                }
            }

            if (busy && busyCount > MAX_BUSY_COUNT) {
                LOG.warn("Waited too long for slice to get ready \"" + slice.getUrnString() + "\"." +
                        " (Waited " + busyCount + " times 10 seconds.)");
                this.status = "Waited too long for slice to get ready. Something probably went wrong. Aborting checks.";

                return;
            }

            if (!busy && statusDetails == null) {
                failCount++;

                if (failCount < MAX_FAIL_COUNT) {
                    LOG.warn("Something went wrong while waiting for slice \"" + slice.getUrnString() + "\" to become ready." +
                            " Trying again in 10 seconds...");
                    tt.scheduleTask(this, PERIOD_MS);
                } else {
                    LOG.warn("Something went wrong while waiting for slice \"" + slice.getUrnString() + "\" to become ready." +
                            " Will stop checking status...");
                    this.status = "Fetching status failed. Aborting checks!";
                }
                return;
            }

            boolean atLeastOneFailure = false;
            if (statusDetails != null) {
                for (StatusDetails.SliverStatus sliverStatus : statusDetails.getStatusBySliverUrn().values())
                    if (sliverStatus == StatusDetails.SliverStatus.FAIL)
                        atLeastOneFailure = true;
                for (StatusDetails.SliverStatus sliverStatus : statusDetails.getStatusByComponentUrn().values())
                    if (sliverStatus == StatusDetails.SliverStatus.FAIL)
                        atLeastOneFailure = true;
            }

            //if not fail or ready, schedule again!
            if (busy || statusDetails.getGlobalStatus().equals(StatusDetails.SliverStatus.UNKNOWN) ||
                    statusDetails.getGlobalStatus().equals(StatusDetails.SliverStatus.CHANGING)) {
                if (!atLeastOneFailure || failCount < MAX_FAIL_COUNT) { //bit of a hack
                    busyCount++;
                    LOG.info("Status of Slice {} is \"{}\": will check again in 10 seconds.",
                            slice.getUrnString(), statusDetails);
                    tt.scheduleTask(this, PERIOD_MS);
                } else {
                    failCount++;
                    LOG.warn("At least one sliver has FAILED in \"" + slice.getUrnString() + "\". Retried a few times, but there's no point to keep retrying." +
                            " Will stop checking status...");
                    this.status = "At least one sliver has failed.";
                }
            } else {
                if (atLeastOneFailure) {
                    this.status = "At least one resource has permanently failed.";
                }

                finished = true;
                for (Task t : tasksToScheduleWhenReady)
                    tt.addTask(t);
            }
        }

        @Override
        public List<Task> initDependsOn() {
            return Collections.singletonList(getSliceCredential(slice));
        }

        public long getMsUntilNextCheck() {
            if (!futureTasks.isEmpty()) {
                return futureTasks.get(0).getTimeLeftMs();
            } else
                return 0;

        }

        public int getTriesCount() {
            return triesCount;
        }

        public boolean isFinished() {
            return finished;
        }

        public String getStatus() {
            return status;
        }

        public StatusDetails getStatusDetails() {
            return statusDetails;
        }

        public SfaAuthority getAuthority() {
            return auth;
        }
    }

    public ListSliceResourcesTask describe(SfaAuthority auth, Slice slice) {
        return new ListSliceResourcesTask(auth, slice);
    }

    private final class ListSliceResourcesTask extends Task {
        private final SfaAuthority auth;
        private final Slice slice;

        public ListSliceResourcesTask(SfaAuthority auth, Slice slice) {
            super("ListResources of Slice @ " + auth.getUrnString());
            this.auth = auth;
            this.slice = slice;
        }

        @Override
        protected void doTask(TaskExecution singleTask) throws JFedException, InterruptedException {
            AggregateManagerWrapper amWrapper = getAggregateManagerWrapper(singleTask, auth);
            String rspec = amWrapper.describe(slice.getUrn(), slice.getCredentials());

            for (Sliver sliver : slice.findSlivers(auth)) {
                sliver.setInitializationFinished(true);
            }
        }
    }

    public final class StitchTask extends Task {
        private final Slice slice;
        private final Instant expirationDate;
        private ParallelStitcher parallelStitcher;
        private final SfaAuthority scsAuth;
        private final boolean includeProjectMemberSshKeys;

        private ParallelStitcher.GiveUpStitchingCallback giveUpStitchingCallback;

        private final List<Task> extraDependencies = new ArrayList<>();

        public StitchTask(Slice slice, Instant expirationDate, SfaAuthority scsAuth, boolean includeProjectMemberSshKeys, ParallelStitcher.GiveUpStitchingCallback giveUpStitchingCallback) {
            super("Create Stitched Slivers" + (scsAuth == null ? "" : " with SCS " + scsAuth.getName()));

            assert slice != null;
            assert scsAuth != null;

            this.slice = slice;
            this.scsAuth = scsAuth;
            this.expirationDate = expirationDate;
            this.giveUpStitchingCallback = giveUpStitchingCallback;
            this.includeProjectMemberSshKeys = includeProjectMemberSshKeys;
        }

        @Override
        public void doTask(TaskExecution singleTask) throws JFedException, InterruptedException {
            List<UserSpec> userSpecs =
                    UserSpecHelper.getMergedUserSpecs(geniUserProvider, easyModel, slice.getRequestRspec().getModelRspec());
            if (includeProjectMemberSshKeys)
                userSpecs =
                        UserSpecHelper.mergeUserSpecs(userSpecs, UserSpecHelper.getSliceMemberUserSpecs(slice.getUrn(), easyModel));

            parallelStitcher =
                    new ParallelStitcher(tt, easyModel, geniUserProvider,
                            slice.getRequestRspec(), connectionProvider,
                            scsAuth, HighLevelController.this, userSpecs, expirationDate, authorityList,
                            jFedPreferences, automaticAggregateManagerWrapperFactory,
                            null /* default time between status checks*/,
                            null /* default ms before giving up on status*/,
                            null /* default ms before retry createsliver after delete*/,
                            JavaFXLogger.wrap(logger, singleTask), stitchingReportFactory);
            parallelStitcher.setGiveUpStitchingCallback(this.giveUpStitchingCallback);

            stitchingReportHistory.addStitchingReport(parallelStitcher.getStitchingReport());

            assert slice.hasCredential() :
                    "Must use a slice with a known credential: " + slice + "   all slices:" + easyModel.getSlices();

            UserAndSliceApiWrapper w = getUserAndSliceApiWrapper();
            if (!w.hasUserCredentials())
                throw new JFedException("UserAndSliceApiWrapper did not have user Credentials.");

            parallelStitcher.setUserCredentialAndSlice(w.getCachedUserCredentialsForAM(), slice);

            for (Runnable r : preStichStartCallbacks)
                r.run();

            parallelStitcher.start();

            LOG.debug("StitchTask: stitching started. Will wait up to 2 hours for stitching to finish");
            parallelStitcher.waitUntilDone(2, TimeUnit.HOURS); //wait up to 2 hours for stitching to finish

            File jFedStitchLogDir = new File(JFedUtils.getUserDataDirectory() + File.separator + "stitch-log");
            if (!jFedStitchLogDir.exists())
                jFedStitchLogDir.mkdirs();

            SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd_hh-mm-ss");
            String logName = "stitch-" + dateFormat.format(new Date()) + ".html";
            File logFile = new File(jFedStitchLogDir, logName);
            LOG.debug("StitchTask: stitching finished! Will write report to: " + logFile.getPath());
            parallelStitcher.getStitchingReport().toHtml(logFile, false);
        }

        @Override
        public List<Task> initDependsOn() {
            List<Task> res = new ArrayList<>();
            res.add(getSliceCredential(slice));
            res.add(getUserSSHKey());
            if (includeProjectMemberSshKeys)
                res.add(getSliceMemberSSHKeys(slice));
            res.addAll(extraDependencies);
            return res;
        }

        public void addExtraDependency(Task t) {
            extraDependencies.add(t);
        }

        public Slice getSlice() {
            return slice;
        }

        public ParallelStitcher getParallelStitcher() {
            return parallelStitcher;
        }

        private List<Runnable> preStichStartCallbacks = new ArrayList<>();

        public void addPreStichStartCallback(Runnable r) {
            preStichStartCallbacks.add(r);
        }
    }

    public class ListSlicesTask extends Task {
        private List<GeniUrn> slices;

        public ListSlicesTask() {
            super("List Slices");
        }

        @Override
        public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
            UserAndSliceApiWrapper w = getUserAndSliceApiWrapper();
            slices = w.getSlicesForUser(JavaFXLogger.wrap(logger, taskExecution),
                    geniUserProvider.getLoggedInGeniUser().getUserUrn());
        }

        @Override
        public List<Task> initDependsOn() {
            List<Task> res = new ArrayList<>();
            res.add(getUserCredential());
            return res;
        }


        public List<GeniUrn> getSlices() {
            return slices;
        }
    }


    /**
     * Convenience method that submits the task into the taskthread
     *
     * @param task
     * @param callbacks
     */
    public TaskExecution submitTask(Task task, TaskExecutionFinishedCallback... callbacks) {
        return tt.addTask(task, callbacks);
    }

    public class SliceMembersSshKeysTask extends Task {
        private final Slice slice;
        private final GeniUrn sliceUrn;

        private List<UserSpec> userSpecs = null;

        public SliceMembersSshKeysTask(@Nonnull Slice slice) {
            super("Get User SSH Keys For users in slice " + slice.getUrnString());
            this.slice = slice;
            this.sliceUrn = slice.getUrn();
            assert sliceUrn != null;
        }

        @Override
        public void doTask(TaskExecution taskExecution) throws JFedException, InterruptedException {
            List<UserSpec> result = new ArrayList<>();

            UserAndSliceApiWrapper w = getUserAndSliceApiWrapper();
            List<GeniUrn> userUrns =
                    w.getUsersForSlice(JavaFXLogger.wrap(logger, taskExecution), sliceUrn);
            if (userUrns == null) userUrns = Collections.emptyList();
            for (GeniUrn userUrn : userUrns) {
                assert (userUrn != null);
                try {
                    List<String> userKeys = w.getSshKeysForUser(JavaFXLogger.wrap(logger, taskExecution), userUrn);
                    if (userKeys != null)
                        result.add(new UserSpec(userUrn.toString(), userKeys));
                } catch (JFedException e) {
                    LOG.warn("Could not get user key for " + userUrn, e);
                }
            }

            userSpecs = result;
        }

        @Override
        public List<Task> initDependsOn() {
            List<Task> res = new ArrayList<>();
            res.add(getSliceCredential(slice));
            return res;
        }

        public List<UserSpec> getUserSpecs() {
            return userSpecs;
        }
    }
}

