/*
 * Copyright 2013 SmartBear Software
 * 
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by the European Commission - subsequent
 * versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * 
 * http://ec.europa.eu/idabc/eupl
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the Licence is
 * distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
 * express or implied. See the Licence for the specific language governing permissions and limitations
 * under the Licence.
 */
package org.guavafx;

import com.google.common.base.Function;
import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import com.google.common.collect.Iterables;
import com.google.common.collect.Lists;
import javafx.beans.InvalidationListener;
import javafx.beans.Observable;
import javafx.beans.WeakInvalidationListener;
import javafx.collections.FXCollections;
import javafx.collections.ListChangeListener;
import javafx.collections.ObservableList;
import javafx.collections.WeakListChangeListener;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Set;

/**
 * Utility class for dealing with JavaFX ObservableLists.
 *
 * @author dain.nilsson
 * @author henrik.olsson
 */
public class ObservableLists {
    /**
     * Creates a readonly ObservableList containing all OSGi published services
     * for the given Class type. The list is dynamically updated to reflect
     * changes in the services.
     *
     * @param type
     * @return
     */

    /**
     * Creates an ObservableList of transformed elements from a given list.
     *
     * @param original
     * @param function
     * @return
     */
    public static <F, T> ObservableList<T> transform(final ObservableList<F> original, final Function<F, T> function) {
        final ListeningList<F, T> listeningList = new ListeningList<>(original);

        final LoadingCache<F, T> cache =
                CacheBuilder.newBuilder().weakKeys().softValues().build(new CacheLoader<F, T>() {
                    @Override
                    public T load(F key) throws Exception {
                        return function.apply(key);
                    }
                });

        final InvalidationListener listener = new InvalidationListener() {
            @Override
            public void invalidated(Observable o) {
                listeningList.list.setAll(Lists.newArrayList(Iterables.transform(original, cache)));
            }
        };
        listeningList.addListener(listener);
        listener.invalidated(original);

        return listeningList.readOnlyList;
    }

    @SuppressWarnings("serial")
    private static class ListeningList<F, T> extends ArrayList<T> {
        private final ObservableList<? extends F> originalList;
        private final ObservableList<T> list;
        private final ObservableList<T> readOnlyList;
        private final Set<Object> hardrefs = new HashSet<>();

        private ListeningList(ObservableList<? extends F> originalList) {
            this.originalList = originalList;
            list = FXCollections.observableList(this);
            readOnlyList = FXCollections.unmodifiableObservableList(list);
        }

        public ListeningList<F, T> addListener(InvalidationListener listener) {
            originalList.addListener(new WeakInvalidationListener(listener));
            hardrefs.add(listener);

            return this;
        }

        public ListeningList<F, T> addListener(ListChangeListener<F> listener) {
            originalList.addListener(new WeakListChangeListener<>(listener));
            hardrefs.add(listener);

            return this;
        }
    }
}
