package be.iminds.ilabt.jfed.lowlevel;

import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.IOUtils;
import be.iminds.ilabt.jfed.util.KeyUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.cert.CertificateParsingException;
import java.security.cert.X509Certificate;
import java.util.Collection;
import java.util.List;

/**
 * SimpleGeniUser
 */
public class SimpleGeniUser implements GeniUser {
    private final static Logger LOG = LoggerFactory.getLogger(SimpleGeniUser.class);

    private List<X509Certificate> certificateChain;
    private PrivateKey privateKey;

    private SfaAuthority authority;
    private GeniUrn userUrn;

    private File certificateKeyFile;
    private File privateKeyFile;

    private String keyCertContent; //may be null, in case the file needs to be read

    //    public SimpleGeniUser(SfaAuthority userAuthority, String username, String keyCertContent, char[] keyPass) {
//        this(userAuthority, "urn:publicid:IDN+"+userAuthority.getNameForUrn()+"+user+"+username, keyCertContent, keyPass);
//    }
    public SimpleGeniUser(SfaAuthority userAuthority, GeniUrn userUrn, String keyCertContent, char[] keyPass, File certificateKeyFile, File privateKeyFile) {
        this.authority = userAuthority;
        this.userUrn = userUrn;
        this.certificateKeyFile = certificateKeyFile;
        this.privateKeyFile = privateKeyFile;
        this.keyCertContent = keyCertContent;

        try {
            privateKey = KeyUtil.pemToAnyPrivateKey(keyCertContent, keyPass);
        } catch (KeyUtil.PEMDecodingException e) {
            privateKey = null;
            throw new RuntimeException("ERROR reading PEM key:" + keyCertContent + " -> " + e.getMessage(), e);
        }
        if (privateKey == null)
            throw new RuntimeException("ERROR: PEM key and certificate does not contain a key:" + keyCertContent);

        certificateChain = KeyUtil.pemToX509CertificateChain(keyCertContent);
        if (certificateChain == null || certificateChain.isEmpty())
            throw new RuntimeException("ERROR: PEM key and certificate does not contain a X509 certificate:" + keyCertContent);
    }

    public SimpleGeniUser(SfaAuthority userAuthority, GeniUrn userUrn, List<X509Certificate> certificateChain, PrivateKey privateKey, File certificateKeyFile, File privateKeyFile) {
        this.authority = userAuthority;
        this.userUrn = userUrn;
        this.privateKey = privateKey;
        this.certificateChain = certificateChain;
        this.certificateKeyFile = certificateKeyFile;
        this.privateKeyFile = privateKeyFile;
    }

    /**
     * make a copy of a GeniUser
     */
    public SimpleGeniUser(GeniUser geniUser) {
        this.certificateChain = geniUser.getClientCertificateChain();
        this.privateKey = geniUser.getPrivateKey();
        this.authority = geniUser.getUserAuthority();
//        this.userName = geniUser.getUserName();
        this.userUrn = GeniUrn.parse(geniUser.getUserUrnString());

        this.certificateKeyFile = geniUser.getCertificateFile();
        this.privateKeyFile = geniUser.getPrivateKeyFile();
    }

    /**
     * make a copy of a GeniUser
     */
    public SimpleGeniUser(SimpleGeniUser geniUser) {
        this.certificateChain = geniUser.certificateChain;
        this.privateKey = geniUser.privateKey;
        this.authority = geniUser.authority;
        this.userUrn = geniUser.userUrn;

        this.certificateKeyFile = geniUser.certificateKeyFile;
        this.privateKeyFile = geniUser.privateKeyFile;
    }


    @Override
    public List<X509Certificate> getClientCertificateChain() {
        return certificateChain;
    }

    @Override
    public PrivateKey getPrivateKey() {
        return privateKey;
    }

    @Override
    public PublicKey getPublicKey() {
        List<X509Certificate> certificateChain = getClientCertificateChain();
        if (certificateChain == null || certificateChain.isEmpty())
            return null;
        return certificateChain.get(0).getPublicKey();
    }

    @Override
    public SfaAuthority getUserAuthority() {
        return authority;
    }

//    @Override
//    public String getUserName() {
//        return userName;
//    }

    @Override
    public GeniUrn getUserUrn() {
        return userUrn;
    }

    @Override
    public File getPrivateKeyFile() {
        return privateKeyFile;
    }

    @Override
    public File getCertificateFile() {
        return certificateKeyFile;
    }

    @Override
    public String getUserUrnString() {
        return userUrn.toString();
    }


    /**
     * doesn't do anything when it fails anywhere
     */
    public void deriveUserAndAuthFromPemContent(AuthorityFinder optionalAuthorityFinder) {
        String privateKeyFileContent = keyCertContent;

        if (privateKeyFileContent == null)
            try {
                privateKeyFileContent = IOUtils.fileToString(privateKeyFile);
            } catch (IOException e) {
                return;
            }
        if (privateKeyFileContent == null) return;


        if (certificateChain == null || certificateChain.isEmpty()) return;

        Collection<List<?>> altNames = null;
        try {
            X509Certificate firstCertificate = certificateChain.get(0);
            altNames = firstCertificate.getSubjectAlternativeNames();
        } catch (CertificateParsingException e) {
            return;
        }

//        String userAuth = null;

        if (altNames != null) {
            boolean hasvalidUrn = false;
            for (List<?> altName : altNames) {
                Integer nameType = (Integer) altName.get(0);
                if (nameType == 6) {
                    //uniformResourceIdentifier
                    String urn = (String) altName.get(1);
                    GeniUrn geniUrn = GeniUrn.parse(urn);
                    if (geniUrn == null || !geniUrn.getEncodedResourceType().equals("user")) {

                    } else {
                        hasvalidUrn = true;
//                        userAuth = geniUrn.getTopLevelAuthority();
                        String userName = geniUrn.getEncodedResourceName();
                        if (userUrn == null) //don't overwrite
                            userUrn = geniUrn;
                        else {
                            if (!userUrn.equals(geniUrn))
                                LOG.error("deriveUserAndAuthFromPemContent-> derived userurn \""+geniUrn+"\" differs from already known userUrn \""+userUrn+"\"");
                        }
                    }
                }
            }
        }

        if (userUrn != null && optionalAuthorityFinder != null && authority == null) {
            authority = optionalAuthorityFinder.findByUrn(userUrn, AuthorityFinder.Purpose.FIND_USERAUTH);
        }
    }
}
