package be.iminds.ilabt.jfed.lowlevel.api_wrapper;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.SfaConnection;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.GeniUrn;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.List;

/**
 * UserAndSliceApiWrapper wraps user and slice CRU calls.
 * <p/>
 * It stores user the credentials needed internally and caches them.
 * The reason for this is that there might be multiple types of credentials, depending on the API.
 * <p/>
 * Slice credentials are not stored. (might change in future version, but no reason to do that yet)
 * <p/>
 * This an currently wrap:
 * - "ProtoGeni Slice Authority API v1"
 * - "Uniform Federation Slice Authority API draft"+"Uniform Federation Member Authority API draft"
 * - Planetlab SfaRegistryInterface
 * <p/>
 * Some calls might result in multiple calls.
 * <p/>
 * Most call returns some "high-level" useful result.
 * But it is also assumes that users of this API process the call results themselves as well if they need more details. (jFed does this in the highlevel-library)
 */
public abstract class UserAndSliceApiWrapper
        extends ApiWrapper {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(UserAndSliceApiWrapper.class);

    /**
     * Note: logger is not a parameter with a very good reason!
     * <p/>
     * if it was, it would be a problem to make parallel calls.
     * The reason is that the UserAndSliceApiWrapper stores credential data, so it cannot be recreated each time its used, like the AMWrapper
     * So the logger has to be changed for each call, but that is hard if the logger is stored in the wrapper.
     * Solution: each call needs the specific logger. (this is also the method least likely to cause bugs due to confusion)
     */
    public UserAndSliceApiWrapper(Logger logger, GeniUserProvider geniUserProvider, JFedConnectionProvider connectionProvider, JFedPreferences jFedPreferences) {
        super(logger, geniUserProvider, connectionProvider, jFedPreferences);
    }


    protected SfaConnection getConnection(ServerType serverType) throws JFedException {
        GeniUser user = geniUserProvider.getLoggedInGeniUser();
        SfaAuthority userAuth = user.getUserAuthority();
        return (SfaConnection) connectionProvider.getConnectionByAuthority(user, userAuth, serverType);
    }

    protected GeniUser getUser() {
        return geniUserProvider.getLoggedInGeniUser();
    }

    protected SfaAuthority getUserAuthority() {
        GeniUser user = geniUserProvider.getLoggedInGeniUser();
        SfaAuthority userAuth = user.getUserAuthority();
        return userAuth;
    }

    public List<AnyCredential> getLoggedinUserCredentials(Logger logger)
            throws JFedException {
        return getUserCredentials(logger, getUser().getUserUrn());
    }

    /**
     * @returns null or empty list if no credentials available
     */
    public abstract List<AnyCredential> getCachedUserCredentialsForAM();

    public abstract boolean hasUserCredentials();

    protected boolean hasUserCredentials(List<AnyCredential> userCredentials) {
        if (userCredentials == null || userCredentials.isEmpty()) {
            return false; //has no valid credential
        }

        boolean hasUnknownCredential = false;
        boolean hasValidCredential = false;
        boolean hasExpiredCredential = false;
        List<AnyCredential> toRemove = new ArrayList<AnyCredential>();
        for (AnyCredential existingCredential : userCredentials) {
            if (existingCredential.getExpiresDate() == null) {
                LOG.warn("UserAndSliceApiWrapper could not determine an expirationDate for the current credential. We assume it is still valid.");
                hasUnknownCredential = true; //has a valid credential, with unknown expire time
            } else {
                //don't compare with now, compare with 2 minutes from now, since credential that expires soon is as good as useless.
                Date now = new Date(System.currentTimeMillis() + 2*60*1000);
                if (existingCredential.getExpiresDate().before(now)) {
                    hasExpiredCredential = true; //has expired credential
                    LOG.info("UserAndSliceApiWrapper.hasUserCredentials encountered an expired credential");
                    toRemove.add(existingCredential);
                } else {
                    hasValidCredential = true; //has a valid credential
                }
            }
        }

        //remove expired just to be sure
        userCredentials.removeAll(toRemove);

        if (hasUnknownCredential && hasExpiredCredential)
            LOG.warn("UserAndSliceApiWrapper.hasUserCredentials encountered a credential with unknown expire time, and one with expired time. Will have to assume there IS a valid credential.");

        if (hasValidCredential) return true; //if at least one is valid and not expired, that's enough
        if (hasUnknownCredential) return true; //TODO because this is true, old credentials might be used. Just implement getExpiresDate for each credential type to make it work!
        if (hasExpiredCredential) return false;
        //has no valid credential at all
        return false;
    }

    /**
     * get one or more credentials for the currently logged in user
     */
    public abstract List<AnyCredential> getUserCredentials(Logger logger, GeniUrn userUrn) throws JFedException;

    /**
     * get one or more credentials for the given slice
     */
    public abstract List<AnyCredential> getSliceCredentials(Logger logger, GeniUrn sliceUrn) throws JFedException;

    /**
     * get the list of slices for the currently given user (typically the logged in user)
     */
    public abstract List<GeniUrn> getSlicesForUser(Logger logger, GeniUrn userUrn) throws JFedException;

    /**
     * get info about the slivers the slice is running on.
     * <p/>
     * This call does not return anything!  (it is assumed this is used from the "HighLevelController", while you are processing the results asynchronously)
     */
    public abstract List<String> getAggregatesForSlice(Logger logger, List<AnyCredential> sliceCredentials, GeniUrn sliceUrn) throws JFedException;

    public abstract boolean isRegisterAggregatesForSliceSupported();

    /**
     * @param logger
     * @param sliceCredentials
     * @param sliceUrn
     * @param aggregateUrn aggregate on which the slivers exist (use multiple calls if slivers on multiple aggregates)
     * @param sliverUrns list of slivers on same aggregate in same slice, with same expiration date
     * @param sliverExpirationDate expiration date of all slivers  (use multiple calls if slivers have different expiration dates)
     * @throws JFedException
     */
    public abstract void registerAggregatesForSlice(Logger logger, List<AnyCredential> sliceCredentials, GeniUrn sliceUrn, GeniUrn aggregateUrn, Collection<GeniUrn> sliverUrns, Date sliverExpirationDate) throws JFedException;
    public abstract void updateSliverExpirationDateInfoForSlice(Logger logger, List<AnyCredential> sliceCredentials, GeniUrn sliceUrn, Collection<GeniUrn> sliverUrns, Date sliverExpirationDate) throws JFedException;
    public abstract void unregisterAggregatesForSlice(Logger logger, List<AnyCredential> sliceCredentials, GeniUrn sliceUrn, GeniUrn aggregateUrn, Collection<GeniUrn> sliverUrns) throws JFedException;

    /**
     * create a new slice with the given name
     */
    public SliceInfo createSlice(Logger logger, String sliceName, String subAuthName) throws JFedException {
        return createSlice(logger, sliceName, null, subAuthName);
    }

    /**
     * create a new slice with the given name
     * <p/>
     * This call does not return anything!  (it is assumed this is used from the "HighLevelController", while you are processing the results asynchronously)
     */
    public abstract SliceInfo createSlice(Logger logger, String sliceName, Date expirationDate, String subAuthName) throws JFedException;

    /**
     * Renew the slice and set the expiration time to newExpirationDate
     *
     * @param sliceCredentials
     * @param newExpirationDate
     */
    public abstract List<AnyCredential> renewSlice(Logger logger, List<AnyCredential> sliceCredentials, Date newExpirationDate) throws JFedException;

    /**
     * get all SSH keys for the given user (typically the logged in user)
     * <p/>
     * This call does not return anything!  (it is assumed this is used from the "HighLevelController", while you are processing the results asynchronously)
     */
    public abstract List<String> getSshKeysForUser(Logger logger, GeniUrn userUrn) throws JFedException;

    public abstract SubAuthoritySupport getSubAuthoritySupport(Logger logger) throws JFedException;

    public abstract List<String> getSubAuthorityNames(Logger logger, GeniUrn user) throws JFedException;


    public abstract void shareSlice(Logger logger, GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, GeniUrn userUrn) throws JFedException;

    public abstract List<GeniUrn> getUsersForSubAuthority(Logger logger, String projectName) throws JFedException;

    public abstract List<GeniUrn> getUsersForSlice(Logger logger, GeniUrn sliceUrn) throws JFedException;

    public abstract void addUsersToSlice(Logger logger, GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, List<GeniUrn> userUrns, String role) throws JFedException;

    /*
         * sub authority support
         *`
         * Some servers/APIs require a subauthority, while others offer the option,a nd other forbid it.
         *
         * What this API hides:
        * For some API's (Protogeni SA),using a subauthority connects to a different URL
        * For others, it connects to the same URL
        * */
    public enum SubAuthoritySupport {
        SUB_AUTHORITY_MANDATORY, SUB_AUTHORITY_OPTIONAL, SUB_AUTHORITY_FORBIDDEN
    }

    public static class SliceInfo {
        private final String name;
        private final GeniUrn urn;
        private final List<AnyCredential> credentials;

        public SliceInfo(String name, GeniUrn urn, List<AnyCredential> credentials) {
            assert name != null;
            assert urn != null;
            assert credentials != null;
            assert !credentials.isEmpty();

            this.name = name;
            this.urn = urn;
            this.credentials = new ArrayList<AnyCredential>(credentials);
        }

        public String getName() {
            return name;
        }

        public GeniUrn getUrn() {
            return urn;
        }

        public List<AnyCredential> getCredentials() {
            return credentials;
        }
    }
}
