package be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.api.AbstractGeniAggregateManager;
import be.iminds.ilabt.jfed.lowlevel.api.AggregateManager2;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.ErrorDetails;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.StatusDetails;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.SfaConnection;
import be.iminds.ilabt.jfed.lowlevel.stitching.StitchingDirector;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.GeniUrn;
import org.slf4j.LoggerFactory;

import java.net.URL;
import java.util.Date;
import java.util.Hashtable;
import java.util.List;

/**
 * AMv2Wrapper
 */
public class AMv2Wrapper extends AggregateManagerWrapper {
    private final static org.slf4j.Logger LOG = LoggerFactory.getLogger(AMv2Wrapper.class);
    private AggregateManager2 am2;

    public AMv2Wrapper(Logger logger, GeniUserProvider geniUserProvider, JFedConnectionProvider connectionProvider, JFedPreferences jFedPreferences, SfaAuthority amAuthority) {
        super(logger, geniUserProvider, connectionProvider, jFedPreferences, amAuthority);

        am2 = new AggregateManager2(logger, jFedPreferences);
    }



    protected SfaConnection getConnection() throws JFedException {
        return (SfaConnection) getConnection(new ServerType(ServerType.GeniServerRole.AM, 2));
    }

    @Override
    public Hashtable getVersion() throws JFedException {
        AbstractGeniAggregateManager.AggregateManagerReply<AggregateManager2.VersionInfo> reply =
                am2.getVersion(getConnection());
        lastAmReply = reply;

        if (!reply.getGeniResponseCode().isSuccess()) {
            throw new JFedException("GetVersion Call not successful", reply);
        }

        return reply.getRawResult();
    }

    @Override
    public String listResources(List<AnyCredential> userCredentials, boolean available, String type, String version) throws JFedException {
        AbstractGeniAggregateManager.AggregateManagerReply<String> reply = am2.listResources(
                getConnection(),
                filterCredentials(userCredentials),
                type, version,
                available, true/*compressed*/,
                null/*sliceUrn*/, null/*extra_options*/);
        lastAmReply = reply;

        if (!reply.getGeniResponseCode().isSuccess()) {
            throw new JFedException("ListResources Call not successful", reply);
//            return null;
        }

        return reply.getValue();
    }

    @Override
    public String allocate(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String rspec, Date startDate, Date endDate) throws JFedException {
       throw new UnsupportedOperationException("AMv2 does not support reservations");

    }

    @Override
    public String createSliver(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String rspec, List<UserSpec> users, Date expirationDate)
            throws JFedException {
        AbstractGeniAggregateManager.AggregateManagerReply<String> reply = am2.createSliver(
                getConnection(),
                filterCredentials(sliceCredentials),
                sliceUrn.getValue(), rspec,
                users, null/*extra_options*/);

        if (reply.getGeniResponseCode().isSuccess() && reply.getValue()  == null) {
            throw new JFedException("CreateSliver Call has null reply value", reply);
        }
        if (reply.getGeniResponseCode().isSuccess() && !(reply.getValue() instanceof String)) {
            throw new JFedException("CreateSliver Call has invalid reply type: "+reply.getValue().getClass().getName(), reply);
        }

        String manifestRspec = reply.getValue();
        lastAmReply = reply;

        if (!reply.getGeniResponseCode().isSuccess()) {
            if (AbstractGeniAggregateManager.isNotEnoughFreeResources(reply))
                throw new NotEnoughFreeResourcesException(reply);

            if (StitchingDirector.isAnyNotSupported(reply))
                throw new AnyVlanUnsupportedException(reply);

            if (StitchingDirector.isUnavailableVlan(reply))
                throw new VlanUnavailableException(StitchingDirector.getUnavailableVlan(reply), reply);

            boolean ignoreError = false;
            if (reply.getOutput().equalsIgnoreCase("Must delete existing slice first")) {
                //if this occurs, the CreateSliver call has typically timed out but still succeeded. The next call then succeeded but with a "Must delete existing slice first" error.
                //we can recover by requesting the manifest and returning it
                manifestRspec = describe(sliceUrn, sliceCredentials);
                ignoreError = true;
            }

            if (!ignoreError)
                throw new JFedException("CreateSliver Call not successful", reply);
        }

        if (expirationDate != null) {
            //AMv2 doesn't support setting the expirationDate on creation, so we now forcefully execute an renewSliver

            //no long wait needed, HighLevelController that calls this now schedules renew itself
            //   if it is still specified, we do it immediately

//            //but we wait first, since it will always be busy if you try immediately.
//            try {
//                Thread.sleep(30 * 1000l);
//            } catch (InterruptedException e) { /*ignore, doesn't matter if sleep worked or not*/ }

            try {
                renewSlivers(sliceUrn, null/*sliverUrns*/, filterCredentials(sliceCredentials), expirationDate);
            } catch (JFedException ex) {
                LOG.error("Error when trying to define expiration date for newly created sliver", ex);
            }
            lastAmReply = reply; //HACK set to last CreateSliver reply!
        }

        return manifestRspec;
    }

    @Override
    public boolean hasSeperateAllocateAndProvision() {
        return false;
    }

    @Override
    public String allocate(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String rspec, Date expirationDate) throws JFedException {
        throw new RuntimeException("Unsupported operation");
    }

    @Override
    public String provisionAndStart(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, List<UserSpec> userSpecs, Date expirationDate) throws JFedException {
        throw new RuntimeException("Unsupported operation");
    }

    @Override
    public void deleteSliver(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        AbstractGeniAggregateManager.AggregateManagerReply<Boolean> reply = am2.deleteSliver(
                getConnection(),
                filterCredentials(sliceCredentials),
                sliceUrn.getValue(), null/*extra_options*/);
        lastAmReply = reply;

        if (!reply.getGeniResponseCode().isSuccess()) {
            if (reply.getGeniResponseCode().equals(GeniAMResponseCode.GENIRESPONSE_SEARCHFAILED))
                return; //not an error, the sliver was already deleted, so all is ok.

            throw new JFedException("Delete Call not successful", reply);
            //            return null;
        }
    }

    @Override
    public StatusDetails status(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        AbstractGeniAggregateManager.AggregateManagerReply<AggregateManager2.SliverStatus> reply = am2.sliverStatus(
                getConnection(),
                filterCredentials(sliceCredentials),
                sliceUrn.getValue(), null/*extra_options*/);
        lastAmReply = reply;

        if (reply.getGeniResponseCode() == GeniAMResponseCode.GENIRESPONSE_SEARCHFAILED) {
            //the AM could not find any slivers associated to the given slice
            return new StatusDetails(StatusDetails.SliverStatus.UNALLOCATED);
        }

        AggregateManager2.SliverStatus status = reply.getValue();

        if ((!reply.getGeniResponseCode().isSuccess())) {
            throw new JFedException("SliverStatus Call not successful", reply);
        }
        if (status == null) {
            throw new JFedException("SliverStatus Call has no return value", reply);
        }

        if (status.getStatus().equals("ready") && status.getResources().isEmpty()) {
            throw new JFedException("SliverStatus Call returned \"ready\", but there are no resources. This is probably a server error.", reply);
        }

        //convert AggregateManager2.SliverStatus to SliverStatusDetails
        StatusDetails res = new StatusDetails(status.getStatus());

        for (AggregateManager2.SliverStatus.ResourceStatus resourceStatus : status.getResources()) {
            if (resourceStatus.getError() != null && !resourceStatus.getError().isEmpty()) {
                res.addError(new ErrorDetails(resourceStatus.getError(), null, resourceStatus.getUrn()));
            }
            res.addSliverStatus(resourceStatus.getUrn(), resourceStatus.getStatus());
        }


        return res;
    }

    @Override
    public String describe(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        String type = "geni";
        String version = "3";
        if (amAuthority != null && amAuthority.getType() != null && amAuthority.getType().equalsIgnoreCase("FOAM"))
            type = "GENI"; //see FEDIBBTDEV-1392: workaround for bug at al2s (it should ignore case but doesn't)

        AbstractGeniAggregateManager.AggregateManagerReply<String> reply = am2.listResources(
                getConnection(),
                filterCredentials(sliceCredentials),
                type, version,
                null /*available*/, true/*compressed*/,
                sliceUrn.getValue(), null/*extra_options*/);
        lastAmReply = reply;

        if (!reply.getGeniResponseCode().isSuccess()) {
            throw new JFedException("ListResources Call not successful", reply);
            //            return null;
        }

        return reply.getValue();
    }

    @Override
    public void renewSlivers(GeniUrn sliceUrn, List<GeniUrn> sliverUrns, List<AnyCredential> sliceCredentials, Date newExpire) throws JFedException {
        if (sliceUrn == null)
            throw new JFedException("RenewSliver not called, because sliceUrn == null and this call supports only slice URN. (sliverUrns=" + sliverUrns + ")");

        AbstractGeniAggregateManager.AggregateManagerReply<Boolean> reply = am2.renewSliver(
                getConnection(),
                filterCredentials(sliceCredentials),
                sliceUrn.getValue(), newExpire, null/*extra_options*/);
        lastAmReply = reply;

        if (!reply.getGeniResponseCode().isSuccess()) {
            throw new JFedException("RenewSliver Call not successful", reply);
//            return null;
        }
    }

    @Override
    public boolean isExirationdateSetByCreateSliver() {
        return false;
    }


    @Override
    public boolean isShareLanSupported() {
        return false;
    }

    @Override
    public void shareLan(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String lanClientId, String newSharedLanName) {
        throw new RuntimeException("Unsupported operation");
    }

    @Override
    public void unshareLan(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, String lanClientId, String sharedLanName) {
        throw new RuntimeException("Unsupported operation");
    }


    @Override
    public boolean isRestartSupported() {
        return false;
    }

    @Override
    public boolean isReloadOSSupported() {
        return false;
    }

    @Override
    public boolean isOpenConsoleSupported() {
        return false;
    }

    @Override
    public void restart(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        throw new RuntimeException("Unsupported operation");
    }

    @Override
    public void reloadOS(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        throw new RuntimeException("Unsupported operation");
    }

    @Override
    public URL openConsole(GeniUrn sliverUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        throw new RuntimeException("Unsupported operation");
    }


    @Override
    public boolean isEditSshKeysSupported() {
        return false;
    }

    @Override
    public void editSshKeys(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, List<UserSpec> userSpecs) throws JFedException {
        throw new RuntimeException("Unsupported operation");
    }

    @Override
    public boolean areUserDiskImageFunctionsSupported() {
        return false;
    }

    @Override
    public ImageInfo createImage(GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, GeniUrn componentUrn, String imageName) throws JFedException {
        throw new RuntimeException("Unsupported operation");
    }
}
