package be.iminds.ilabt.jfed.lowlevel.authority;

import be.iminds.ilabt.jfed.preferences.CorePreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.GeniUrn;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;

/**
 * AuthorityFinder
 */
public class AuthorityFinder {
    private static final Logger LOG = LoggerFactory.getLogger(AuthorityFinder.class);

    private final AuthorityListModel model;
    private final JFedPreferences jFedPreferences;

    @Inject
    public AuthorityFinder(AuthorityListModel model, JFedPreferences jFedPreferences) {
        this.model = model;
        this.jFedPreferences = jFedPreferences;
    }

    public static enum Purpose { CREATE_SLIVER, STITCHING, STATUS, DELETE, RECOVER, RECOVER_BOTH_FIRST, RECOVER_BOTH_SECOND, REQUEST_RSPEC, FIND_USERAUTH };
    private static enum ExogeniChoice { FORCE_TOP, FORCE_SUBAUTH, USE_GLOBAL_SETTING/*typically this is preference in testbeds.xml*/ };

    public SfaAuthority findByAnyUrn(String urnString, Purpose purpose) {
        return findByUrn(urnString, purpose, true/*ignoreTypeAndName*/);
    }
    public SfaAuthority findByUrn(String urnString, Purpose purpose) {
        return findByUrn(urnString, purpose, false/*ignoreTypeAndName*/);
    }
    public SfaAuthority findByUrn(String urnString, Purpose purpose, boolean ignoreTypeAndName) {
        GeniUrn u = GeniUrn.parse(urnString);
        if (u == null) {
            LOG.warn("Tried to lookup SfaAuthority using the invalid URN \"" + urnString + "\" -> solving problem by returning null");
            return null;
        }
        return findByUrn(u, purpose, ignoreTypeAndName);
    }
    public SfaAuthority findByUrn(GeniUrn urn, Purpose purpose) {
        return findByUrn(urn, purpose, false/*ignoreTypeAndName*/);
    }
    public SfaAuthority findByAnyUrn(GeniUrn urn, Purpose purpose) {
        return findByUrn(urn, purpose, true/*ignoreTypeAndName*/);
    }
    public SfaAuthority findByUrn(GeniUrn urn, Purpose purpose, boolean ignoreTypeAndName) {
        String prefenceForceExoSM = jFedPreferences.getString(CorePreferenceKey.PREF_TESTBEDSPECIFIC_EXOGENI_FORCE_EXOSM).trim().toLowerCase(); //AUTO, YES or NO
        String prefenceRecoverExoSM = jFedPreferences.getString(CorePreferenceKey.PREF_TESTBEDSPECIFIC_EXOGENI_RECOVER_METHOD).trim().toLowerCase(); //BOTH or SINGLE

        AuthorityListModel.SubAuthMatchAllowed allowed;
        AuthorityListModel.SubAuthMatchPreference preferred;
        ExogeniChoice exogeniChoice;
        switch (purpose) {
//            case CREATE_SLIVER: {
//                allowed = AuthorityListModel.SubAuthMatchAllowed.ALLOW_TOPLEVEL;
//                exogeniChoice = ExogeniChoice.USE_GLOBAL_SETTING;
//                break;
//            }
            case STITCHING: {
                //Basically, the SCS chooses, we can't be flexible!!

                //this is only used in one case. See StichngData for details on how AM is chosen when stitching.
//                allowed = AuthorityListModel.SubAuthMatchAllowed.ALLOW_TOPLEVEL;
//                exogeniChoice = ExogeniChoice.FORCE_TOP;
                allowed = AuthorityListModel.SubAuthMatchAllowed.ALLOW_ONLY_EXACT_SUBAUTHORITY;
                exogeniChoice = null; //nothing special for exogeni, so prefer exact
                break;
            }
            case RECOVER:
                //assert prefenceRecoverExoSM.equalsIgnoreCase("single");
                LOG.warn("Expected "+CorePreferenceKey.PREF_TESTBEDSPECIFIC_EXOGENI_RECOVER_METHOD+" to be \"single\", but it is \""+prefenceRecoverExoSM+"\"");
                //fall through: same as STATUS
            case CREATE_SLIVER:
                //fall through: same as STATUS
            case DELETE:
                //fall through: same as STATUS
            case STATUS: {
                allowed = AuthorityListModel.SubAuthMatchAllowed.ALLOW_TOPLEVEL;
                if (prefenceForceExoSM.equals("auto"))
                    exogeniChoice = ExogeniChoice.USE_GLOBAL_SETTING;
                else
                    exogeniChoice = prefenceForceExoSM.equals("yes") ? ExogeniChoice.FORCE_TOP : ExogeniChoice.FORCE_SUBAUTH;
                break;
            }
            case RECOVER_BOTH_FIRST: {
                assert prefenceRecoverExoSM.equalsIgnoreCase("both");
                allowed = AuthorityListModel.SubAuthMatchAllowed.ALLOW_TOPLEVEL;
                exogeniChoice = ExogeniChoice.FORCE_TOP;
                break;
            }
            case RECOVER_BOTH_SECOND: {
                assert prefenceRecoverExoSM.equalsIgnoreCase("both");
                allowed = AuthorityListModel.SubAuthMatchAllowed.ALLOW_TOPLEVEL;
                exogeniChoice = ExogeniChoice.FORCE_SUBAUTH;
                break;
            }
            case REQUEST_RSPEC: {
                allowed = AuthorityListModel.SubAuthMatchAllowed.ALLOW_ONLY_EXACT_SUBAUTHORITY;
                exogeniChoice = ExogeniChoice.FORCE_SUBAUTH;
                break;
            }
            case FIND_USERAUTH: {
                allowed = AuthorityListModel.SubAuthMatchAllowed.ALLOW_TOPLEVEL;
                exogeniChoice = null; //do nothing special for exogeni
                break;
            }
            default: throw new RuntimeException("BUG: Unimplemented functionality for "+purpose);
        }

        if (urn.getTopLevelAuthority_withoutSubAuth().equalsIgnoreCase("exogeni.net") && exogeniChoice != null) {
            switch (exogeniChoice) {
                case FORCE_TOP: {
                    preferred = AuthorityListModel.SubAuthMatchPreference.PREFER_TOPLEVEL;
                    break;
                }
                case FORCE_SUBAUTH: {
                    preferred = AuthorityListModel.SubAuthMatchPreference.PREFER_EXACT_SUBAUTHORITY;
                    break;
                }
                case USE_GLOBAL_SETTING: {
                    //use testbed.xml settings
                    SfaAuthority exogeniAuth = model.getByUrn(urn, AuthorityListModel.SubAuthMatchAllowed.ALLOW_OTHER_SUBAUTHORITY, AuthorityListModel.SubAuthMatchPreference.PREFER_EXACT_SUBAUTHORITY, false/*ignoreTypeAndName*/);
                    if (exogeniAuth != null && exogeniAuth.hasSpecialCase(SfaAuthority.SpecialCase.PREFER_TOPLEVEL))
                        preferred = AuthorityListModel.SubAuthMatchPreference.PREFER_TOPLEVEL;
                    else
                        preferred = AuthorityListModel.SubAuthMatchPreference.PREFER_EXACT_SUBAUTHORITY;
                    break;
                }
                default:
                    throw new RuntimeException("Unsupported exogeniChoice=" + exogeniChoice);
            }
        } else {
            preferred = AuthorityListModel.SubAuthMatchPreference.PREFER_EXACT_SUBAUTHORITY;
        }

        return model.getByUrn(urn, allowed, preferred, ignoreTypeAndName);
    }

}
