package be.iminds.ilabt.jfed.lowlevel.authority;

import be.iminds.ilabt.jfed.lowlevel.Gid;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.ServerType;
import be.iminds.ilabt.jfed.util.GeniUrn;
import org.slf4j.LoggerFactory;

import java.net.URL;
import java.util.*;

/**
 * AuthorityListModel is a list of GeniAuthority
 * It can be empty.
 * <p/>
 * Authorities can be added and removed, or "merged".
 * "merging" authorities is useful when there are 2 sources of data about an authority.
 * The data can then be combined.
 * <p/>
 * You can listen for changes in the list.
 */
public class AuthorityListModel {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(AuthorityListModel.class);
    private List<SfaAuthority> authorities;
    private List<AuthorityListModelListener> changeListeners = new ArrayList<AuthorityListModelListener>();

    private Boolean forceExoSM = null; //null means "no exception for exogeni", so do what testbeds.xml tells

    //was package scope, so that only JFedAuthorityList can create a new AuthorityListModel
    //however, the JavaFX gui needs to have an external list
    public AuthorityListModel() {
        authorities = new ArrayList<SfaAuthority>();
    }

    public List<SfaAuthority> getAuthorities() {
//        assert authorities.size() > 0;
        assert !authorities.contains(null);
        return Collections.unmodifiableList(authorities);
    }


//    public SfaAuthority getByUrn(GeniUrn urn, SubAuthMatchingMode matchingMode) {
//        return getByUrn(urn, matchingMode, forceExoSM);
//    }
//    public SfaAuthority getByUrn(GeniUrn urn, SubAuthMatchingMode matchingMode, @Nonnull ExogeniChoice exogeniChoice) {
//        return getByUrn(urn.toString(), matchingMode, forceExoSM);
//    }
//
//    private static boolean exogeniForceSM(String topTld, String subAuth, @Nonnull ExogeniChoice exogeniChoice) {
//        if (exogeniChoice == ExogeniChoice.USE_GLOBAL_SETTING) return false;
//        if (!topTld.equals("exogeni.net"))
//            return false;
//        return forceExoSM;
//    }
//    private static boolean exogeniForceNonSM(String topTld, String subAuth, @Nonnull ExogeniChoice exogeniChoice) {
//        if (exogeniChoice == ExogeniChoice.USE_GLOBAL_SETTING) return false;
//        if (!topTld.equals("exogeni.net"))
//            return false;
//        return !forceExoSM;
//    }
//
//    public void setForceExoSM(Boolean forceExoSM) {
//        this.forceExoSM = forceExoSM;
//    }

    public enum SubAuthMatchAllowed {
        ALLOW_ONLY_EXACT_SUBAUTHORITY, /*Sub authorities (or lack of sub authority) must be identical */
        ALLOW_TOPLEVEL, /* Sub authority either identical, or there exists an authority that is the same but without subauthority.
                                    (includes same matches as ALLOW_ONLY_EXACT_SUBAUTHORITY) */
        ALLOW_OTHER_SUBAUTHORITY /* match all with same authority, regardless of no or different sub authority.
                                    (includes same matches as ALLOW_OTHER_SUBAUTHORITY and ALLOW_ONLY_EXACT_SUBAUTHORITY) */
    }
    public enum SubAuthMatchPreference {
        PREFER_EXACT_SUBAUTHORITY /*Sub authorities (or lack of sub authority) must be identical */,
        PREFER_TOPLEVEL /* Returns authority without subauth, or if that is not found, the authority with sub authority will be returned if found. */,
        /* if ALLOW_OTHER_SUBAUTHORITY, an auth with other subauthority will only be returned when there is no matching toplevel or subauth athority. */
        /* if an urn without sub auth is given, both choices prefer the toplevel (as PREFER_EXACT_SUBAUTHORITY then prefers toplevel)*/
    }
    /**
     * Retreive a GeniAuthority by URN.
     *
     * @param urn    the URN of the GeniAuthority to retreive
     * @param matchAllowed which matches are allowed
     * @param matchPreferred if there are multiple matches, which is preferred?
     * @return the GeniAuthority by URN. If no authority is known for the URN, return null
     */
    public SfaAuthority getByUrn(final GeniUrn urn, SubAuthMatchAllowed matchAllowed, SubAuthMatchPreference matchPreferred, boolean ignoreUrnTypeAndName) {
        LOG.debug("AuthorityListModel.getByUrn("+urn+", "+matchAllowed+", "+matchPreferred+", "+ignoreUrnTypeAndName+")");

        if (!ignoreUrnTypeAndName && !urn.getResourceType().equals("authority"))
            LOG.warn("Tried to lookup SfaAuthority with non-authority URN \"" + urn + "\". It will likely not be found in list of authority URNs");

        List<SfaAuthority> exactMatches = new ArrayList<SfaAuthority>();
        List<SfaAuthority> topLevelMatches = new ArrayList<SfaAuthority>();
        List<SfaAuthority> ignoreSubAuthMatches = new ArrayList<SfaAuthority>();

        int urnSubAuthCount = urn.getSubAuthCount();

        for (SfaAuthority auth : authorities) {
            GeniUrn authUrn = auth.getUrn();
            int authSubAuthCount = authUrn.getSubAuthCount();

            if (!ignoreUrnTypeAndName) {
                if (!authUrn.getResourceType().equals(urn.getResourceType()))
                    continue;
                if (!authUrn.getResourceName().equals(urn.getResourceName()))
                    continue;
            }

            //if TLD doesn't match, it won't ever match
            if (!authUrn.getTopLevelAuthority_withoutSubAuth().equals(urn.getTopLevelAuthority_withoutSubAuth()))
                continue;

            //first search for exact auth and sub auth matches
            if (authUrn.getTopLevelAuthority().equals(urn.getTopLevelAuthority())) {
                exactMatches.add(auth);
            }

            //search again, and consider authorities without less sub authorities (= more top level), that are the same
            if (!authUrn.hasSubAuthority())
                topLevelMatches.add(auth); //matches, since TLD matches
            if (authSubAuthCount > 0 && authSubAuthCount < urnSubAuthCount && urn.getSubAuthNames(authSubAuthCount).equals(authUrn.getSubAuthName()))
                topLevelMatches.add(auth);

            //finally, search again ignoring sub authorities
            ignoreSubAuthMatches.add(auth); //matches, since TLD matches
        }

        switch (matchAllowed) {
            case ALLOW_ONLY_EXACT_SUBAUTHORITY: {
                topLevelMatches.clear();
                ignoreSubAuthMatches.clear();
                break;
            }
            case ALLOW_TOPLEVEL: {
                ignoreSubAuthMatches.clear();
                break;
            }
            case ALLOW_OTHER_SUBAUTHORITY:  break;
        }

        switch (matchPreferred) {
            case PREFER_EXACT_SUBAUTHORITY: {
                for (SfaAuthority auth : exactMatches)
                    return auth;
                for (SfaAuthority auth : sortBySubAuthDifference(topLevelMatches, urn))
                    return auth;
                for (SfaAuthority auth : sortBySubAuthDifference(ignoreSubAuthMatches, urn))
                    return auth;
                return null;
            }
            case PREFER_TOPLEVEL: {
                for (SfaAuthority auth : sortBySubAuthNumber(topLevelMatches))
                    return auth;
                for (SfaAuthority auth : exactMatches)
                    return auth;
                for (SfaAuthority auth : sortBySubAuthNumber(ignoreSubAuthMatches))
                    return auth;
                return null;
            }
            default: throw new RuntimeException("match preference not implemented: "+matchPreferred);
        }
    }

    /**
     * sort by distance from number of subauths of provided urn. The first has the closest possible subauth to the provided urn.
     * */
    static List<SfaAuthority> sortBySubAuthDifference(List<SfaAuthority> l, final GeniUrn urn) {
        l.sort(new Comparator<SfaAuthority>() {
            @Override
            public int compare(SfaAuthority a, SfaAuthority b) {
                return Math.abs(a.getUrn().getSubAuthCount() - urn.getSubAuthCount())
                        - Math.abs(b.getUrn().getSubAuthCount() - urn.getSubAuthCount());
            }
        });
        return l;
    }

    /**
     * Sort by number of sub auths. With smallest number of sub auths first.
     * */
    static List<SfaAuthority> sortBySubAuthNumber(List<SfaAuthority> l) {
        l.sort(new Comparator<SfaAuthority>() {
            @Override
            public int compare(SfaAuthority a, SfaAuthority b) {
                return a.getUrn().getSubAuthCount() - b.getUrn().getSubAuthCount();
            }
        });
        return l;
    }

    public SfaAuthority getByUrn(String urnString, SubAuthMatchAllowed matchAllowed, SubAuthMatchPreference matchPreferred, boolean ignoreUrnTypeAndName) {
        GeniUrn u = GeniUrn.parse(urnString);
        if (u == null) {
            LOG.warn("Tried to lookup SfaAuthority using the invalid URN \"" + urnString + "\" -> solving problem by returning null");
            return null;
        }
        return getByUrn(u, matchAllowed, matchPreferred, ignoreUrnTypeAndName);
    }
    public SfaAuthority getByUrn(String urnString, SubAuthMatchAllowed matchAllowed, SubAuthMatchPreference matchPreferred) {
        return getByUrn(urnString, matchAllowed, matchPreferred, false/*ignoreUrnTypeAndName*/);
    }
    public SfaAuthority getByUrn(GeniUrn urn, SubAuthMatchAllowed matchAllowed, SubAuthMatchPreference matchPreferred) {
        return getByUrn(urn, matchAllowed, matchPreferred, false/*ignoreUrnTypeAndName*/);
    }
    public SfaAuthority getByUrnExact(String urnString) {
        return getByUrn(urnString, AuthorityListModel.SubAuthMatchAllowed.ALLOW_ONLY_EXACT_SUBAUTHORITY, AuthorityListModel.SubAuthMatchPreference.PREFER_EXACT_SUBAUTHORITY, false/*ignoreUrnTypeAndName*/);
    }
    public SfaAuthority getByUrnExact(GeniUrn urn) {
        return getByUrn(urn, AuthorityListModel.SubAuthMatchAllowed.ALLOW_ONLY_EXACT_SUBAUTHORITY, AuthorityListModel.SubAuthMatchPreference.PREFER_EXACT_SUBAUTHORITY, false/*ignoreUrnTypeAndName*/);
    }
    public SfaAuthority getFromAnyUrn(String urnString, SubAuthMatchAllowed matchAllowed, SubAuthMatchPreference matchPreferred) {
        return getByUrn(urnString, matchAllowed, matchPreferred, true/*ignoreUrnTypeAndName*/);
    }
    public SfaAuthority getFromAnyUrn(GeniUrn urn, SubAuthMatchAllowed matchAllowed, SubAuthMatchPreference matchPreferred) {
        return getByUrn(urn, matchAllowed, matchPreferred, true/*ignoreUrnTypeAndName*/);
    }
    public SfaAuthority getFromAnyUrnExact(String urnString) {
        return getByUrn(urnString, AuthorityListModel.SubAuthMatchAllowed.ALLOW_ONLY_EXACT_SUBAUTHORITY, AuthorityListModel.SubAuthMatchPreference.PREFER_EXACT_SUBAUTHORITY, true/*ignoreUrnTypeAndName*/);
    }
    public SfaAuthority getFromAnyUrnExact(GeniUrn urn) {
        return getByUrn(urn, AuthorityListModel.SubAuthMatchAllowed.ALLOW_ONLY_EXACT_SUBAUTHORITY, AuthorityListModel.SubAuthMatchPreference.PREFER_EXACT_SUBAUTHORITY, true/*ignoreUrnTypeAndName*/);
    }



    public SfaAuthority getByUrl(URL scsUrl) {
        for (SfaAuthority auth : authorities) {
            if (auth.getUrls().values().contains(scsUrl))
                return auth;
        }
        return null;
    }


    public boolean removeByUrn(String urn) {
        for (SfaAuthority auth : authorities) {
            if (auth.getUrnString().equals(urn)) {
                authorities.remove(auth);
                fireChange();
                return true;
            }
        }
        return false;
    }

    void removeAll() {
        if (authorities.isEmpty()) return;
//        List<SfaAuthority> allAuth = new ArrayList<SfaAuthority>(authorities);
//        for (SfaAuthority auth : allAuth)
//            authorities.remove(auth);
        authorities.clear();
        fireChange();
    }

    /**
     * Does not call fireChange. do this manually afterwards!
     */
    public SfaAuthority createOrUpdate(String urn, String name, Map<ServerType, URL> urls, List<SfaAuthority.ProxyInfo> proxies, Gid gid, String type) {
        for (SfaAuthority auth : authorities) {
            if (auth.getUrnString().equals(urn)) {
                auth.updateAll(name, urls, gid, type);
                return auth;
            }
        }
        try {
            SfaAuthority newAuth = new SfaAuthority(urn, name, null, urls, proxies, gid, type);
            authorities.add(newAuth);
            return newAuth;
        } catch (JFedException e) {
            LOG.warn("WARNING: Error adding authority (=> authority not added!): \"" + urn + "\": " + e.getMessage(), e);
            return null;
        }
    }

    /**
     * Does not call fireChange. do this manually afterwards!
     */
    public SfaAuthority mergeOrAdd(SfaAuthority auth) {
        for (SfaAuthority existingAuth : authorities) {
            if (existingAuth.getUrnString().equals(auth.getUrnString())) {
                existingAuth.updateAll(auth.getName(), auth.getUrls(), auth.getGid(), auth.getType());
                return existingAuth;
            }
        }
        authorities.add(auth);
        return auth;
    }

    public void addAuthority(SfaAuthority auth) {
        assert !authorities.contains(auth);
        this.authorities.add(auth);
        fireChange();
    }

    public void fireChange() {
        for (AuthorityListModelListener l : changeListeners)
            l.onAuthorityListChanged();
    }

    public void addAuthorityListModelListener(AuthorityListModelListener l) {
        changeListeners.add(l);
    }

    public void removeAuthorityListModelListener(AuthorityListModelListener l) {
        changeListeners.remove(l);
    }

    public static interface AuthorityListModelListener {
        //        public void onAuthorityAdded();
        public void onAuthorityListChanged();
    }
}
