package be.iminds.ilabt.jfed.lowlevel.connection;

import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.util.JFedTrustStore;
import org.apache.http.impl.client.DefaultHttpClient;

import java.security.KeyStore;
import java.security.PrivateKey;
import java.security.cert.X509Certificate;
import java.util.List;

/**
 * SfaSslWithClientAuthConnection: SSL HTTP
 */
public class RestSslClientAuthConnection extends HttpConnection {
    private DefaultHttpClient httpClient;

    /**
     * Create a connection. This constructor assumes the url is to the specified sfaAuthority and the client authentication uses the specified user
     *
     * @param serverUrl               url to connect to (may not be null)
     * @param sfaAuthority            the authority connected to this URL. Will be used to get trusted certificate info, and to fill in debug info. (may not be null)
     * @param user                    the user that will be used to authenticate this connection. This users certificate and private key will be used, and it will be used for debugging. (may not be null)
     * @param proxyInfo               the proxy to use (may be null if no proxy)
     * @param handleUntrustedCallback the callback to use when the servers certificate is not trusted. This is typically used to communicate with the user about untrusted self-signed certificates. (this may be null, which will cause connection failure for untrusted certificates)
     * @param debugMode               if true, additional logging is enabled
     * @throws be.iminds.ilabt.jfed.lowlevel.JFedException
     */
    public RestSslClientAuthConnection(String serverUrl,
                                       SfaAuthority sfaAuthority,
                                       GeniUser user,
                                       ProxyInfo proxyInfo,
                                       HandleUntrustedCallback handleUntrustedCallback,
                                       boolean debugMode) throws JFedException {
        this(serverUrl, new JFedTrustStore(sfaAuthority), user.getClientCertificateChain(),
                user.getPrivateKey(), proxyInfo, handleUntrustedCallback, new DebugInfo(debugMode, sfaAuthority, user));
    }

    /**
     * Create a connection. This constructor assumes the url is to the specified sfaAuthority and the client authentication uses the specified user
     * This constructor does the same as {@link be.iminds.ilabt.jfed.lowlevel.connection.SfaSslWithClientAuthConnection#SfaSslWithClientAuthConnection(String, be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority, be.iminds.ilabt.jfed.lowlevel.GeniUser, be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection.ProxyInfo, HandleUntrustedCallback, boolean)} but with debugMode set to false.
     *
     * @param serverUrl               url to connect to (may not be null)
     * @param sfaAuthority            the authority connected to this URL. Will be used to get trusted certificate info, and to fill in debug info. (may not be null)
     * @param user                    the user that will be used to authenticate this connection. This users certificate and private key will be used, and it will be used for debugging. (may not be null)
     * @param proxyInfo               the proxy to use (may be null if no proxy)
     * @param handleUntrustedCallback the callback to use when the servers certificate is not trusted. This is typically used to communicate with the user about untrusted self-signed certificates. (this may be null, which will cause connection failure for untrusted certificates)
     * @throws be.iminds.ilabt.jfed.lowlevel.JFedException
     */
    public RestSslClientAuthConnection(String serverUrl,
                                       SfaAuthority sfaAuthority,
                                       GeniUser user,
                                       ProxyInfo proxyInfo,
                                       HandleUntrustedCallback handleUntrustedCallback) throws JFedException {
        this(serverUrl, sfaAuthority, user, proxyInfo, handleUntrustedCallback, false);
    }

    /**
     * Create a connection by specifying all low level details needed.
     *
     * @param serverUrl               url to connect to (may not be null)
     * @param jFedTrustStore          The trust store to use to verify the server's identity. (may not be null, you can always use the {@link be.iminds.ilabt.jfed.util.JFedTrustStore#JFedTrustStore()} constructor to use the system trust store)
     * @param clientCertificateChain  the certificate (or chain of certificates) used to identify the client at the server. (may not be null)
     * @param privateKey              the client his private key used to setup a client authenticated connection (must match public key in certificate) (may not be null)
     * @param proxyInfo               the proxy to use (may be null if no proxy)
     * @param handleUntrustedCallback the callback to use when the servers certificate is not trusted. This is typically used to communicate with the user about untrusted self-signed certificates. (this may be null, which will cause connection failure for untrusted certificates)
     * @param debugInfo               debug information, such as authority and user, and whether debug mode is enabled. (may be null)
     * @throws JFedException
     */
    public RestSslClientAuthConnection(String serverUrl,
                                       JFedTrustStore jFedTrustStore,
                                       List<X509Certificate> clientCertificateChain,
                                       PrivateKey privateKey,
                                       ProxyInfo proxyInfo,
                                       HandleUntrustedCallback handleUntrustedCallback,
                                       DebugInfo debugInfo) throws JFedException {
        super(serverUrl, jFedTrustStore, proxyInfo, debugInfo);

        error = true;


        if (serverUrl == null) throw new IllegalArgumentException("Illegal argument: serverURL == null");
        if (clientCertificateChain == null)
            throw new IllegalArgumentException("Illegal argument: clientCertificateChain == null");
        if (clientCertificateChain.isEmpty())
            throw new IllegalArgumentException("Illegal argument: clientCertificateChain is empty");
        if (privateKey == null) throw new IllegalArgumentException("Illegal argument: privateKey == null");

        KeyStore trustStore = jFedTrustStore.getTrustStore();

        try {
            httpClient = HttpsClientWithUserAuthenticationFactory.getHttpClient(
                    proxyInfo,
                    clientCertificateChain,
                    privateKey,
                    serverUrl,
                    trustStore,
                    jFedTrustStore.getAllowedServerCertificateHostnameAliases(),
                    this.debugInfo.isDebugMode(),
                    handleUntrustedCallback
            );
            error = false;
        } catch (Exception e) {
            throw new JFedException("Error creating XmlRpcClient in GeniConnection constructor: " + e.getMessage(), e);
        }
    }

    @Override
    public DefaultHttpClient getHttpClient() {
        assert !debugInfo.isDebugMode() :
                "Error: This connection is a fake debugging connection, yet the httpClient was requested";
        return httpClient;
    }

    public void close() {
        error = true;
        httpClient = null;
    }

}