package be.iminds.ilabt.jfed.lowlevel.connection;

import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.util.JFedTrustStore;
import be.iminds.ilabt.jfed.util.SSLCertificateDownloader;
import be.iminds.ilabt.jfed.util.SocksProxyHelper;
import org.apache.http.conn.scheme.Scheme;
import org.apache.http.conn.ssl.SSLSocketFactory;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.params.BasicHttpParams;
import org.apache.http.params.CoreConnectionPNames;
import org.apache.http.params.HttpParams;
import org.slf4j.LoggerFactory;

import java.net.MalformedURLException;
import java.net.URL;
import java.security.KeyStore;
import java.security.SecureRandom;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * SfaSslWithClientAuthConnection: SSL HTTP
 */
public class RestSslNoAuthConnection extends HttpConnection {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(RestSslNoAuthConnection.class);
    private static int extraTrustCount = 7000;

    private DefaultHttpClient httpClient;

    /**
     * Create a connection. This constructor assumes the url is to the specified authority and creates in the trust store based on that.
     *
     * @param serverUrlStr url to connect to (may not be null)
     * @param authority    the authority connected to this URL. Will be used to get trusted certificate info, and to fill in debug info. (may not be null)
     * @param proxyInfo    the proxy to use (may be null if no proxy)
     * @param debugMode    if true, additional logging is enabled
     * @throws be.iminds.ilabt.jfed.lowlevel.JFedException
     */
    public RestSslNoAuthConnection(String serverUrlStr, SfaAuthority authority, ProxyInfo proxyInfo, boolean debugMode) throws JFedException {
        this(serverUrlStr, new JFedTrustStore(authority), proxyInfo, new DebugInfo(debugMode, authority, null), null);
    }

    /**
     * Create a connection by specifying all low level details needed.
     *
     * @param serverUrlStr   url to connect to (may not be null)
     * @param jFedTrustStore The trust store to use to verify the server's identity. (may not be null, you can always use the {@link be.iminds.ilabt.jfed.util.JFedTrustStore#JFedTrustStore()} constructor to use the system trust store)
     * @param proxyInfo      the proxy to use (may be null if no proxy)
     * @param debugInfo      debug information, such as authority and user, and whether debug mode is enabled. (may be null)
     */
    public RestSslNoAuthConnection(String serverUrlStr,
                                   JFedTrustStore jFedTrustStore,
                                   ProxyInfo proxyInfo,
                                   DebugInfo debugInfo,
                                   HandleUntrustedCallback handleUntrustedCallback) throws JFedException {
        super(serverUrlStr, jFedTrustStore, proxyInfo, debugInfo);

        LOG.debug("initialising RestSslPasswordConnection to " + serverUrlStr);

        final Collection<String> allowedCertificateHostnameAliasesOrig = jFedTrustStore.getAllowedServerCertificateHostnameAliases();
        final List<String> allowedCertificateHostnameAliases =
                new ArrayList<String>(allowedCertificateHostnameAliasesOrig);

        error = true;

        if (serverUrlStr == null) throw new IllegalArgumentException("Illegal argument: serverURL == null");

        URL serverUrl;
        try {
            serverUrl = new URL(serverUrlStr);
        } catch (MalformedURLException e) {
            LOG.error("ERROR: MalformedURLException url=\"" + serverUrlStr + "\"", e);
            error = true;
            return;
        }

        HttpParams params = new BasicHttpParams();
        params.setParameter(CoreConnectionPNames.SO_TIMEOUT, 120000); //read timeout in ms (default: 0 = infinite)   (long timeout, as some calls take time to finish)
        params.setParameter(CoreConnectionPNames.CONNECTION_TIMEOUT, 10000); //connection timeout in ms (default: 0 = infinite)

        if (System.getProperty("proxySet") != null &&
                System.getProperty("proxySet").equals("true") &&
                System.getProperty("socksProxyHost") != null)
            httpClient = SocksProxyHelper.getHttpClientOverSocksProxy(params);
        else
            httpClient = new DefaultHttpClient(params);

        String algo = SSLSocketFactory.TLS;
        SecureRandom random = new SecureRandom();

        KeyStore trustStore = jFedTrustStore.getTrustStore();

        SSLSocketFactory socketFactory = null;
        try {
            //was::
//            socketFactory = new SSLSocketFactory(trustStore);
            if (handleUntrustedCallback != null) {
                //                socketFactory = new SSLSocketFactory(TRUSTALL, new AllowAllHostnameVerifier()); //doesn't work as we need client authentication

                //didn't work for some reason. Probably it's missing client certificate (need to check apache source code of SSLSocketFactory to correctly write this)
                //                socketFactory = new MyInsecureSSLAcceptAllSocketFactory(algo, keyStore, privatekey_store_pass/*password is used to encrypt the key inside the keystore*/, trustStore, random, hostnameVerifier);

                //Solved by connecting 2 times. Not best solution, but it works.
                SSLCertificateDownloader.SSLCertificateJFedInfo certInfo =
                        SSLCertificateDownloader.getCertificateInfo(serverUrl);
                if (certInfo != null && certInfo.isSelfSigned() != null && certInfo.isSelfSigned()) {
                    if (handleUntrustedCallback.trust(certInfo)) {
                        trustStore.setCertificateEntry("allTrustCert" + extraTrustCount, certInfo.getCert());
                        if (!certInfo.getSubjectMatchesHostname())
                            allowedCertificateHostnameAliases.add(certInfo.getSubject());
                    } else {
                        LOG.info("User does not trust certificate -> Not adding anything to trust store.");
                    }
                }
            }

            org.apache.http.conn.ssl.X509HostnameVerifier hostnameVerifier = HttpsClientWithUserAuthenticationFactory.makeHostNameVerifier(allowedCertificateHostnameAliases);
            socketFactory = new SSLSocketFactory(
                    "TLS", null,
                    null, trustStore,
                    null, null,
                    hostnameVerifier/*default: BROWSER_COMPATIBLE_HOSTNAME_VERIFIER*/);
        } catch (Exception e) {
            LOG.error("error creating SSLSocketFactory", e);
            throw new RuntimeException(e);
        }
        Scheme sch1 = new Scheme("https", 443, socketFactory); //port is the default port for the given protocol
        httpClient.getConnectionManager().getSchemeRegistry().register(sch1);

        int port = serverUrl.getPort();
        if (serverUrl.getPort() <= 0) port = serverUrl.getDefaultPort();

        LOG.debug("Constructed RestSslPasswordConnection");
        error = false;
    }

    @Override
    public DefaultHttpClient getHttpClient() {
        assert !debugInfo.isFakeForDebugging() :
                "Error: This connection is a fake debugging connection, yet the httpClient was requested";
        return httpClient;
    }

    public void close() {
        error = true;
        httpClient = null;
    }

}
