package be.iminds.ilabt.jfed.lowlevel.connection;

import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.ServerType;
import be.iminds.ilabt.jfed.lowlevel.api.*;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.fedmon.Fedmon;
import be.iminds.ilabt.jfed.preferences.CorePreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.preferences.PreferenceChangeListener;
import be.iminds.ilabt.jfed.util.ExtraInfoCallback;
import be.iminds.ilabt.jfed.util.JFedTrustStore;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.inject.Singleton;
import java.lang.management.ManagementFactory;
import java.lang.management.ThreadInfo;
import java.lang.management.ThreadMXBean;
import java.net.URL;
import java.util.*;

/**
 * SfaConnectionPool
 *
 * A connection pool used for reusing connections. Users of this class can request connections, and the SfaConnectionPool
 * will either create them, or return an already existing connection.
 *
 * Connections with errors are automatically removed.
 *
 * Only connections with the same settings will be selected for reuse. See {@link be.iminds.ilabt.jfed.lowlevel.connection.SfaConnectionPool.ConnectionSpecification} for details.
 * This includes these settings: authority, serverType, disableClientAuth, proxyInfo
 *
 * The SfaConnectionPool also stores a bae trust store, to use for all new connections.
 * When connections to an authority are requested, a trust store to which the certificates of that authority are added will be used.
 * When connections without any specified authority are used, the caller can specify a trust store to use (or null to use the SfaConnectionPool trust store).
 *
 * The SfaConnectionPool stores default proxy info to use. Some calls allow this default to be overwritten.
 */
@Singleton
public class SfaConnectionPool implements JFedConnectionProvider {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(SfaConnectionPool.class);
    private final Map<ConnectionSpecification, JFedConnection> conPool;
    private JFedConnection.ProxyInfo defaultProxy = null;

    /* debug will print out a info during the call */
    private boolean debugMode = false;

    private final JFedTrustStore baseTrustStore;
    private final ConnectionFactory connectionFactory;

    @Inject
    public SfaConnectionPool(JFedTrustStore baseTrustStore, final JFedPreferences jFedPreferences, ConnectionFactory connectionFactory) {
        conPool = new HashMap<ConnectionSpecification, JFedConnection>();
        this.baseTrustStore = baseTrustStore;
        this.connectionFactory = connectionFactory;
        startAutoExpire();

        if (jFedPreferences.hasKey(CorePreferenceKey.PREF_DEBUG_CPU)) {
            setDebugCpuUsage(jFedPreferences.getBoolean(CorePreferenceKey.PREF_DEBUG_CPU));
            jFedPreferences.addPreferenceChangeListener(CorePreferenceKey.PREF_DEBUG_CPU, new PreferenceChangeListener() {
                @Override
                public void onPreferenceChanged() {
                    setDebugCpuUsage(jFedPreferences.getBoolean(CorePreferenceKey.PREF_DEBUG_CPU));
                }
            });
        }
    }

    public int size() {
        synchronized (conPool) {
            return conPool.size();
        }
    }

    public JFedConnection.ProxyInfo getDefaultProxy() {
        return defaultProxy;
    }
    public void setDefaultProxy(JFedConnection.ProxyInfo defaultProxy) {
        this.defaultProxy = defaultProxy;
    }

    /**
     * debug mode will print out a info during the call
     */
    public boolean isDebugMode() {
        return debugMode;
    }

    /**
     * Forget all cached connections, so that existing connections cannot be reused and new connections will be made.
     */
    public void flush() {
        //we don't want to close active connections
//        for (JFedConnection con : conPool.values())
//            con.close();

        synchronized (conPool) {
            conPool.clear();
        }
    }

    /**
     * debug mode will print out a info during the call
     */
    public void setDebugMode(boolean debugMode) {
        boolean prevMode = this.debugMode;
        this.debugMode = debugMode;

        if (debugMode != prevMode)
            flush();
    }

    @Override
    public JFedConnection getConnectionByAuthority(GeniUser user, SfaAuthority authority, Class targetClass) throws JFedException {
        return getConnectionByAuthority(user, authority, targetClass, false);
    }

    public static ServerType classToServerType(Class targetClass) {
        if (targetClass.equals(AggregateManager3.class))
            return new ServerType(ServerType.GeniServerRole.AM, 3);
        if (targetClass.equals(AggregateManager2.class))
            return new ServerType(ServerType.GeniServerRole.AM, 2);
        if (targetClass.equals(ProtogeniSliceAuthority.class))
            return new ServerType(ServerType.GeniServerRole.PROTOGENI_SA, 1);
        if (targetClass.equals(ProtoGeniClearingHouse1.class))
            return new ServerType(ServerType.GeniServerRole.PROTOGENI_CH, 1);
        if (targetClass.equals(PlanetlabSfaRegistryInterface.class))
            return new ServerType(ServerType.GeniServerRole.PlanetLabSliceRegistry, 1);
        if (targetClass.equals(StitchingComputationService.class))
            return new ServerType(ServerType.GeniServerRole.SCS, 1);
        if (targetClass.equals(Fedmon.class))
            return new ServerType(ServerType.GeniServerRole.Fedmon, 1);
        if (targetClass.equals(FederationRegistryApi1.class))
            return new ServerType(ServerType.GeniServerRole.GENI_CH, 1);
        if (targetClass.equals(FederationMemberAuthorityApi1.class))
            return new ServerType(ServerType.GeniServerRole.GENI_CH_MA, 1);
        if (targetClass.equals(FederationSliceAuthorityApi1.class))
            return new ServerType(ServerType.GeniServerRole.GENI_CH_SA, 1);
        if (targetClass.equals(FederationRegistryApi2.class))
            return new ServerType(ServerType.GeniServerRole.GENI_CH, 2);
        if (targetClass.equals(FederationMemberAuthorityApi2.class))
            return new ServerType(ServerType.GeniServerRole.GENI_CH_MA, 2);
        if (targetClass.equals(FederationSliceAuthorityApi2.class))
            return new ServerType(ServerType.GeniServerRole.GENI_CH_SA, 2);
        if (targetClass.equals(OCCI.class))
            return new ServerType(ServerType.GeniServerRole.OCCI, 1);
        return null;
    }

    public JFedConnection getConnectionByAuthority(GeniUser user, SfaAuthority authority, Class targetClass, boolean disableClientAuthentication) throws JFedException {
//        assert user != null;
        assert authority != null;
        assert targetClass != null;

        if (authority == null) throw new RuntimeException("authority == null");
        ServerType serverType = classToServerType(targetClass);
        if (serverType != null)
            return getConnectionByAuthority(user, authority, serverType, disableClientAuthentication);
        if (targetClass.equals(ProtoGeniAMExtensions.class)) {
            //AMv2 OR AMv3
            JFedConnection con =
                    getConnectionByAuthority(user, authority, new ServerType(ServerType.GeniServerRole.AM, 2), disableClientAuthentication);
            if (con != null) return con;
            return getConnectionByAuthority(user, authority, new ServerType(ServerType.GeniServerRole.AM, 3), disableClientAuthentication);
        }
        throw new JFedException("Cannot get connection for " + targetClass.getName());
    }

    public JFedConnection getConnectionByUserAuthority(GeniUser user, ServerType serverType) throws JFedException {
        return getConnectionByUserAuthority(user, serverType, false);
    }

    public JFedConnection getConnectionByUserAuthority(GeniUser user, ServerType serverType, boolean disableClientAuthentication) throws JFedException {
        assert serverType != null;
        assert user != null;

        assert user.getUserAuthority() != null :
                "getConnectionByUserAuthority requires that the user's authority is known";
        SfaAuthority authority = user.getUserAuthority();
        assert authority != null;

        return getConnectionByAuthority(user, authority, serverType, disableClientAuthentication);
    }

    /* TODO: connections by URL are not cached yet*/
    @Override
    public JFedConnection getConnectionByUrl(JFedTrustStore overrideTrustStore, GeniUser user, URL serverUrl, HandleUntrustedCallback handleUntrustedCallback, ServerType serverType, SfaAuthority optionalAuthorityForDebugInfo) throws JFedException {
        return getConnectionByUrl(overrideTrustStore, user, serverUrl, handleUntrustedCallback, serverType, false/*disableClientAuthentication*/, false/*overrideProxy*/, null, optionalAuthorityForDebugInfo);
    }

    public JFedConnection getConnectionByUrl(JFedTrustStore overrideTrustStore, GeniUser user, URL serverUrl, HandleUntrustedCallback handleUntrustedCallback, ServerType serverType, JFedConnection.ProxyInfo overrideProxyInfo, SfaAuthority optionalAuthorityForDebugInfo) throws JFedException {
        return getConnectionByUrl(overrideTrustStore, user, serverUrl, handleUntrustedCallback, serverType, false/*disableClientAuthentication*/, true/*overrideProxy*/, overrideProxyInfo, optionalAuthorityForDebugInfo);
    }

    public JFedConnection getConnectionByUrl(JFedTrustStore overrideTrustStore, GeniUser user, URL serverUrl, HandleUntrustedCallback handleUntrustedCallback, ServerType serverType, boolean disableClientAuthentication, SfaAuthority optionalAuthorityForDebugInfo) throws JFedException {
        return getConnectionByUrl(overrideTrustStore, user, serverUrl, handleUntrustedCallback, serverType, disableClientAuthentication, false/*overrideProxy*/, null, optionalAuthorityForDebugInfo);
    }

    public JFedConnection getConnectionByUrl(JFedTrustStore overrideTrustStore, GeniUser user, URL serverUrl, HandleUntrustedCallback handleUntrustedCallback, ServerType serverType, boolean disableClientAuthentication, boolean overrideProxy, JFedConnection.ProxyInfo overrideProxyInfo, SfaAuthority optionalAuthorityForDebugInfo) throws JFedException {
        assert serverType != null;

        final JFedTrustStore conTrustStore = overrideTrustStore == null ? baseTrustStore : overrideTrustStore;

        JFedConnection.ProxyInfo usedProxyInfo;
        if (overrideProxy)
            usedProxyInfo = overrideProxyInfo;
        else
            usedProxyInfo = defaultProxy;

        JFedConnection con = null;

        if (serverUrl == null)
            throw new RuntimeException("getConnectionByUrl: serverUrl null for type: " + serverType + "    url: " + serverUrl);

        if (serverType.getRole().equals(ServerType.GeniServerRole.Fedmon)) {
            assert serverUrl.getProtocol().equals("https") : "Only HTTPS is supported for a FedMon connection";
            con = connectionFactory.createRestSslNoAuthConnection(
                    serverUrl.toString(), conTrustStore,
                    usedProxyInfo, handleUntrustedCallback,
                    new JFedConnection.DebugInfo(debugMode, optionalAuthorityForDebugInfo/*auth*/, user));
        } else {
            if (serverType.getRole().equals(ServerType.GeniServerRole.OCCI)) {
                assert serverUrl.getProtocol().equals("https") : "Only HTTPS is supported for an OCCI connection";
                ExtraInfoCallback.Login bonfireLogin = ExtraInfoCallback.getBonfireLogin();
                if (bonfireLogin == null) return null;
                con = connectionFactory.createRestSslPasswordConnection(
                        serverUrl.toString(), conTrustStore,
                        usedProxyInfo, bonfireLogin,
                        handleUntrustedCallback,
                        new JFedConnection.DebugInfo(debugMode, optionalAuthorityForDebugInfo/*auth*/, user));
            } else {
                if (serverUrl.getProtocol().equals("https")) {
                    if (!disableClientAuthentication) {
                        assert user != null : "HTTPS connections require that a user is logged in";

                        con = connectionFactory.createSfaSslWithClientAuthConnection(serverUrl.toString(), conTrustStore, user.getClientCertificateChain(),
                                user.getPrivateKey(), usedProxyInfo, handleUntrustedCallback,
                                new JFedConnection.DebugInfo(debugMode, optionalAuthorityForDebugInfo/*auth*/, user));
                    } else
                        con = connectionFactory.createSfaSslWithoutClientAuthConnection(serverUrl.toString(), conTrustStore,
                                usedProxyInfo, handleUntrustedCallback, new JFedConnection.DebugInfo(debugMode, optionalAuthorityForDebugInfo/*auth*/, user));
                } else
                    con = connectionFactory.createSfaPlainHttpConnection(serverUrl.toString(), usedProxyInfo,
                            new JFedConnection.DebugInfo(debugMode, optionalAuthorityForDebugInfo/*auth*/, user));
            }
        }

        assert con != null;
        return con;
    }

    @Override
    public JFedConnection getConnectionByAuthority(GeniUser user, SfaAuthority authority, ServerType serverType) throws JFedException {
        return getConnectionByAuthority(user, authority, serverType, false, false, null);
    }

    public JFedConnection getConnectionByAuthority(GeniUser user, SfaAuthority authority, ServerType serverType,
                                                   boolean disableClientAuthentication) throws JFedException {
        return getConnectionByAuthority(user, authority, serverType, disableClientAuthentication, false, null);
    }
    public JFedConnection getConnectionByAuthority(GeniUser user, SfaAuthority authority, ServerType serverType,
                                                   boolean disableClientAuthentication, JFedConnection.ProxyInfo overwriteProxyInfo) throws JFedException {
        return getConnectionByAuthority(user, authority, serverType, disableClientAuthentication, true, overwriteProxyInfo);
    }
    /**
     * @param overrideProxy if true, use parameter overwriteProxyInfo as proxyInfo (even if null!). If false use proxy settings in SfaConnectionPool (and overwriteProxyInfo is ignored)
     * @param overwriteProxyInfo the proxy info to overwrite the default SfaConnectionPool proxy info with. Note that this can be null while overrideProxy=true, to disable proxy.
     * */
    public JFedConnection getConnectionByAuthority(GeniUser user, SfaAuthority authority, ServerType serverType,
                                                   boolean disableClientAuthentication,
                                                   boolean overrideProxy, JFedConnection.ProxyInfo overwriteProxyInfo) throws JFedException {
        JFedConnection.ProxyInfo usedProxyInfo = overrideProxy ? overwriteProxyInfo : defaultProxy;

        assert serverType != null;
        if (authority == null) throw new RuntimeException("authority == null");
        assert authority != null;

        ConnectionSpecification conSpec = new ConnectionSpecification(user, authority, serverType, disableClientAuthentication, usedProxyInfo);
        return getConnectionByConnectionSpecification(conSpec);
    }

    public JFedConnection getConnectionByConnectionSpecification(ConnectionSpecification connectionSpecification) throws JFedException {
        final GeniUser user = connectionSpecification.getUser();
        final SfaAuthority authority = connectionSpecification.getAuthority();
        final ServerType serverType = connectionSpecification.getServerType();
        final boolean disableClientAuthentication = connectionSpecification.isDisableClientAuth();
        final JFedConnection.ProxyInfo usedProxyInfo = connectionSpecification.getProxyInfo();

        assert authority != null;

        JFedConnection con;
        synchronized (conPool) {
            con = conPool.get(connectionSpecification);
        }

        if (con == null)
            LOG.trace("No cached connection for " + connectionSpecification);
        if (con != null && (con.isError() || con.isExpired())) {
            //something went wrong, we'll kill the connection

            synchronized (conPool) {
                conPool.remove(connectionSpecification);
            }
            con.close();
            con = null;
            LOG.debug("NOT reusing cached connection for " + connectionSpecification + " due to error in connection.");
        }
        if (con != null)
            LOG.trace("Re-using cached connection for " + connectionSpecification);

        if (con == null) {
            JFedConnection.JFedConnectionFactory conCopier;
            final URL serverUrl = authority.getUrl(serverType);
            if (serverUrl == null)
                throw new RuntimeException("Connection type has no URL: serverType: " + serverType + "   authority: " + (
                        authority == null ? "null" : authority.getUrn()));

            final JFedTrustStore conTrustStore = new JFedTrustStore(baseTrustStore);

            if (serverType.getRole().equals(ServerType.GeniServerRole.Fedmon)) {
                assert serverUrl.getProtocol().equals("https") : "Only HTTPS is supported for a FedMon connection";

                if (authority.hasPemSslTrustCerts()) {
                    conTrustStore.addTrustedPemCertificateIfNotAdded(authority.getPemSslTrustCerts());
                    conTrustStore.addAllowedServerCertificateHostnameAliases(authority.getAllowedCertificateHostnameAliases());
                }
                if (user.getUserAuthority() != null) {
                    if (user.getUserAuthority().hasPemSslTrustCerts())
                        conTrustStore.addTrustedPemCertificateIfNotAdded(user.getUserAuthority().getPemSslTrustCerts());
                    conTrustStore.addAllowedServerCertificateHostnameAliases(user.getUserAuthority().getAllowedCertificateHostnameAliases());
                }

                conCopier = new JFedConnection.JFedConnectionFactory() {
                    @Override
                    public JFedConnection create() throws JFedException {
                        JFedConnection res = connectionFactory.createRestSslNoAuthConnection(serverUrl.toString(), conTrustStore, usedProxyInfo,
                                null/*handleUntrustedCallback*/, new JFedConnection.DebugInfo(debugMode, authority/*auth*/, user));
                        assert res != null;
                        res.setConnectionFactory(this);
                        return res;
                    }
                };
            } else {
                if (serverType.getRole().equals(ServerType.GeniServerRole.OCCI)) {
                    assert serverUrl.getProtocol().equals("https") : "Only HTTPS is supported for an OCCI connection";
                    assert user != null : "HTTPS connections require that a user is logged in";

                    if (authority.hasPemSslTrustCerts()) {
                        conTrustStore.addTrustedPemCertificateIfNotAdded(authority.getPemSslTrustCerts());
                        conTrustStore.addAllowedServerCertificateHostnameAliases(authority.getAllowedCertificateHostnameAliases());
                    }
                    if (user.getUserAuthority() != null) {
                        if (user.getUserAuthority().hasPemSslTrustCerts())
                            conTrustStore.addTrustedPemCertificateIfNotAdded(user.getUserAuthority().getPemSslTrustCerts());
                        conTrustStore.addAllowedServerCertificateHostnameAliases(user.getUserAuthority().getAllowedCertificateHostnameAliases());
                    }

                    LOG.debug("Getting bonfire login info");
                    final ExtraInfoCallback.Login bonfireLogin = ExtraInfoCallback.getBonfireLogin();
                    if (bonfireLogin == null) {
                        LOG.error("Failed to get bonfire login info, will not create connection.");
                        return null;
                    }


                    conCopier = new JFedConnection.JFedConnectionFactory() {
                        @Override
                        public JFedConnection create() throws JFedException {
                            JFedConnection res = connectionFactory.createRestSslPasswordConnection(serverUrl.toString(), conTrustStore, usedProxyInfo, bonfireLogin,
                                    null/*handleUntrustedCallback*/, new JFedConnection.DebugInfo(debugMode, authority/*auth*/, user));
                            assert res != null;
                            res.setConnectionFactory(this);
                            return res;
                        }
                    };

                } else {
                    if (serverUrl.getProtocol().equals("https")) {
                        if (authority.hasPemSslTrustCerts()) {
                            conTrustStore.addTrustedPemCertificateIfNotAdded(authority.getPemSslTrustCerts());
                            conTrustStore.addAllowedServerCertificateHostnameAliases(authority.getAllowedCertificateHostnameAliases());
                        }
                        if (user.getUserAuthority() != null) {
                            if (user.getUserAuthority().hasPemSslTrustCerts())
                                conTrustStore.addTrustedPemCertificateIfNotAdded(user.getUserAuthority().getPemSslTrustCerts());
                            conTrustStore.addAllowedServerCertificateHostnameAliases(user.getUserAuthority().getAllowedCertificateHostnameAliases());
                        }

                        if (!disableClientAuthentication) {
                            assert user != null : "HTTPS connections require that a user is logged in";
                            conCopier = new JFedConnection.JFedConnectionFactory() {
                                @Override
                                public JFedConnection create() throws JFedException {
                                    JFedConnection res = connectionFactory.createSfaSslWithClientAuthConnection(
                                            serverUrl.toString(),
                                            conTrustStore,
                                            user.getClientCertificateChain(),
                                            user.getPrivateKey(),
                                            usedProxyInfo,
                                            null/*handleUntrustedCallback*/,
                                            new JFedConnection.DebugInfo(debugMode, authority, user)
                                    );
                                    assert res != null;
                                    res.setConnectionFactory(this);
                                    return res;
                                }
                            };
                        } else {
                            conCopier = new JFedConnection.JFedConnectionFactory() {
                                @Override
                                public JFedConnection create() throws JFedException {
                                    JFedConnection res = connectionFactory.createSfaSslWithoutClientAuthConnection(
                                            serverUrl.toString(),
                                            conTrustStore,
                                            usedProxyInfo,
                                            null/*handleUntrustedCallback*/,
                                            new JFedConnection.DebugInfo(debugMode, authority, user)
                                    );
                                    assert res != null;
                                    res.setConnectionFactory(this);
                                    return res;
                                }
                            };
                        }
                    } else {
                        conCopier = new JFedConnection.JFedConnectionFactory() {
                            @Override
                            public JFedConnection create() throws JFedException {
                                JFedConnection res = connectionFactory.createSfaPlainHttpConnection(serverUrl.toString(), usedProxyInfo,
                                        new JFedConnection.DebugInfo(debugMode, authority, user));
                                assert res != null;
                                res.setConnectionFactory(this);
                                return res;
                            }
                        };
                    }
                }
            }

            assert conCopier != null;
            con = conCopier.create(); //use copier to get initial connection itself
            assert con != null;
            con.setConnectionFactory(conCopier);
            assert con.canDuplicate();
            assert con.getConnectionFactory() != null;

            //reuse connection next time, if allowed
            if (!authority.isReconnectEachTime())
                synchronized (conPool) {
                    conPool.put(connectionSpecification, con);
                }
        }

        assert con == null || con.getDebugInfo().getAuth() != null : "con.getDebugInfo() = "+con.getDebugInfo();

        return con;
    }

    public void clear() {
        synchronized (conPool) {
            for (JFedConnection conn : conPool.values()) {
                LOG.debug("Closing connection to {}", conn.getServerUrl());
                conn.close();
            }

            conPool.clear();
        }
        LOG.debug("Cleared connection pool");
    }

    private long lastThreadCpuDump = 0;

    private Map<Long, Long> prev_thread_cpuNs_sinceStart_ByThreadId = new HashMap<Long, Long>();
    private long prevCpuNsSinceStart = 0;
    private final long firstDate = System.currentTimeMillis();
    private long prevDate = System.currentTimeMillis();

    private boolean debugCpuUsage = false;

    public void clearExpired() {
        int connectionsCleared = 0;

        //TODO: perhaps not a good idea to lock conPool while doing this. (though this should be a fast operation)
        synchronized (conPool) {

            for (Iterator<Map.Entry<ConnectionSpecification, JFedConnection>> it =
                         conPool.entrySet().iterator(); it.hasNext();) {
                Map.Entry<ConnectionSpecification, JFedConnection> entry =  it.next();
                JFedConnection con = entry.getValue();
                if (con.isExpired()) {
                    LOG.debug("Closing expired connection to {}. inUse={} lastUse={} ms ago",
                            con.getServerUrl(),
                            con.isInUse(),
                            (System.currentTimeMillis() - con.getLastCallTime().getTime()));
                    con.close();
                    it.remove();
                    connectionsCleared++;
                }
            }
        }

        int conCount = conPool.size();

        if (connectionsCleared > 0)
            LOG.debug("Periodical check of expired connections in connection pool: Closed "+connectionsCleared+" ("+conCount+" remain open)");
        else
            LOG.debug("Periodical check of expired connections in connection pool: Did not find any. ("+conCount+" remain open)");

//        if (true) return;
        if (debugCpuUsage)
            dumpCpuUsage();
    }

    public void setDebugCpuUsage(boolean debugCpuUsage) {
        this.debugCpuUsage = debugCpuUsage;
    }

    public void dumpCpuUsage() {
        ThreadMXBean tmxb = ManagementFactory.getThreadMXBean();
        if (!tmxb.isThreadCpuTimeSupported())
            LOG.debug("Thread CPU measurements are not supported");
        else {
            boolean dumpThreadCpuNow = (System.currentTimeMillis() - lastThreadCpuDump) > 60000; //once each minute

            StringBuilder logString = new StringBuilder();
            StringBuilder highCpulogString = new StringBuilder();

            long[] threadIDs = tmxb.getAllThreadIds();
            long cpuNsSinceStart = 0;
            for (int i = 0; i < threadIDs.length; i++) {
                long threadID = threadIDs[i];
                long cpuTime = tmxb.getThreadCpuTime(threadID);
                cpuNsSinceStart += cpuTime;
            }
            long curDate = System.currentTimeMillis();
            long realMsSincePrev = curDate - prevDate;
            long realMsSinceStart = curDate - firstDate;
            long cpuNsSincePrev = cpuNsSinceStart - prevCpuNsSinceStart;

            double cpuPercentOfReal_sinceStart = (cpuNsSinceStart*100.0)/(realMsSinceStart*1000000.0);
            double cpuPercentOfReal_sincePrev = (cpuNsSincePrev*100.0)/(realMsSincePrev*1000000.0);

            for (int i = 0; i < threadIDs.length; i++) {
                long threadID = threadIDs[i];
                ThreadInfo threadInfo = tmxb.getThreadInfo(threadID);
                if (threadInfo != null) {
                    String threadName = threadInfo.getThreadName();
                    long thread_cpuNs_sinceStart = tmxb.getThreadCpuTime(threadID); //in nanoseconds

                    Long prev_thread_cpuNs_sinceStart = prev_thread_cpuNs_sinceStart_ByThreadId.get(threadID); //in nanoseconds
                    if (prev_thread_cpuNs_sinceStart == null) prev_thread_cpuNs_sinceStart = 0L;

                    long thread_cpuNs_sincePrev = thread_cpuNs_sinceStart - prev_thread_cpuNs_sinceStart;

                    double thread_cpuPercentOfProcess_sinceStart = (thread_cpuNs_sinceStart*100.0)/cpuNsSinceStart;
                    double thread_cpuPercentOfProcess_sincePrev = (thread_cpuNs_sincePrev*100.0)/cpuNsSincePrev;

                    double thread_cpuPercentOfReal_sinceStart = (thread_cpuNs_sinceStart*100.0)/(realMsSinceStart*1000000.0);
                    double thread_cpuPercentOfReal_sincePrev = (thread_cpuNs_sincePrev*100.0)/(realMsSincePrev*1000000.0);


                    logString.append("Thread ").append(threadID).append(" \"").append(threadName).append("\"\n")
                            .append("    from start:  cpuNs=").append(thread_cpuNs_sinceStart).append(" ")
                            .append(" (").append(thread_cpuPercentOfProcess_sinceStart).append("% of process cputime)")
                            .append(" (").append(thread_cpuPercentOfReal_sinceStart).append("% of real time)\n")
                            .append("    from prev:  cpuNS=").append(thread_cpuNs_sincePrev).append(" ")
                            .append(" (").append(thread_cpuPercentOfProcess_sincePrev).append("% of process cputime)")
                            .append(" (").append(thread_cpuPercentOfReal_sincePrev).append("% of real time)\n");

                    if (thread_cpuPercentOfProcess_sinceStart > 20 || thread_cpuPercentOfProcess_sincePrev > 20) {
                        highCpulogString.append("NOTE: HIGH CPU USE Thread ").append(threadID).append(" \"").append(threadName).append("\"").append("\n")
                                .append(" from start:  cpuNs=").append(thread_cpuNs_sinceStart).append(" ")
                                .append(" (").append(thread_cpuPercentOfProcess_sinceStart)
                                .append("% of process cputime)").append(" (").append(thread_cpuPercentOfReal_sinceStart)
                                .append("% of real time)").append("\n")
                                .append("  from prev:  cpuNS=").append(thread_cpuNs_sincePrev).append(" ")
                                .append(" (").append(thread_cpuPercentOfProcess_sincePrev)
                                .append("% of process cputime)")
                                .append(" (").append(thread_cpuPercentOfReal_sincePrev)
                                .append("% of real time)").append("\n")
                                .append("       Running for ").append(realMsSinceStart).append(" ms.     Used ")
                                .append(cpuNsSinceStart).append(" ns of cpu time in total. =").append(cpuPercentOfReal_sinceStart).append("%").append("\n")
                                .append("Last CPU print was ").append(realMsSincePrev).append(" ms ago. Used ").append(cpuNsSincePrev).append(" ns of cpu time in that time. =").append(cpuPercentOfReal_sincePrev).append("%\n");
                    }

                    prev_thread_cpuNs_sinceStart_ByThreadId.put(threadID, thread_cpuNs_sinceStart);
                }
            }
            logString.append("       Running for " + realMsSinceStart + " ms.     Used " + cpuNsSinceStart + " ns of cpu time in total. =" + cpuPercentOfReal_sinceStart + "%\n");
            logString.append("Last CPU print was " + realMsSincePrev + " ms ago. Used " + cpuNsSincePrev + " ns of cpu time in that time. =" + cpuPercentOfReal_sincePrev + "%");
            prevCpuNsSinceStart = cpuNsSinceStart;
            prevDate = curDate;

            if (dumpThreadCpuNow)
                LOG.info("Thread overview:\n"+logString.toString());
            else {
                if (LOG.isDebugEnabled())
                    LOG.debug("Thread overview:\n"+logString.toString());
                else
                    LOG.warn(highCpulogString.toString());
            }

            if (dumpThreadCpuNow)
                lastThreadCpuDump = System.currentTimeMillis();
        }
    }

    private void startAutoExpire() {
        TimerTask expireConnectionsTask = new TimerTask() {
            @Override
            public void run() {
                clearExpired();
            }
        };

        //execute clearExpired each 10 seconds
        Timer timer = new Timer("SfaConnectionPool-AutoExpire",true /*is daemon*/);
        timer.schedule(
                expireConnectionsTask,
                10* 1000L/*first delay (ms)*/,
                10 * 1000L /*period (ms)*/);
    }

    private class ConnectionSpecification {
        private final GeniUser user;
        private final SfaAuthority authority;
        private final ServerType serverType;
        private final boolean disableClientAuth;
        private final JFedConnection.ProxyInfo proxyInfo;

        private ConnectionSpecification(GeniUser user, SfaAuthority authority, ServerType serverType, boolean disableClientAuth, JFedConnection.ProxyInfo proxyInfo) {
            if (authority == null) throw new RuntimeException("authority == null");
            this.user = user;
            this.authority = authority;
            this.serverType = serverType;
            this.disableClientAuth = disableClientAuth;
            this.proxyInfo = proxyInfo; //may be null
        }

        public SfaAuthority getAuthority() {
            return authority;
        }

        public ServerType getServerType() {
            return serverType;
        }

        public boolean isDisableClientAuth() {
            return disableClientAuth;
        }

        public JFedConnection.ProxyInfo getProxyInfo() {
            return proxyInfo;
        }

        public GeniUser getUser() {
            return user;
        }

        @Override
        public boolean equals(Object o) {
            if (this == o) return true;
            if (o == null || getClass() != o.getClass()) return false;

            ConnectionSpecification that = (ConnectionSpecification) o;

            if ((user == null) != (that.user == null)) return false;
            if ((user != null) && ((user.getUserUrn() == null) != (that.user.getUserUrn() == null))) return false;
            if (user != null && user.getUserUrn() != null && !user.getUserUrn().equals(that.user.getUserUrn())) return false;
            if ((authority == null) != (that.authority == null)) return false;
            if (!authority.equals(that.authority)) return false;
            if ((proxyInfo == null) != (that.proxyInfo == null)) return false;
            if (proxyInfo != null && that.proxyInfo != null && !proxyInfo.equals(that.proxyInfo)) return false;
            if (serverType != that.serverType) return false;
            if (disableClientAuth != that.disableClientAuth) return false;

            return true;
        }

        @Override
        public int hashCode() {
            int result = authority.hashCode();
            if (proxyInfo != null) result = 31 * result + proxyInfo.hashCode();
            result = 31 * result + serverType.hashCode();
            result = 31 * result + (disableClientAuth ? 1 : 0);
            return result;
        }

        @Override
        public String toString() {
            return "ConnectionSpecification{" +
                    "user=" + (user == null ? "null" : "") +
                    "authority=" + authority +
                    ", serverType=" + serverType +
                    ", disableClientAuth=" + disableClientAuth +
                    ", defaultProxy=" + proxyInfo +
                    '}';
        }
    }
}
