package be.iminds.ilabt.jfed.rspec.model;

import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.util.TextUtil;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.transform.stream.StreamSource;
import java.io.StringReader;
import java.util.*;

/**
 * BasicStringRspec provides some very basic RSpec parsing, without requiring the use of ModelRspec (which is in the "rspec" module).
 * <p/>
 * See the rspec module for advanced RSpec parsing.
 * (the rspec module uses java7 (or later) with javaFX (for Property and other javaFX "model" classes, it does not use any javaFX GUI code).
 * But since it uses javaFX, it does require that any jar that references the rspec module jar, correctly includes the javaFX classpath.
 * This can only be done using a javaFX packager. See http://stackoverflow.com/questions/15577735/manifest-classpath-for-jar-files )
 */
public class BasicStringRspec {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(BasicStringRspec.class);
    protected final String xmlString;
    protected boolean wellFormed;
    protected boolean validRspec;
    protected Boolean isPossiblyStitching = null;
    protected final List<Set<String>> stitchingCms = new ArrayList<Set<String>>(); //only links with linktype "lan" are added to this
    protected List<String> allComponentManagers;

    public BasicStringRspec(String xmlString) {
        this.xmlString = xmlString;
        try {
            wellFormed = true;
            validRspec = true;
            parse();
        } catch (Exception e) {
            LOG.warn("XML is not well-formed and cannot be parsed: " + e.getMessage(), e);
            wellFormed = false;
        }
    }

    //TODO: move RspecXmlConstanst to where I can statically import these from it:
    public final static String NAMESPACE_RSPEC = "http://www.geni.net/resources/rspec/3";
    public final static String NAMESPACE_OPENFLOW = "http://www.geni.net/resources/rspec/ext/openflow/3";
    public final static QName Q_RSPEC = new QName(NAMESPACE_RSPEC, "rspec");
    public final static QName Q_NODE = new QName(NAMESPACE_RSPEC, "node");
    public final static QName Q_NODE_AVAILABLE = new QName(NAMESPACE_RSPEC, "available");
    public final static QName Q_ATT_NODE_AVAILABLE_NOW = new QName("now");
    public final static QName Q_LINK = new QName(NAMESPACE_RSPEC, "link");
    public final static QName Q_INTERFACE = new QName(NAMESPACE_RSPEC, "interface");
    public final static QName Q_ATT_CLIENT_ID = new QName("client_id");
    public final static QName Q_ATT_COM_MAN_ID = new QName("component_manager_id");
    public final static QName Q_ATT_COM_ID = new QName("component_id");
    public final static QName Q_LINK_COM_MAN = new QName(NAMESPACE_RSPEC, "component_manager");
    public final static QName Q_ATT_LINK_COM_MAN_NAME = new QName("name");
    public final static QName Q_LINK_TYPE = new QName(NAMESPACE_RSPEC, "link_type");
    public final static QName Q_ATT_LINK_TYPE_NAME = new QName("name");
    public final static QName Q_NODE_SERVICES = new QName(NAMESPACE_RSPEC, "services");
    public final static QName Q_NODE_SERVICES_LOGIN = new QName(NAMESPACE_RSPEC, "login");
    public final static QName Q_ATT_NODE_SERVICES_LOGIN_AUTH = new QName("authentication");
    public final static QName Q_ATT_NODE_SERVICES_LOGIN_HOSTNAME = new QName("hostname");
    public final static QName Q_ATT_NODE_SERVICES_LOGIN_PORT = new QName("port");
    public final static QName Q_ATT_NODE_SERVICES_LOGIN_USERNAME = new QName("username");

    /**
     * Very fault tolerant parsing.
     * <p/>
     * Parses ONLY openflow component managers
     */
    public static List<String> parseOpenflowComponentManagers(String xmlString) {
        //LOG.trace("parseOpenflowComponentManagers starts");
        Set<String> allComponentManagersSet = new HashSet<String>();

        //namespace: "http://www.geni.net/resources/rspec/ext/openflow/3"

        //found 2 componentManagerUrns:
        // <rspec><openflow:sliver><openflow:group><openflow:datapath component_manager_id="urn:publicid:IDN+openflow:foam:foam.instageni.gpolab.bbn.com+authority+am"...
        // <rspec><openflow:datapath component_manager_id="

        //we could simply do: look for any openflow node, and find component_manager_id in it.
        //                    then look for any datapath node, and find component_manager_id in it.
        //
        //But we do it even moer simple: look for any element in the openflow namespace, and find component_manager_id in it.

        try {
            XMLInputFactory xif = XMLInputFactory.newFactory();
            StreamSource xml = new StreamSource(new StringReader(xmlString));
            XMLStreamReader streamReader = xif.createXMLStreamReader(xml);

            while (streamReader.hasNext()) {
                streamReader.next();
                switch (streamReader.getEventType()) {
                    case XMLStreamReader.START_ELEMENT: {
                        if (streamReader.getName().getNamespaceURI().equals(NAMESPACE_OPENFLOW)) {
                            String comManId = streamReader.getAttributeValue(null, Q_ATT_COM_MAN_ID.getLocalPart());
                            if (comManId != null)
                                allComponentManagersSet.add(comManId);
                        }
                        break;
                    }
                    case XMLStreamReader.CHARACTERS: {
                        break;
                    }
                    default:
                        break;
                }
            }
        } catch (XMLStreamException e) {
            //ignore any error parsing RSpec
            LOG.trace("Error parsing openflow RSpec: ignored", e);
        }

        return new ArrayList<String>(allComponentManagersSet);
    }

    public String getXmlString() {
        return xmlString;
    }

    public boolean isWellFormed() {
        return wellFormed;
    }

    public boolean isValidRspec() {
        return wellFormed && validRspec;
    }

    /**
     * Very fault tolerant parsing.
     * <p/>
     * This initialises:
     * - wellFormed
     * - isStitching
     * - allComponentManagers
     */
    protected void parse() {
        Set<String> allComponentManagersSet = new HashSet<String>();

        wellFormed = true;
        validRspec = true;
        int rspecCount = 0;
        try {
            XMLInputFactory xif = XMLInputFactory.newFactory();
            StreamSource xml = new StreamSource(new StringReader(xmlString));
            XMLStreamReader streamReader = xif.createXMLStreamReader(xml);

            int link = 0;
            Set<String> curLinkComMan = new HashSet<String>();
            int rspecLevel = 0;

            boolean currentLinkCannotBeStitching = false;

            while (streamReader.hasNext()) {
                streamReader.next();
                switch (streamReader.getEventType()) {
                    case XMLStreamReader.END_ELEMENT: {
                        if (streamReader.getName().equals(Q_RSPEC)) {
                            rspecLevel--;
                        }
                        if (streamReader.getName().equals(Q_LINK)) {
                            if (curLinkComMan.size() > 1) {
                                if (!currentLinkCannotBeStitching) {
                                    isPossiblyStitching = true;
                                    stitchingCms.add(curLinkComMan);
                                }
                            }
                            curLinkComMan = new HashSet<String>();
                        }
                        break;
                    }
                    case XMLStreamReader.START_ELEMENT: {
                        if (streamReader.getName().equals(Q_RSPEC)) {
                            //rspecs without any rspec node are not wel formed.
                            rspecCount++;
                            rspecLevel++;
                        }
                        if (rspecLevel > 0 && streamReader.getName().equals(Q_NODE)) {
                            String comManId = streamReader.getAttributeValue(null, Q_ATT_COM_MAN_ID.getLocalPart());
                            if (comManId != null)
                                allComponentManagersSet.add(comManId);
                            break;
                        }
                        if (rspecLevel > 0 && streamReader.getName().equals(Q_LINK)) {
                            link++;
                            currentLinkCannotBeStitching = false;
                            break;
                        }
                        if (rspecLevel > 0 && streamReader.getName().equals(Q_LINK_TYPE)) {
                            String linkTypeName = streamReader.getAttributeValue(null, Q_ATT_LINK_TYPE_NAME.getLocalPart());
                            if (linkTypeName != null && !linkTypeName.equals("lan"))
                                currentLinkCannotBeStitching = true;
                        }
                        if (rspecLevel > 0 && streamReader.getName().equals(Q_LINK_COM_MAN)) {
                            String comManId =
                                    streamReader.getAttributeValue(null, Q_ATT_LINK_COM_MAN_NAME.getLocalPart());
                            if (comManId != null) {
                                allComponentManagersSet.add(comManId);
                                curLinkComMan.add(comManId);
                            }
                            break;
                        }
                        break;
                    }
                    case XMLStreamReader.CHARACTERS: {
                        break;
                    }
                    default:
                        break;
                }
            }
        } catch (XMLStreamException e) {
            LOG.error("Error parsing rspec. Will be ignored: " + e.getMessage(), e);
            wellFormed = false;
            return;
        }
        if (rspecCount <= 0)
            validRspec = false;
        if (isPossiblyStitching == null)
            isPossiblyStitching = false;

        allComponentManagersSet.addAll(parseOpenflowComponentManagers(xmlString));

        allComponentManagers = new ArrayList<String>(allComponentManagersSet);

        assert stitchingCms.isEmpty() == !isPossiblyStitching;
    }

    public List<String> getAllComponentManagerUrns() {
        if (allComponentManagers != null) return allComponentManagers;
        if (!isWellFormed()) return null;
        parse();
        return allComponentManagers;
    }

    public List<SfaAuthority> getAllComponentManager(AuthorityListModel authListModel, AuthorityFinder authorityFinder, AuthorityFinder.Purpose purpose) {
        List<String> allComponentManagerUrns = getAllComponentManagerUrns();
        if (allComponentManagerUrns == null) return null;
        List<SfaAuthority> res = new ArrayList<SfaAuthority>();
        for (String cm : allComponentManagerUrns) {
            SfaAuthority auth = authorityFinder.findByUrn(cm, purpose);
            res.add(auth);
        }
        return res;
    }

    /**
     * does any link have multiple component managers?
     *
     * AuthorityListModel is used to determine if two different URNs are actually the same authority.
     * This happens for exogeni for example. In that case, a link with multiple component managers can be a non-stitched link!
     *
     * if authorityListModel is null, it is assumed that any link with multiple component managers is a stitched link
     */
    public Boolean isStitching(AuthorityListModel authorityListModel, AuthorityFinder authorityFinder) {
        if (isPossiblyStitching == null) {
            if (!isWellFormed()) return null;
            parse();
        }
        if (!isPossiblyStitching) return false;
        if (authorityListModel == null)  return isPossiblyStitching;
        //possibly stitching, we need to check

        assert !stitchingCms.isEmpty();
        for (Set<String> urns : stitchingCms) {
            SfaAuthority firstAuth = null;
            for (String urn : urns) {
                //TODO this is actually complexer, because it will depend on exogeni settings...
                SfaAuthority auth = authorityListModel.getByUrn(urn, AuthorityListModel.SubAuthMatchAllowed.ALLOW_TOPLEVEL, AuthorityListModel.SubAuthMatchPreference.PREFER_TOPLEVEL);
                if (auth == null) {
                    LOG.warn("Authority not found for "+urn+" during stitch check -> falling back to assuming link is stitched");
                    return true; //fall back to assuming it is stitched, if any authority is not found
                }
                if (firstAuth == null)
                    firstAuth = auth;
                else {
                    if (!firstAuth.getUrn().equals(auth.getUrn()))
                        return true; //authorities have different URN, so stitched link
                }
            }
        }
        return false;
    }


    public static String makeUniqueNodeId(String clientId, String componentId, String componentManagerId) {
        String compManPart = "@" + (componentManagerId == null ? "unspecified" : componentManagerId);

        if (clientId != null)
            return clientId + compManPart;
        else
            return (componentId == null ? "nodeWithoutAnyId" : componentId) + compManPart;
    }

    public static String makeUniqueLinkId(String clientId, String componentId, Collection<String> componentManagerIds) {
        String compManPart;
        if (componentManagerIds != null && !componentManagerIds.isEmpty()) {
            compManPart = "@";
            List<String> sortedComponentManagerIds = new ArrayList<String>(componentManagerIds);
            Collections.sort(sortedComponentManagerIds);
            boolean first = true;
            for (String componentManagerId : sortedComponentManagerIds) {
                if (first)
                    compManPart += "" + componentManagerId;
                else
                    compManPart += "&" + componentManagerId;
                first = false;
            }
        } else
            compManPart = "@unspecified";

        if (clientId != null)
            return clientId + compManPart;
        else
            return (componentId == null ? "linkWithoutAnyId" : componentId) + compManPart;
    }

    public static String makeUniqueInterfaceId(String clientId, String componentId, String nodeUniqueId) {
        String atPart = nodeUniqueId == null ? "@unknownNode" : "@" + nodeUniqueId;

        if (clientId != null)
            return clientId + atPart;
        else
            return (componentId == null ? "ifaceWithoutAnyId" : componentId) + atPart;
    }


    /**
     * helper to only extract login info for a specific node, without full parsing and thus no need for ModelRspec
     *
     * @param nodeUniqueId the client_id of the node
     */
    public List<LoginService> findNodeLoginInfoByUniqueId(String nodeUniqueId) {
        //TODO: fallback to clientId only if nodeUniqueId fails
        assert nodeUniqueId != null;
        if (!isWellFormed()) return null;

        List<LoginService> res = new ArrayList<LoginService>();

        try {
            XMLInputFactory xif = XMLInputFactory.newFactory();
            StreamSource xml = new StreamSource(new StringReader(xmlString));
            XMLStreamReader streamReader = xif.createXMLStreamReader(xml);

            boolean atCorrectNode = false;
            while (streamReader.hasNext()) {
                streamReader.next();
                switch (streamReader.getEventType()) {
                    case XMLStreamReader.START_ELEMENT: {
                        if (streamReader.getName().equals(Q_NODE)) {
                            String clientId = streamReader.getAttributeValue(null, Q_ATT_CLIENT_ID.getLocalPart());
                            String compManId = streamReader.getAttributeValue(null, Q_ATT_COM_MAN_ID.getLocalPart());
                            String uniqueId = makeUniqueNodeId(
                                    clientId == null ? null : clientId.trim(),
                                    null,
                                    compManId == null ? null : compManId.trim()
                            );
                            if (uniqueId.equals(nodeUniqueId.trim())) {
                                atCorrectNode = true;
                            }
                            break;
                        }
                        if (atCorrectNode && streamReader.getName().equals(Q_NODE_SERVICES_LOGIN)) {
                            String attAuth =
                                    streamReader.getAttributeValue(null, Q_ATT_NODE_SERVICES_LOGIN_AUTH.getLocalPart());
                            String attHostname =
                                    streamReader.getAttributeValue(null, Q_ATT_NODE_SERVICES_LOGIN_HOSTNAME.getLocalPart());
                            String attPort =
                                    streamReader.getAttributeValue(null, Q_ATT_NODE_SERVICES_LOGIN_PORT.getLocalPart());
                            String attUsername =
                                    streamReader.getAttributeValue(null, Q_ATT_NODE_SERVICES_LOGIN_USERNAME.getLocalPart());

                            try {
                                res.add(new LoginService(attAuth, attHostname,
                                        attPort == null ? 22 : Integer.parseInt(attPort), attUsername));
                            } catch (NumberFormatException e) {
                                LOG.error("Found matching login details but with non numeric port (\"" + attPort + "\") -> These login details will be ignored.", e);
                            }
                            break;
                        }
                        break;
                    }
                    case XMLStreamReader.END_ELEMENT: {
                        if (streamReader.getName().equals(Q_NODE)) {
                            atCorrectNode = false;
                        }
                    }
                    default:
                        break;
                }
            }
        } catch (XMLStreamException e) {
            LOG.error("Error parsing rspec. Will be ignored: " + e.getMessage(), e);
        }

        return res;
    }

    public List<BasicNodeInfo> getBasicNodeInfo() {
        if (!isWellFormed()) return null;

        List<BasicNodeInfo> res = new ArrayList<BasicNodeInfo>();
        try {
            XMLInputFactory xif = XMLInputFactory.newFactory();
            StreamSource xml = new StreamSource(new StringReader(xmlString));
            XMLStreamReader streamReader = xif.createXMLStreamReader(xml);

            String comManId = null;
            String comId = null;
            String clientId = null;
            Boolean available = null;

            while (streamReader.hasNext()) {
                streamReader.next();
                switch (streamReader.getEventType()) {
                    case XMLStreamReader.START_ELEMENT: {
                        if (streamReader.getName().equals(Q_NODE)) {
                            comManId = streamReader.getAttributeValue(null, Q_ATT_COM_MAN_ID.getLocalPart());
                            comId = streamReader.getAttributeValue(null, Q_ATT_COM_ID.getLocalPart());
                            clientId = streamReader.getAttributeValue(null, Q_ATT_CLIENT_ID.getLocalPart());
                        }
                        if (streamReader.getName().equals(Q_NODE_AVAILABLE)) {
                            available = TextUtil.objectToBoolean(streamReader.getAttributeValue(null, Q_ATT_NODE_AVAILABLE_NOW.getLocalPart()));
                        }
                        break;
                    }
                    case XMLStreamReader.END_ELEMENT: {
                        if (streamReader.getName().equals(Q_NODE)) {
                            if (clientId != null || comId != null || comManId != null )
                                res.add(new BasicNodeInfo(clientId, comId, comManId, available));

                            comManId = null;
                            comId = null;
                            clientId = null;
                            available = null;
                        }
                    }
                    default:
                        break;
                }
            }
        } catch (XMLStreamException e) {
            LOG.error("Error parsing rspec. Will be ignored: " + e.getMessage(), e);
        }

        return res;
    }

    public static class LoginService {
        private final String authentication;
        private final String hostname;
        private final int port;
        private final String username;

        public LoginService(String authentication, String hostname, int port, String username) {
            assert (authentication != null);
            assert (hostname != null);
            assert (username != null);
            this.authentication = authentication;
            this.hostname = hostname;
            this.port = port;
            this.username = username;
        }

        public String getAuthentication() {
            return authentication;
        }

        public String getHostname() {
            return hostname;
        }

        public int getPort() {
            return port;
        }

        public String getUsername() {
            return username;
        }

        @Override
        public String toString() {
            return "LoginService{" +
                    "authentication='" + authentication + '\'' +
                    ", hostname='" + hostname + '\'' +
                    ", port=" + port +
                    ", username='" + username + '\'' +
                    '}';
        }
    }

    public static class BasicNodeInfo {
        private final String clientId;
        private final String componentId;
        private final String componentManagerId;
        private final Boolean available;

        public BasicNodeInfo(@Nullable String clientId, @Nullable String componentId, @Nullable String componentManagerId, @Nullable Boolean available) {
            this.clientId = clientId;
            this.componentId = componentId;
            this.componentManagerId = componentManagerId;
            this.available = available;
        }

        public String getClientId() {
            return clientId;
        }

        public String getComponentId() {
            return componentId;
        }

        public String getComponentManagerId() {
            return componentManagerId;
        }

        public Boolean isAvailable() {
            return available;
        }

        public String getUniqueId() {
            return makeUniqueNodeId(
                    clientId == null ? null : clientId.trim(),
                    componentId == null ? null : componentId.trim(),
                    componentManagerId == null ? null : componentManagerId.trim()
            );
        }
    }
}
