package be.iminds.ilabt.jfed.rspec.convert;

import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.parser.ExtraXml;
import be.iminds.ilabt.jfed.rspec.parser.RspecXmlConstants;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collections;
import java.util.ListIterator;

/**
 * RspecConverter
 */
public class RspecConverter {
    private static final Logger LOG = LoggerFactory.getLogger(RspecConverter.class);

    /**
     * Change the provided manifest RSpec into a request RSpec.
     *
     * @param source the Rspec to change. The original is NOT changed.
     * @param keepNodeBindings keep bound nodes are not
     * @return a new request ModelRspec
     */
    public ModelRspec convertManifestToRequest(ModelRspec source, boolean keepNodeBindings, RspecFactory rspecFactory) {
        if (source.getType().equals("request")) {
            LOG.warn("convertManifestToRequest was asked to convert a request rspec. This will be ignored, and we will act as if it was a manifest.");
        }
        if (source.getType().equals("advertisement")) {
            LOG.warn("convertManifestToRequest was asked to convert an advertisement rspec. This will be ignored, and we will act as if it was a manifest.");
        }
        ModelRspec res = rspecFactory.copyModelRspec(source, "request");

        //remove manifest specific stuff (and add request specific stuff?)
        for (RspecNode node : res.getNodes()) {
            node.setComponentName(null);
            if (!keepNodeBindings)
                node.setComponentId((String) null);
            node.setLoginServices(Collections.<LoginService>emptyList());
            node.setSliverId((String)null);

            for (ListIterator<ExtraXml> it = node.getExtraXml().listIterator(); it.hasNext();) {
                ExtraXml extraXml = it.next();
                if (extraXml.getStartElementName().equals(RspecXmlConstants.Q_NODE_HOST))
                    it.remove();
                if (extraXml.getStartElementName().equals(RspecXmlConstants.Q_NODE_VNODE))
                    it.remove();
            }

            for (RspecInterface iface : node.getInterfaces()) {
                iface.setSliverId((String) null);
                iface.setMacAddress(null);
                iface.setComponentId(null); //note that node bindings can be kept by this method, but interface bindings can not. (add that functionality if needed)
            }
        }

        for (RspecLink link : res.getLinks()) {
            link.setSliverId((String) null);

            for (RspecInterface iface : link.getInterfaces()) {
                iface.setSliverId((String) null);
                iface.setComponentId(null); //note that node bindings can be kept by this method, but interface bindings can not. (add that functionality if needed)
                iface.setMacAddress(null); //does not actually appear in manifest
            }
        }
        for (ListIterator<ExtraXml> it = res.getExtraXml().listIterator(); it.hasNext();) {
            ExtraXml extraXml = it.next();
            //remove stitching data. It makes no sense in user request. (SCS adds it)
            if (extraXml.getStartElementName().equals(RspecXmlConstants.Q_STITCHING))
                it.remove();
        }

        return res;
    }

    /**
     * Create a copy in which the Component manager bindings of all nodes, links and interfaces are removed.
     * @param source
     * @param rspecFactory
     * @return
     */
    public ModelRspec removeComponentManagerBindings(ModelRspec source, RspecFactory rspecFactory) {
        ModelRspec res = rspecFactory.copyModelRspec(source, null);

        //remove manifest specific stuff (and add request specific stuff?)
        for (RspecNode node : res.getNodes()) {
            node.setComponentName(null);
            node.setComponentId((String) null);
            node.setComponentManagerId((String) null);

            for (RspecInterface iface : node.getInterfaces()) {
                iface.setComponentId(null);
            }
        }

        for (RspecLink link : res.getLinks()) {
            link.getComponentManagerUrns().clear();

            for (RspecInterface iface : link.getInterfaces()) {
                iface.setComponentId(null);
            }
        }

        return res;
    }
}
